program calc (input, output);
(* a calculator that propagates errors

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

*)

const

(* begin module version *)
   version = 2.71; (* of calc.p 2014 Oct 03
2014 Oct 03, 2.71: give pointer to calcline
2014 Oct 03, 2.70: improve log documentation
2013 Jun 09, 2.69: Dudley's Criteria, '?' command
2012 Nov 01, 2.68: truncate function
2008 Jul 25, 2.67: backup
2008 Jul 25, 2.66: record crash bug
2007 Mar 21, 2.65: update simpson
2.64, 2003 Jul 15: remove warning if : mode is off
2.60, 2002 Aug 17: bug in push: spare must be initialized!
2.60, 2001 Jan  2: calc now compiles under gpc
2.60, 2000 Apr 13: In showingerror mode, may skip second number.
2.59, 2000 Apr 12: major bug found! computation of error for
   logs was WRONG.  For base 2 logs, for example, the error
   should have been divided by 1/ln(2).  All computations
   based on versions less than 2.59 MUST BE REDONE.
2.58, 2000 Jan 11: L command introduced
previous changes: 1997 April 8
origin 1987 Sep 8 *)
(* end module version *)

(* begin module describe.calc *)
(*
name
   calc: a calculator that propagates errors

synopsis
   calc(input: in, output: out)

files
   input: reverse polish calculator input
   output: results

description

   The program is based on the idea of the dc program under UNIX.  That
   program takes input as reverse polish and calculates values.  This program
   does the same, but values have estimates so one may calculate and
   propagate errors.  Tokens (commands and numbers) are usually separated by
   spaces or carriage returns.  Tokens that begin with a digit or a dash (-)
   are numbers.  Numbers always come in pairs, the first is the estimate and
   the second is the error.

   Some of the commands are:
     '?' or 'h' give current list of all commands and functions
     numbers (as pairs) are entered on the stack
             5 2
          means 5 +/- 2
             5'
          means 5 +/- 0, so you can avoid giving the estimate if you want.
          any other legal command may replace the single quote as "5p".
     + add the top two numbers on the stack together
     _ (UNDERSCORE) subtract the top number from the next number
       on the stack (underscore is used to be distinct from minus sign, -)
     * multiply the top two numbers on the stack together
     / divide the top number on the stack by the next number on the stack
     s print the stack, top down
     p print the top number on the stack
     . do not show error ranges and do not read new ones
     : do show and do error ranges

   Note:  When the program is asked to do calculations silently,
   (using the t command) it immediately shuts up and does not say that it
   is doing so.  This makes it easier to write programs without having them
   announce in the output that they are doing silent calculations.

   If one has turned on showing error ranges with the ":" command, then
   numbers must be input as pairs, the estimate and the error.  However, if a
   second number is not found during reading, the error is set to zero.

documentation
   An Introduction to Error Analysis,  John R. Taylor
   University Science Books, Mill Valley, CA. 1982.

see also

   {Tcsh script to call calc from the command line:}
   https://alum.mit.edu/www/toms/ftp/calcline

author
   Thomas Schneider

bugs
   Pascal numeric input is used, so anything that can make Pascal
   bomb will bomb this program.  For example, "- ", will cause
   the program to think there is a number after the dash, and (our)
   Pascal will object.  This should be protected against now,
   so the program should never bomb (famous last words).
   2008 Jul 25
   ... Indeed famous.  A simple string like '31.' causes the GPCC
   compiler to stop with:
   "calc: digit expected after decimal point (error #554 at 186db)"
   As this is done by a direct read of the number,
   it cannot be prevented without writing additional protective code.

   The u (uncertainty) function error estimate is set to zero when
   the probability is zero.  This is a guess.

   Version 2.59, 2000 Apr 12: major bug found! computation of error for logs
   was WRONG.  For base 2 logs, for example, the error should have been
   divided by 1/ln(2).  All computations of log errors based on versions less
   than 2.59 MUST BE REDONE.

technical notes
   The Gaussian integration function will create a real number
   that is too large if the Z value is larger than constant zbound.
   The program returns zero in these cases, to avoid bombing.

*)
(* end module describe.calc *)

(* begin module calc.const *)
   defrwidth = 9;   (* default width of numbers *)
   defrdecimal = 2; (* default decimal places of numbers *)
   zbound = 38; (* largest z that the g function can tolerate without bombing
      The value zbound = 38 works for a Sun Sparc workstation. *)
   pmstring1 = ' \pm '; (* latex format *)
   pmstring2 = ' +/- '; (* regular format *)
(* end module calc.const *)

type
(* begin module calc.type *)
   range = record (* representation of x +/- error *)
      estimate: real; (* the central value x *)
      error: real; (* the most probable values above and below x *)
   end;

   stackptr = ^stack;
   stack = record (* a stack of numbers *)
      n: range; (* a number *)
      next: stackptr (* next numbers on the stack *)
   end;
(* end module calc.type *)

var
(* begin module calc.var *)
   base: real; (* the base for taking logarithms *)
   notation: boolean; (* true = scientific notation on output *)
   rwidth: integer;   (* width of numbers *)
   rdecimal: integer; (* decimal places of numbers *)
   printer: char; (* ' ': silent, 'p': print top of stack all the time,
                     's': print whole stack all the time *)
   showingerrors: boolean; (* true if error ranges are shown *)
   latex: boolean; (* if true show "\pm" otherwise use "+/-" for range. *)
   pmstring: packed array[1..5] of char; (* string to show depending on latex
                                            variable *)
   fromto: boolean;   (* true = [5 to 11] display,
                         false = 8 +/- 3  display *)

   thestack: stackptr; (* the global stack on which numbers are kept *)
   spare: stackptr; (* spare records for the stack *)
   carriagereturn: boolean; (* to carriage return or not on lines *)
(* end module calc.var *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module skipblanks *)
procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
      while (thefile^ = ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
      while (thefile^ <> ' ') and not eoln(thefile) do get(thefile);
end;
(* end module skipblanks version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module setrange *)
procedure setrange(f: range; var t: range);
(* set t to the value of f (from and to) *)
begin
   t.estimate := f.estimate;
   t.error := f.error;
end;
(* end module setrange *)

(* begin module readrange *)
procedure readrange(var fin: text; var r: range; var ok: boolean;
                    showingerrors: boolean);
(* read the range from fin, if the number is ok, then ok is true
otherwise ok is false.  If the range is in from-to form, then 
read that way.  If it is in mean +/- deviation, then read that way.
If showingerrors is true, accept a non-numeric for the range and
assume error is zero. *)
var
   hold: real; (* temporary storage for reversing the numbers *)
   first, second: real; (* the first and second numbers read in *)

procedure getnumber(var f: text; var number: real; var ok: boolean);
(* get a number from file f, return ok=true if the number was read properly *)
var
   negative: boolean; (* true if the number is to be negative *)
begin
   (* read the negative sign separately as a protection; otherwise
      a person could type '- 5' and the program will bomb *)
   if f^='-' then begin
      get(f);
      negative := true
   end
   else negative := false;
 
   if not (f^ in ['0','1','2','3','4','5','6','7','8','9'])
   then begin
{
      writeln(output,'badly formed number, not read');
}
      writeln(output,'Error set to zero');
      number := 0.0; (* force zero return *)
      if showingerrors
      then ok := true (* accept it and set to zero *)
      else ok := false;
   end
   else begin
      ok := true;
      read(f, number);
      if negative then number := - number;
   end;
end;
begin
   getnumber(fin,first,ok);
   if ok then begin
      (* skip to the next number if we are reading them *)
      if showingerrors then begin
         (* skip blanks and lines until we get to some stuff *)
         skipblanks(fin);
         while eoln(fin) do begin
            readln(fin);
            skipblanks(fin);
         end;

         if fin^='''' then begin
            get(fin); (* skip the ' *)
            if fromto then second := first
                      else second := 0.0;
            if printer <> ' ' (* silencer *)
            then writeln(output, 'error assumed zero');
         end
         else getnumber(fin,second,ok)
      end
      else second := 0.0 (* set to zero silently if not reading them *)
   end;

   if ok then begin
      (* If we got a second number and it is not to be from-to, make sure
         it is positive *)
      if (not fromto) and (second < 0.0) then begin
         second := abs(second);
         writeln(output,'(Note: the error was made to be positive number.)');
      end;

      (* make the order correct if it is from-to and showingerrors *)
      if fromto and showingerrors and (first > second) then begin
         hold := first;
         first := second;
         second := hold;
         writeln(output,'(Note: order of the pair was reversed.)');
      end;

      (* now we need to interpret what we read. *)
      if fromto and showingerrors then begin
         r.estimate := (first + second)/2.0;
         r.error    := r.estimate - first;
      end
      else begin
         r.estimate := first;
         r.error    := second;
      end;
   end
   else begin (* be sure to return zeros *)
      r.estimate := 0.0;
      r.error    := 0.0;
   end;
end;
(* end module readrange *)

(* begin module writerange *)
procedure writerange(var fout: text; r: range; width, decimal: integer;
                     exponential: boolean);
(* write the range to fout, using the given width and decimal places
for the numbers.  If exponential is true, then use exponential form *)
var zerodecimal: boolean; (* true if there are to be no decimal places *)
procedure makeplusminus;
begin
  if exponential
  then write(fout, r.estimate:width        , pmstring, r.error:width        )
  else if zerodecimal
       then write(fout, round(r.estimate):width, pmstring, round(r.error):width)
       else write(fout, r.estimate:width:decimal, pmstring, r.error:width:decimal)
{
  then write(fout, r.estimate:width        ,' +/- ', r.error:width        )
  else if zerodecimal
       then write(fout, round(r.estimate):width,' +/- ', round(r.error):width)
       else write(fout, r.estimate:width:decimal,' +/- ', r.error:width:decimal)
}
end;
procedure makefromto;
begin
  write(fout,'[');

  if exponential
  then write(fout,(r.estimate-r.error):width        ,' to ',
                  (r.estimate+r.error):width        )
  else if zerodecimal
       then write(fout,round(r.estimate-r.error):width,' to ',
                       round(r.estimate+r.error):width)
       else write(fout,(r.estimate-r.error):width:decimal,' to ',
                       (r.estimate+r.error):width:decimal);

  write(fout,']')
end;
begin (* writerange *)
   zerodecimal := (decimal <= 0);

   if showingerrors then begin
      if fromto then makefromto
                else makeplusminus
   end
   else begin
      if exponential
      then write(fout, r.estimate:width        )
      else if zerodecimal
           then write(fout, round(r.estimate):width)
           else write(fout, r.estimate:width:decimal)
   end
end; (* writerange *)
(* end module writerange *)

(* begin module addrange *)
procedure addrange(a, b: range; var aplusb: range);
(* add a to b and put result in aplusb *)
begin
   aplusb.estimate := a.estimate + b.estimate;
   aplusb.error := sqrt( sqr(a.error)
                        +sqr(b.error) )
end;
(* end module addrange *)

(* begin module minusrange *)
procedure minusrange(a, b: range; var aminusb: range);
(* subtract b from a and put result in aminusb *)
begin
   aminusb.estimate := a.estimate - b.estimate;
   aminusb.error := sqrt( sqr(a.error)
                         +sqr(b.error) )
end;
(* end module minusrange *)

(* begin module timesrange *)
procedure timesrange(a, b: range; var atimesb: range);
(* multiply a by b and put result in atimesb *)
begin
   atimesb.estimate := a.estimate * b.estimate;
   if (a.estimate = 0) or (b.estimate = 0)
   then begin
      (* 2003 July 15: only give the warning if the : mode is on *)
      if showingerrors then begin
         writeln(output,'WARNING: one of the estimates to be multiplied is zero');
         writeln(output,'so the error will be set to zero');
      end;
      atimesb.error := 0
   end
   else atimesb.error := abs(atimesb.estimate) *
                        sqrt( sqr(a.error/a.estimate)
                             +sqr(b.error/b.estimate) )
end;
(* end module timesrange *)

(* begin module dividerange *)
procedure dividerange(a, b: range; var aoverb: range);
(* divide a by b and put result in aoverb *)
begin
   aoverb.estimate := a.estimate / b.estimate;
   if (a.estimate = 0) or (b.estimate = 0)
   then begin
      (* 2003 July 15: only give the warning if the : mode is on *)
      if showingerrors then begin
         writeln(output,'WARNING: one of the estimates to be divided is zero');
         writeln(output,'so the error will be set to zero');
      end;
      aoverb.error := 0
   end
   else aoverb.error := abs(aoverb.estimate) *
                        sqrt( sqr(a.error/a.estimate)
                             +sqr(b.error/b.estimate) )
end;
(* end module dividerange *)

(* begin module logrange *)
procedure logrange(a: range; base: real; var loga: range);
(* take log base of a as loga.
| 
| Given x + \delta x
| and   q(x) = ln x
| we want to find q + \delta q.
| 
| Taylor1982 page 65 gives:
| \delta q = abs(dq/dx) \delta x
| 
| For this case:
| 
| dq/dx = 1/x
| 
| so
| 
| \deltaq = \delta x/x.
| 
| For q(x) = ln(x)/ln(2) - ie log base 2
| dq/dx = 1/(x ln 2)
| 
| so \deltaq = \delta x / (x ln 2)
| 
| This was not done correctly before 2000 April 12,
| version 2.59 of the calc program.
*)
begin
   loga.estimate := ln(a.estimate)/ln(base);
(* original computation: incorrect!!
   loga.error := a.error/ a.estimate ;
*)
   loga.error := a.error/ (a.estimate * ln(base)) ;
end;
(* end module logrange *)

(* begin module exprange *)
procedure exprange(a: range; b: range; var atotheb: range);
(* find a to the b, Tayler page 72.  Let & mean the partial
derivative.  x=a and y=b.  For q = x^y,
&q/&x = y*x^(y-1)
      = b*a^(b-1)
      = b*exp((b-1)*ln(a))
Multiply this by the error in x = a to get the first part.

 ln q = y ln(x) so q = exp(y * ln(x)) so
&q/&y = exp(y * ln x) * ln(x)
      = exp(b * ln a) * ln(a)

Multiply this by the error in y = b to get the second part.
 *)
begin
   if a.estimate <= 0.0
   then writeln(output,'exponentiation refused: non-positive base')
   else begin
{
writeln(output,'  exprange, a.estimate = ',a.estimate);
writeln(output,'  exprange, a.error    = ',a.error   );
writeln(output,'  exprange, b.estimate = ',b.estimate);
writeln(output,'  exprange, b.error    = ',b.error   );
}
      atotheb.estimate := exp(b.estimate*ln(a.estimate));
      atotheb.error :=
         (b.estimate*exp((b.estimate-1)*ln(a.estimate)))     * a.error
          +
         (exp(b.estimate * ln(a.estimate)) * ln(a.estimate)) * b.error
{
;
writeln(output,'  exprange, atotheb.estimate = ',atotheb.estimate);
writeln(output,'  exprange, atotheb.error    = ',atotheb.error   );
}
   end
end;
(* end module exprange *)

(**************)

(* begin module initialize *)
procedure initialize;
(* initialize all global variables *)
begin
   base := exp(1.0); (* start with base e *)
   notation := false;  (* start with regular notation rather than scientific *)
   rwidth := defrwidth;     (* width of numbers *)
   rdecimal := defrdecimal; (* decimal places of numbers *)
   printer := 's'; (* print state *)
   showingerrors := false; (* showing of error ranges *)
   latex := true; (* latex display mode *)
   pmstring := pmstring1;
   fromto := false; (* use +/- display at first *)
   
   thestack := nil; (* no numbers on the stack *)
   spare := nil; (* no spare numbers *)
   carriagereturn := true; (* carriage return at ends of lines *)
end;
(* end module initialize *)

(* begin module push *)
procedure push(a: range);
(* push the number a onto the stack *)
var hold: stackptr; (* for moving a spare stack element *)
begin
   (* make a new spare if there isn't one there *)
   if spare = nil then begin
      new(spare);
      (* initialize it properly: *)
      spare^.n.estimate := 0.0;
      spare^.n.error := 0.0;
      spare^.next := nil
   end;

   (* pick up an element from the spare stack *)
   hold := spare;
   if spare^.next <> nil
      then spare := spare^.next
      else spare := nil;

   (* move the element onto the stack *)
   hold^.next := thestack;
   thestack := hold;

   (* put the number into the top of the stack *)
   thestack^.n := a
end;
(* end module push *)

(* begin module pop *)
procedure pop;
(* pop a number off the stack onto the spare stack *)
var hold: stackptr; (* for moving a spare stack element *)
begin
   if thestack <> nil then begin
      (* pick up the top element *)
      hold := thestack;
      thestack := thestack^.next; (* note this handles nil *)

      (* put the top element onto the spare stack *)
      hold^.next := spare;
      spare := hold
   end
   else writeln(output,'you cannot remove numbers from an empty stack');
end;
(* end module pop *)

(* begin module rolldown *)
procedure rolldown;
(* take top number on stack and put on the bottom *)
var hold: stackptr; (* for moving a spare stack element *)
    here: stackptr; (* where we are in the stack *)
begin
   (* no use rolling if there is nothing to roll! *)
   if thestack <> nil then begin
      (* no use rolling if there is nothing else! *)
      if thestack^.next <> nil then begin
         (* pick up the top element *)
         hold := thestack;
         thestack := thestack^.next;
         hold^.next := nil;

         (* find the end of the stack *)
         here := thestack;
         while here^.next <> nil do here := here^.next;

         here^.next := hold
      end
   end
end;
(* end module rolldown *)

(* begin module getone *)
function getone(var first: range): boolean;
(* check that one number is on the stack
give message if not and return false, return the number otherwise *)
begin
   if thestack=nil then begin
(* formerly showed this message:
      writeln(output,'stack is empty');
*)
      getone := false
   end
   else begin
      first := thestack^.n;
      getone := true
   end
end;
(* end module getone *)

(* begin module gettwo *)
function gettwo(var first, second: range): boolean;
(* check that two numbers are on the stack
give message if not and return false, return the numbers otherwise *)
begin
   if thestack=nil then begin
      writeln(output,'stack is empty');
      gettwo := false
   end
   else if thestack^.next = nil then begin
      writeln(output,'there is only one number on the stack');
      gettwo := false
   end
   else begin
      first := thestack^.n;
      second := thestack^.next^.n;
      gettwo := true
   end
end;
(* end module gettwo *)

(********)

(* begin module toggleprint *)
procedure toggleprint;
(* toggle the printer state *)
begin
   case printer of
      ' ': printer := 'p';
      'p': printer := 's';
      's': printer := ' ';
   end;
   case printer of
      ' ': ; (* writeln(output,'silent calculations'); *)
(*
formerlly:
      ' ': writeln(output,'silent - nothing will be shown');
      ' ': writeln(output,'silent - no stack will be shown');
*)
      'p': writeln(output,'printing top of stack all the time');
      's': writeln(output,'printing whole stack all the time');
   end
end;
(* end module toggleprint *)

(* begin module clear *)
procedure clear;
(* clear the stack *)
begin
   while thestack <> nil do pop
end;
(* end module clear *)

(* begin module decimal *)
procedure decimal;
(* set the number of decimal places *)
var n: integer; (* a tentative number *)
    top: range; (* top of the stack *)
begin
   if getone(top) then begin
      n := round(top.estimate);
      if n < 0 then begin
         writeln(output,'negative number of decimal places is not allowed');
      end
      else begin
         rdecimal := n;
         pop;
         if printer <> ' ' (* silencer *)
         then writeln(output,rdecimal:1,' decimal places');
      end
   end
end;
(* end module decimal *)

(* begin module duplicate *)
procedure duplicate;
var top: range; (* top of the stack *)
(* make two copies of the first number on the stack *)
begin
   if getone(top) then push(top)
end;
(* end module duplicate *)

(* begin module width *)
procedure width;
(* set the number of width places *)
var n: integer; (* a tentative number *)
    top: range; (* top of the stack *)
begin
   if getone(top) then begin
      n := round(top.estimate);
      pop; (* pop even if there is an error *)
      if n < 1 then begin
         if printer <> ' ' (* silencer *)
         then writeln(output,'non-positive width is not allowed');
      end
      else begin
         rwidth := n;
         if printer <> ' ' (* silencer *)
         then writeln(output,rwidth:1,' width characters');
      end
   end
end;
(* end module width *)

(* begin module dobase *)
procedure dobase;
(* set the number of base places *)
var n: real; (* a tentative number *)
    top: range; (* top of the stack *)
begin
   if getone(top) then begin
      n := top.estimate;
      if n <= 1.0 then begin
         writeln(output,'bases less than or equal to 1 are not allowed');
      end
      else begin
         base := n;
         pop;
         if printer <> ' ' then begin (* silencer *)
            write(output, ' base for logarithm and exponentiation is ');
            if rdecimal > 0 then writeln(output,base:rwidth:rdecimal)
                            else writeln(output,round(base):rwidth);
         end
      end;
   end
end;
(* end module dobase *)

(* begin module donotation *)
procedure donotation;
(* set the number of notation places *)
begin
   notation := not notation;
   if printer <> ' ' (* silencer *)
      then if notation
           then writeln(output,'scientific notation')
           else writeln(output,'normal numeric notation')
end;
(* end module donotation *)

(* begin module absolute *)
procedure absolute;
(* replace the top number with its absolute value *)
var a: range; (* number to take the absolute value of *)
begin
   if getone(a) then begin
      pop;
      a.estimate := abs(a.estimate);
      push(a)
   end
end;
(* end module absolute *)

(* begin module truncate *)
procedure truncate;
(* replace the top number with its truncated value *)
var a: range; (* number to truncate *)
begin
   if getone(a) then begin
      pop;
      a.estimate := trunc(a.estimate);
      push(a)
   end
end;
(* end module truncate *)

(* begin module add *)
procedure add;
(* add the two top numbers on the stack together,
and replace the two of them with the answer. *)
var a, b: range; (* numbers to add *)
    aplusb: range; (* the numbers added *)
begin
   if gettwo(a, b) then begin
      addrange(a,b,aplusb);
      pop; pop;
      push(aplusb)
   end
end;
(* end module add *)

(* begin module subtract *)
procedure subtract;
(* subtract the top number on the stack from the second number
and replace the two of them with the answer. *)
var a, b: range; (* numbers to subtract *)
    bminusa: range; (* the numbers subtracted *)
begin
   if gettwo(a, b) then begin
      minusrange(b,a,bminusa);
      pop; pop;
      push(bminusa)
   end
end;
(* end module subtract *)

(* begin module squareroot *)
procedure squareroot;
(* squareroot the top number on the stack from the second number
and replace the two of them with the answer. *)
var a, roota: range; (* number to squareroot and its root *)
    half: range; (* 0.5 *)
begin
   if getone(a) then begin
      half.estimate := 0.5;
      half.error    := 0.0;
      if a.estimate < 0.0 then begin
         writeln(output,'cannot take square root of negative real number');
      end
      else begin
         if a.estimate > 0.0
         then exprange(a,half,roota)
         else begin
            roota.estimate := 0.0;
            roota.error := 0.0;
         end;

         pop;
         push(roota)
      end
   end
end;
(* end module squareroot *)

(* begin module multiply *)
procedure multiply;
(* multiply the two top numbers on the stack together,
and replace the two of them with the answer. *)
var a, b: range; (* numbers to multiply *)
    atimesb: range; (* the numbers multiplyed *)
begin
   if gettwo(a, b) then begin
      timesrange(a,b,atimesb);
      pop; pop;
      push(atimesb)
   end
end;
(* end module multiply *)

(* begin module divide *)
procedure divide;
(* divide the top number on the stack from the second number
and replace the two of them with the answer. *)
var a, b: range; (* numbers to divide *)
    bovera: range; (* the numbers divided *)
begin
   if gettwo(a, b) then begin
      dividerange(b,a,bovera);
      pop; pop;
      push(bovera)
   end
end;
(* end module divide *)

(* begin module dolog *)
procedure dolog;
(* replace the top of the stack with its log *)
var a: range; (* the answer *)
    top: range; (* top of the stack *)
begin
   if getone(top) then begin
      if top.estimate > 0.0 then begin
         logrange(top,base,a);
         setrange(a,thestack^.n);
      end
      else begin
         writeln(output,'top number on the stack is not positive,',
                        ' cannot take logarithm');
      end
   end
end;
(* end module dolog *)

(* begin module dolatex *)
procedure dolatex;
(* toggle LaTeX mode on or off *)
begin
   latex := not latex;
   if latex
   then pmstring := pmstring1
   else pmstring := pmstring2;
{writeln(output,'the pmstring is "',pmstring,'"');}
end;
(* end module dolatex *)

(* begin module doexp *)
procedure doexp;
(* replace the top of the stack with its exponent.
Method: page 72 of Taylor1982.
For function q(Xbest +/- deltaX),
deltaq ~  (dq/dx) deltaX.
Here q = exp(x), so dq/dx = exp(x), so
deltaq ~  exp(x) deltaX.
*)
var a: range; (* the answer *)
    top: range; (* top of the stack *)
begin
   if getone(top) then begin
      pop;
      a.estimate := exp(top.estimate);
      a.error := a.estimate * top.error;
      push(a)
   end
end;
(* end module doexp *)

(* begin module douncertainty *)
procedure douncertainty;
(* replace the top of the stack with its uncertainty, - p log2(p).
In the special case of p = 0, return 0 because that is the limit. *)
var a: range; (* the answer *)
    logp: range; (* partial answer *)
    top: range; (* top of the stack *)
begin
   if getone(top) then begin
      if top.estimate < 0.0 then begin
         writeln(output,'u command: cannot take log because p is',
                        ' less than zero');
      end
      else begin
         if top.estimate > 1.0
         then writeln(output,'u command: WARNING:',
                             ' p was greater than or 1.0');
         pop;
         if top.estimate > 0.0 then begin
            logrange(top,2.0,logp);
            timesrange(top,logp,a);
            a.estimate := abs(a.estimate); (* make it positive *)
         end
         else begin
            a.estimate := 0.0;
            a.error := 0.0;  (* a guess! *)
         end;
         push(a)
      end;
   end
end;
(* end module douncertainty *)

(* begin module simpson *)
procedure simpson(upper : real; var answer,tol : real);
(* Perform a numerical integration of the normal distribution using
Simpson's rule.  The variable "lower" is the lower bound of the
integration.  The variable "upper" is the the upper bound of the
integration.  "answer" is the value of the integration, and "tol"
is the tolerance of the integration procedure, and thus the error
of the calculation.  Written by Mark Shaner, 1992.  Taken from
Miller, Alan R. PASCAL PROGRAMS FOR SCIENTISTS AND ENGINEERS. 
SYBEX Inc., 1981. pgs 260-291

2007 Mar 21:  Tom Schneider:  Using a tolerance of 1/maxint caused an
infinite loop on a 64 bit machine.  Presumably it was attempting to
compute the value to 20 or so decimal places.  To avoid this, the
tolerance is set to a more reasonable value.

*)

var
  i  (* a counter *): integer;

  x,x1,x2, (* the independent variables for calculating the value of
              functions *)
  pi, (* the value of pi *)
  val, (* the value of 1/sqrt(2*pi), it is used in calculating the value of the
          gaussian for any x value, and is defined as a variable in order to
          speed calculations. *)
  deltax, (* the distance between every x value and the subsequent one *)
  evensum, (* the sum of the area under each of the even numbered parabolas *)
  oddsum, (* the sum of the area under each of the odd numbered parabolas *)
  endsum, (* the sum of the area under the first and last parabolas *)
  endcor, (*  the value of the end correction, it is determined using dgauss *)
  sum1 : real; (* a place to store the previous sum so that it can be compared
                with the subsequent sum to determine if the tolerance level has
                been reached *)
  pieces : integer; (* the number of parabolas under the curve *)
  lower, (* the lower bound of the integration *)
  sum :real; (* the value of the area under the curve *)

begin
  pi := 4.0*arctan(1.0);
  val := 1/sqrt (2*pi);
  (* activate both files to be used *)
  lower := 0;
  upper := abs(upper);
{
  tol := 1/maxint;
}
  tol := 1.0E-6;
  pieces := 2;
  deltax := (upper-lower)/pieces;
  x := lower + deltax;
  oddsum := val*exp(-0.5*x*x);
  evensum := 0.0;
  x1 := lower;
  x2 := upper;
  endsum := val*exp(-0.5*x1*x1) + val*exp(-0.5*x2*x2);
  endcor := -(x1*val)*exp(-0.5*x1*x1) - (-(x2*val)*exp(-0.5*x2*x2));
  sum := (endsum + 4.0*oddsum)*deltax/3.0;

  repeat
    pieces := pieces*2;
    sum1 := sum;
    deltax := (upper - lower)/pieces;
    evensum := evensum + oddsum;
    oddsum := 0.0;
    for i := 1 to (pieces div 2) do
    begin
      x := lower + deltax*(2.0*i-1.0);
      oddsum := oddsum + val*exp(-0.5*x*x);
    end;
    sum := (7.0*endsum + 14.0*evensum + 16.0*oddsum + endcor*deltax)
          *deltax/15.0;
  until abs(sum-sum1) <= abs(tol*sum);
  answer := (0.5 - sum);
  if answer < 0.0 then answer := 0.0; (* safety for roundoff errors *)
(*
  writeln (upper:7:5,'      ',answer:7:5,'+/-',tol)
*)
end;
(* end module simpson version = 'matmod 2.07 2007mar21 tds/gds'; *)

(* begin module dogaussian *)
procedure dogaussian;
(* replace the top of the stack, a Normal Gaussian Z
value, with its probability *)
var a: range; (* the answer *)
    Z: range; (* top of the stack *)
begin
   if getone(Z) then begin
      pop;
      if abs(Z.estimate) > zbound then begin
         a.estimate := 0;
         a.error := 1/maxint;
         writeln(output,'warning: z > ',zbound:1,'; ',
                        'returning zero for Gaussian');
      end
      else simpson(Z.estimate, a.estimate, a.error);
      push(a)
   end
end;
(* end module dogaussian *)

(* begin module doz *)
procedure doz;
(* find the distance of the first number on the stack from the mean
and standard deviation of the second number on the stack.  The second
number is not disturbed. *)
var first, second: range; (* two numbers on the top of the stack *)
begin
   if gettwo(first,second) then begin
      pop; pop;
      if second.error <= 0 then begin
         writeln(output,'error of second number on stack must be positive',
                        ' to do Z function');
      end
      else begin
         first.estimate := (first.estimate - second.estimate) / second.error;
         first.error := 0.0;
      end;
      push(second);
      push(first)
   end
end;
(* end module doz *)

(* begin module doraise *)
procedure doraise;
(* raise the second stack to the first stack number *)
var a: range; (* the answer *)
    first, second: range; (* two numbers on the top of the stack *)
begin
   if gettwo(first,second) then begin
      pop; pop;
      exprange(second,first,a); (* second raised to the first *)
      push(a)
   end
end;
(* end module doraise *)

(* begin module swap *)
procedure swap;
(* swap the top two numbers on the stack *)
var first, second: range; (* numbers to swap *)
begin
   if gettwo(first, second) then begin
      pop; pop;
      push(first);
      push(second)
   end
end;
(* end module swap *)

(* begin module join *)
procedure join;
(* join the top two numbers on the stack,
the second number becomes the error of the first.
In from-to mode, the estimates of the top two numbers become the range. *)
var
   first, second: range; (* numbers to join *)
   average: real; (* average of first and second numbers *)
   sigma: real; (* half of the range *)
begin
   if gettwo(first, second) then begin
      pop; pop;

      if fromto then begin
         average := (first.estimate + second.estimate) / 2;
         sigma := abs(average - first.estimate);
         first.estimate := average;
         first.error := sigma;
         push(first);
	 if printer <> ' '
         then writeln(output,'joined as from-to');
      end
      else begin
         second.error := abs(first.estimate); (* force to be positive *)

         if first.estimate < 0
         then writeln(output,
             '(Note: the error was made to be positive number.)');

         push(second);
	 if printer <> ' ' then writeln(output,'joined as estimate-error');
      end
   end
end;
(* end module join *)

(* begin module helpalways *)
procedure helpalways(var f: text);
(* some help we always give *)
begin
writeln(f);
writeln(f,'For more help type one of these commands:');
writeln(f,'1h gives general help for calc');
writeln(f,'2h gives help for display of numbers');
writeln(f,'3h gives help for functions');
writeln(f,'4h gives help for commands');
writeln(f);
writeln(f,'NOTE: use the form 4''h (for example) when in ":" mode.');
writeln(f);
end;
(* end module helpalways *)

(* begin module helpcommand *)
procedure helpcommand(var f: text);
(* Give help for commands *)
begin
writeln(f,'Help for Commands');
writeln(f,'-----------------');

writeln(f,'c clear the stack');
writeln(f,'b top number of stack is used as the base for taking logarithms');
if rdecimal > 0 then writeln(f,'  (currently ', base:rwidth:rdecimal,')')
                else writeln(f,'  (currently ', round(base):rwidth,')');
if base = exp(1.0)
then writeln(f,'  (this is e, which gives the natural logarithm)')
else writeln(f,'  (Note: to get back to natural logs use the commands "1eb")');
writeln(f,'d top number of stack is used as',
          ' the number of decimal places to print');
writeln(f,'  (currently ', rdecimal:1,')');
writeln(f,'j join: first number on stack becomes error of the second number');
writeln(f,'  In from-to mode, the estimates of the top two numbers');
writeln(f,'  become the range.');
writeln(f,'L toggle between LaTeX and regular number notation: \pm vs +/-');
writeln(f,'n toggle between scientific and regular number notation');
write  (f,'  (currently ');
             if notation then write(f, 'scientific')
                         else write(f, 'regular');
writeln(f, ' notation)');
writeln(f,'p print top number on the stack');
writeln(f,'s print the entire stack');

writeln(f,'t toggles the way the stack is shown between three states:');
write  (f,'  silent, print top number only, print whole stack.');
write  (f,'  (currently ');
   case printer of
      ' ': writeln(f,'SILENT)');
      'p': writeln(f,'PRINT TOP)');
      's': writeln(f,'PRINT STACK)');
   end;

writeln(f,'q quit the program');
writeln(f,'w top number of stack is used as',
          ' the width of numbers to print');
writeln(f,'  (currently ', rwidth:1,')');
writeln(f,'x throw away the top number on the stack');
writeln(f,'y rolldown: top number on stack goes to bottom of stack');
writeln(f,'~ swap the top two numbers on the stack');
writeln(f,', duplicate the top number on the stack');
writeln(f,'. do not show error ranges and do not read new ones,');
writeln(f,'  (but keep calculating them if they had already been read)');
writeln(f,': do show and do error ranges');
writeln(f,'# make a comment: calc will ignore the rest of the line');
writeln(f,'m make a message: like #, but appears in silent mode.');
writeln(f,'  A "\" at the end of the line suppresses carriage returns');
writeln(f,'\ as a command toggles carriage returns on and off in p and s');

end;
(* end module helpcommand *)

(* begin module helpfunction *)
procedure helpfunction(var f: text);
(* Give help for functions *)
begin
writeln(f,'Help for Functions');
writeln(f,'------------------');

writeln(f,'a absolute value of top number on stack, put on top of stack');
writeln(f,'T truncate top number on stack, put on top of stack');
writeln(f,'+ ADD top two numbers on the stack, put on top of stack');
writeln(f,'_ (UNDERSCORE) SUBTRACT top two numbers on the stack,',
             ' put on top of stack');
writeln(f,'* MULTIPLY top two numbers on the stack, put on top of stack');
writeln(f,'/ DIVIDE second number by first number on the stack',
          ' put on top of stack');
writeln(f,'^ raise second number to first number on the stack, put on top');
writeln(f,'e raise e to the first number on the stack, put on top');
writeln(f,'g convert the Gaussian Z value to a probability, put on top');
writeln(f,'l replace top number on stack by its LOG to the base ');
if rdecimal > 0 then writeln(f,'  (currently ', base:rwidth:rdecimal,')')
                else writeln(f,'  (currently ', round(base):rwidth,')');
if base = exp(1.0)
then writeln(f,'  (this is e, which gives the natural logarithm)');
writeln(f,'r replace top number on stack by its SQUARE ROOT');
writeln(f,'u uncertainty: if p is the top of the stack, replace it with',
          ' -p*log2(p)');
writeln(f,'z replace top number on stack by its Z distance from second number');
writeln(f,'% poisson: error of top number on stack is square root of estimate');
writeln(f,'@ toggle range display between "8 +/- 3" and "[5 to 11]".');
writeln(f,'  Not only will numbers show up differently, but they will');
writeln(f,'  be READ IN DIFFERENTLY when the : is in operation.');
writeln(f,'  eg: after starting calc, "0d1w :@ 5 11" gives "[5 to 11]"');
writeln(f,'! extra special cowculation');

end;
(* end module helpfunction *)

(* begin module helpgeneral *)
procedure helpgeneral(var f: text);
(* give general help *)
begin
writeln(f,'General Help for Calc version ',version:4:2);
writeln(f,'----------------------------------');

writeln(f);
writeln(f,'The calc program is a reverse polish calculator that');
writeln(f,'propagates errors.  The program is based on the idea of the');
writeln(f,'dc program under UNIX.  That program takes input as reverse');
writeln(f,'polish and calculates values.  This program does the same,');
writeln(f,'but values have estimates so one may calculate and');
writeln(f,'propagate errors.  Tokens (commands and numbers) are');
writeln(f,'usually separated by spaces or carriage returns.  Tokens');
writeln(f,'that begin with a digit or a dash (-) are numbers.');
writeln(f);

writeln(f,'Normally the program does not show you the errors, and you');
writeln(f,'run it like a regular calculator.  For example, to average');
writeln(f,'three numbers, one enters them, adds them together, and then');
writeln(f,'divides by three:  "3 4 5 + + 3 /".  Try this with carriage');
writeln(f,'returns between the commands to see how the stack changes.');
writeln(f);

writeln(f,'The methods of calculation are given in:');
writeln(f,'An Introduction to Error Analysis,  John R. Taylor');
writeln(f,'University Science Books, Mill Valley, CA. 1982.');
writeln(f);
writeln(f,'Tom Schneider');
writeln(f,'toms@alum.mit.edu');

end;
(* end module helpgeneral *)

(* begin module helpdisplay *)
procedure helpdisplay(var f: text);
(* give display help *)
begin
writeln(f,'Help for displaying numbers in Calc');
writeln(f,'-----------------------------------');

writeln(f,'To enter errors and display them, type ":".  The prompt of');
writeln(f,'the program will change from "." to ":" to indicate this.');
writeln(f,'Now every pair of numbers you type represents a number and');
writeln(f,'its error range.  Thus "5 2" means "5 +/- 2" and it will be');
writeln(f,'displayed that way.  A single quote, as in "5''" means to');
writeln(f,'set the error to zero.');
writeln(f);
writeln(f,'You can also enter numbers as a range by flipping the toggle "@".');
writeln(f,'The prompt will become "@:" to indicate this.');
writeln(f,'Now every pair of numbers you type represents a range.');
writeln(f,'Thus "1 5" means the range [1 to 5] or "3 +/- 2".');
writeln(f,'The numbers shown on the stack will reflect this.');
writeln(f);
writeln(f,'You can see things in scientific notation with the toggle "n".');
writeln(f,'This has no effect on how you enter numbers.');
writeln(f);

writeln(f,'Calculation is not affected by the display.  So you can enter');
writeln(f,'numbers as errors or ranges, switch to "period" mode, do your');
writeln(f,'calculations without the clutter, then switch back to "colon"');
writeln(f,'to see them.');

end;
(* end module helpdisplay *)

(* begin module help *)
procedure help(var f: text);
(* give general help *)
var top: range; (* top number from the stack *)
    t: integer; (* the round of the estimation portion of top *)
begin
   writeln(f);
   (* although I hesitate to refer directly to the stack here, it is
      best not to bother the user with 'stack is empty' when giving help. *)
   if thestack=nil
   then helpgeneral(f)
   else begin
      if getone(top) then begin
         t := round(top.estimate);
         pop;
         (* change the upper bound in the following if statement
            when there are more help 'screens' *)
         if (t < 1) or (t > 4)
         then helpgeneral(f)
         else case t of
            1: helpgeneral(f);
            2: helpdisplay(f);
            3: helpfunction(f);
            4: helpcommand(f);
         end;
      end
      else writeln('program error in procedure help')
   end;

   helpalways(f);
end;
(* end module help *)

(* begin module print *)
procedure print;
(* print the top number on the stack *)
var top: range;
begin
   if getone(top) then begin
      writerange(output, top, rwidth, rdecimal, notation);
      if carriagereturn then writeln(output)
(*
I don't know why one would want a space after the number!
                        else write(output,' ');
*)
   end
end;
(* end module print *)

(* begin module printstack *)
procedure printstack;
(* print the entire stack *)
var top: range;
    s: stackptr;
begin
   if getone(top) then begin
      writeln(output,'stack:'); (* separate from previous stuff *)
      s := thestack;
      while s <> nil do begin
         writerange(output, s^.n, rwidth, rdecimal, notation);
         if carriagereturn then writeln(output)
                           else write(output,' ');
         s := s^.next
      end
   end
end;
(* end module printstack *)

(* begin module prompt *)
procedure prompt(var fout: text);
(* show the prompt character *)
begin
   if printer <> ' ' then begin (* otherwise be totally silent!! *)
      if notation then write(fout,'n');
      if showingerrors
      then begin
         if fromto then write(fout,'@');
         write(fout,': ')
      end
      else write(fout,'. ')
   end
end;
(* end module prompt *)

(* begin module poisson *)
procedure poisson;
(* set the error of the top number of the stack to be
the square root of the estimate (poisson distribution). *)
begin
                 (* x                      *)
   duplicate;    (* x              x       *)
   squareroot;   (* sqrt(x)        x       *)
   join          (* x +/- sqrt(x)          *)
end;
(* end module poisson *)

(* begin module dofromto *)
procedure dofromto;
(* toggle the global fromto variable *)
begin
   fromto := not fromto
end;
(* end module dofromto *)

(* begin module docomment *)
procedure docomment(var fin, fout: text);
(* do a comment by copying the line to the output *)
begin
  if printer <> ' ' then write(fout,'# '); (* replace the read symbol *)
  (* copyaline(fin,fout); *)
      while not eoln(fin) do begin
         fout^ := fin^;
         if printer <> ' ' then put(fout);
         get(fin)
      end;
      (* readln(fin); *) (* let this be done elsewhere *)
      if printer <> ' ' then writeln(fout);
end;
(* end module docomment *)

(* begin module docow *)
procedure docow;
(* have a cow!
from
Part 2 of the 2-part COMPLETE GUIDE TO COWS!
Eric W. Tilenius // etileniu@Oracle.COM *)
begin
   writeln(output);
   writeln(output,'                         (__)');
   writeln(output,'                         (oo)');
   writeln(output,'          / ------------- \/');
   writeln(output,'         / (   |7 8 9 +|  )');
   writeln(output,'        *  (   |4 5 6 -|  )');
   writeln(output,'            (  |1 2 3 X|  )');
   writeln(output,'             ( |C 0 = /| )');
   writeln(output,'              ^---------^');
   writeln(output,'              ||       ||');
   writeln(output,'              ||       ||');
   writeln(output,'              ^^       ^^');
   writeln(output,'              "COWCULATOR"');
   writeln(output);

writeln(output,
'Dudley''s Criteria');
writeln(output);
writeln(output,
'The proof may seem to be unsatisfying: each step is correct, and hence');
writeln(output,
'the conclusion is true, but it is not clear why the steps are there');
writeln(output,
'and where they came from. That is because there are at least eight');
writeln(output,
'levels of mathematical understanding, and it is hard for someone on a');
writeln(output,
'lower level to appreciate what goes on at a higher level. The levels');
writeln(output,
'are, I think:');
writeln(output);
writeln(output,
'1. Being able to do arithmetic.');
writeln(output);
writeln(output,
'2. Being able to substitute numbers in formulas.');
writeln(output);
writeln(output,
'3. Given formulas, being able to get other formulas.');
writeln(output);
writeln(output,
'4. Being able to understand the hypotheses and conclusions of');
writeln(output,
'theorems.');
writeln(output);
writeln(output,
'5. Being able to understand the proofs of theorems, step by step.');
writeln(output);
writeln(output,
'6. Being able to _really_ understand the proofs of theorems: that is,');
writeln(output,
'seeing why the proof is as it is, and comprehending the inwardness of');
writeln(output,
'the theorem and its relation to other theorems.');
writeln(output);
writeln(output,
'7. Being able to generalize and extend theorems.');
writeln(output);
writeln(output,
'8. Being able to see new relationships and discover and prove entirely');
writeln(output,
'new theorems.');
writeln(output);
writeln(output,
'Those of us stuck on level 5 can no more understand the workings of a');
writeln(output,
'level 8 mind than a cow could understand calculus.');
writeln(output);
writeln(output,
'Elementary Number Theory, 2nd ed, page 103-104, by Underwood Dudley.');
writeln(output,
'W. H. Freeman & co. 1969, second edition 1978 DUDLEY@DEPAUW.EDU');

end;
(* end module docow *)

(* begin module togglereturn *)
procedure togglereturn;
(* toggle switch for carriage returns *)
begin
   carriagereturn := not carriagereturn
end;
(* end module togglereturn *)

(* begin module domessage *)
procedure domessage(var fin, fout: text);
(* do a message by copying the line to the output.  This is like docomment,
but it always appears, even in silent mode, and no # or m symbol is duplicated
to the output (gives the user complete control).  Allow more control
by suppressing carriage return at ends of lines if \ (backslash) is the
last character of the line. *)
var
  lastcharacter: char; (* the previous character in the file *)
  suppress: boolean; (* the previously seen character was a backslash
     and was at the end of the line.  Suppress the backslash
     and writing the linefeed carriage return *)
begin
   suppress := false;
   while not eoln(fin) do begin
      read(fin,lastcharacter);
      suppress := (lastcharacter = '\') and eoln(fin);
      if not suppress then write(fout,lastcharacter);
   end;
   (* readln(fin); *) (* let this be done elsewhere *)
   if not suppress then writeln(fout);
end;
(* end module domessage *)

(* begin module calc.themain *)
procedure themain (var fin, fout: text);
(* the main procedure of the program *)
var
   a: range; (* a number *)
   achar: char; (* a character read in *)
   done: boolean; (* a loop is done when true *)
   ok: boolean; (* the number read in was ok *)
begin
   writeln(fout,'calc ',version:4:2);
   initialize;
   prompt(fout);

   (* copy the first lines that begin with * to fout *)
   (*while fin^='*' do copyaline(fin,fout);*)

   done := false;
   repeat
      skipblanks(fin);

      if eoln(fin)
      then begin
         readln(fin);
         prompt(fout);
      end
      else begin
         if fin^ in ['-','0','1','2','3','4','5','6','7','8','9'] then begin
            readrange(fin, a, ok, showingerrors);
            if ok then push(a)
         end

         else begin
            read(fin, achar);
                 if achar='a' then   absolute
            else if achar='c' then   clear
            else if achar='b' then dobase
            else if achar='d' then   decimal
            else if achar=',' then   duplicate
            else if achar='^' then doraise
            else if achar='e' then doexp
            else if achar='g' then dogaussian
            else if achar='h' then   help(fout)
            else if achar='?' then   help(fout)
            else if achar='j' then   join
            else if achar='l' then dolog
            else if achar='L' then dolatex
            else if achar='m' then domessage(fin,fout)
            else if achar='n' then donotation
            else if achar='p' then   print
            else if achar='s' then   printstack
            else if achar='t' then   toggleprint
            else if achar='u' then douncertainty
            else if achar='q' then   done := true
            else if achar='w' then   width
            else if achar='x' then   pop
            else if achar='y' then   rolldown
            else if achar='z' then doz
            else if achar='T' then   truncate
            else if achar='+' then   add
            else if achar='_' then   subtract
            else if achar='r' then   squareroot
            else if achar='*' then   multiply
            else if achar='/' then   divide
            else if achar='~' then   swap
            else if achar='.' then   showingerrors := false
            else if achar=':' then   showingerrors := true
            else if achar='%' then   poisson
            else if achar='@' then dofromto
            else if achar='#' then docomment(fin,fout)
            else if achar='!' then docow
            else if achar='\' then togglereturn
            else writeln(fout,'unrecognized command: ',achar)
         end;

         if not (achar in ['p','s','q']) then case printer of
            ' ': ;
            'p': print;
            's': printstack;
         end
      end
   until done
end;
(* end module calc.themain *)

begin
   themain(input,output);
end.
