program chacha(fin,fout,chachap,output);
(* chacha: character changing program
     by patrick roche
     module libraries: delman *)

label  1;   (* end of chacha *)

const
(* begin module version *)
version = 3.12; (* of chacha 1996 July 29
origin 1981 november 8 *)
(* end module version *)

(* begin module describe.chacha *)
(*
name
      chacha: changes characters in a file

synopsis
      chacha(fin: in, fout: out, chachap: in, output: out)

files
      fin:     any file in which one wants to translate one set of
               characters into another set.
      fout:    the file to which the translated copy of fin is written.
      chachap: the chacha parameter file which contains the translation sets.
               chachap must only contain 2 lines.
               the first line contains the characters used in fin, typed one
                  right after the next with no blanks at the beginning.
               the second line contains the characters that the
                  characters in the first line are to be translated
                  into, typed in the same way and in corresponding order.
               if you want to change a character to blanks,
                  or vice versa, then you must have the blank character
                  in between other characters in chachap.
      output:  where error messages will appear.

description
      chacha translates characters in a file to a new set of characters.
      also, more than one character can be translated in one run
      of the program.

examples
      to convert between double and single quotes, use:
         '"
         "'
      to convert blanks to periods, use:
         j j
         j.j
      in the chachap file.
      each character on the first line on chachap will be translated
      into the character directly beneath it on the second line in
      the output file.

documentation
      delman.assembly.intro  and  delman.assembly.chacha

see also
      worcha.p

author
      patrick r. roche

bugs
      none known

technical notes
      the maximum number of characters that can be translated is
      constant top.  caution: top is also the maximum line length.

*)
(* end module describe.chacha *)

      top = 100;      (* max length of line *)

type
      buffer = record  (* holds the characters from chachap *)
         place: array[1..top] of char;
         length: integer (* length of place *)
      end;

var
      fin,  (* the input file to be translated *)
      fout, (* the file to which translated fin will be written *)
      chachap: (* the parameter file containing the translation definition *)
       text;

      chafrom,    (* the first line of chachap, characters to "change from" *)
      chato,      (* the second line of chachap, characters to "change to" *)
      b:          (* line buffer, holds a line from the input file *)
       buffer;

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.51 85 apr 17 tds/gds' *)

procedure readintobuffer(var fin: text; var b: buffer);
      (* read a line into a buffer *)

      var i: integer;   (* index for b.place[i] and counter *)

      begin
         i:=0;          (* starting at begining of the line *)
         b.length :=0;  (* keeping track of the length of the buffer *)

         while ( i<=top ) and not (eoln(fin) or eof(fin) ) do begin
         (* keeping line length below max and going to end of line *)
            i:= i+1 ;               (* next character *)
            read(fin, b.place[i]);   (* read in next character to buffer *)
            b.length := i           (* keeping track of length of the buffer *)
         end;
         if not eof(fin) then readln(fin)    (* go to beginning of next line *)

      end;

procedure removeblanks(var b: buffer);
      (* remove blanks from the end of a buffer *)
      begin
         if b.length=0
         then begin
            writeln(output);
            writeln(output,' no blank lines allowed in chachap.');
            writeln(output);
            halt   (* stop program and notify user *)
         end;
         if b.place[b.length]=' '   (* if the last place is blank *)
         then begin   (* remove the blank and check next place etc. *)
            repeat
               b.length:= b.length-1
            until b.place[b.length]<>' '
         end;
      end;

procedure checkchachap;
      (* check chachap---the parameter file containing the transdef *)

      begin
         if (chafrom.length <=0) or (chato.length <=0)
            (* if chachap contains blank lines *)
         then begin
            writeln(output);
            writeln(output,' no blank lines allowed in chachap.');
            writeln(output);
            halt   (* stop program and notify user *)
         end;
         if chafrom.length <> chato.length    (* if lines do not match up *)
         then begin
            writeln(output);
            write  (output,' the 2 lines in chachap do not contain');
            writeln(output,' the same number of characters.');
            writeln(output);
            halt   (* stop program and notify user *)
         end
      end;


procedure translate(var fin, fout: text);

      var i: integer;   (* index of buffer.places *)
          inptchar,  (* inputcharacter *)
          tranchar:  (* translated character *)
           char;

      begin
      (* read the input char, look for input char in the transdef,
      if input char is not found copy input char to output, in input
      char is found write translated char to output. *)

         while not eof(fin) do begin
            while not eoln(fin) do begin
               read(fin,inptchar);
               i:=0;
               repeat
                  i:=i+1
               until (chafrom.place[i] = inptchar)  or  (chafrom.length=i);

               if chafrom.place[i] <> inptchar
               then write(fout,inptchar)
               else begin
                  tranchar:= chato.place[i];
                  write(fout,tranchar)
               end
            end;
            writeln(fout);   (* go to next output line *)
            readln(fin)   (* go to next input line *)
         end
      end;
begin  (* chacha *)
      writeln(output,' chacha ',version:4:2);
      reset(fin);
      reset(chachap);
      rewrite(fout);

      readintobuffer(chachap,chafrom);
      removeblanks(chafrom);
      readintobuffer(chachap,chato);
      removeblanks(chato);
      checkchachap;

      translate(fin,fout);
1: end.
