program datarange(datain, datarangep, dataout, output);
(* datarange: limit the range of a data table

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.09; (* of datarange.p 2018 Nov 22
2018 Nov 22, 1.09: find out why pc/fpc objects
2010 Oct 20, 1.08: invert logic
2010 Oct 20, 1.07: polish
2010 Oct 20, 1.07: new modules skipstringblanks, stringtoreal from
                   prgmod now
2010 Oct 20, 1.06: clean up
2010 Oct 20, 1.05: convert datum to real hopefully ... functional!
2010 Oct 20, 1.04: clean up
2010 Oct 20, 1.03: process data, it gets the column now!
2010 Oct 20, 1.02: switch to standard strings! - identical input to output!
2010 Oct 20, 1.01: obtain string routines!
2010 Oct 20, 1.00: origin
*)
updateversion = 1.08; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.datarange *)
(*
name
   datarange: limit the range of a data table

synopsis
   datarange(datain: in, datarangep: in, dataout: out, output: out)

files

   datain:  input data file.  '*' are comments, blank lines ignored.
      data are separated by spaces.

   dataout:  output data file with data lines removed that are
      in the range defined in datarangep.

   datarangep:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      column: column to use for making the cut

      mincut: minimum data value allowed

      maxcut: maximum data value allowed

      deleteinclude (character):
        'd' to delete the range
        'i' to include the range

   output: messages to the user

description

   Cut data lines out of a file based on values in a data column. 
   Data files in the delila system have columns separated by blanks. 
   This program reads the datain file and determines the value of the
   data (integer or real) in a particular column.  If the datum is
   outside the range defined in the parameter file, that data line is
   copied to the dataout file.

examples

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.datarange *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

var
   datain, (* file used by this program *)
   datarangep, (* file used by this program *)
   dataout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module skipstringblanks *)

procedure skipstringblanks(var line: string);
(* skip blanks in the line *)
begin
   with line do while (letters[current] = ' ') and (current < maxstring)
   do current := succ(current)
end; (* skipstringblanks *)

procedure skipstringnonblanks(var line: string);
(* skip nonblanks in the line *)
begin
   with line do while (letters[current] <> ' ') and (current < maxstring)
   do current := succ(current)
end; (* skipstringblanks *)

procedure skipstringcolumn(var line: string);
(* skipstring over a data column in the line *)
begin
   skipstringblanks(line); skipstringnonblanks(line);
end;

procedure stringcolumn(line: string; column: integer; var datum: string);
(* extract the column from the line and put into the datum string.
Ignore leading blanks in defining a column. *)
const
{
   debug = true;
}
   debug = false;
var
   c: integer; (* index to the datum characters *)
   columnnumber: integer; (* current column in the line *)
   done: boolean; (* did we find the datum column *)
   positionbeg: integer; (* position of the datum, begin *)
   positionend: integer; (* position of the datum, end *)
begin
   columnnumber := 0;
   done := false;
   if debug then begin writestring(output, line); writeln(output); end;
   while not done do begin
       if (columnnumber = column) or (line.current >= line.length) then begin
          done := true;
       end
       else begin
          skipstringblanks(line);
          positionbeg := line.current;
          if debug then write(output,'|':line.current,' ',line.current:1);
          if debug then write(output,'"',line.letters[positionbeg],'"');
          if debug then writeln(output);
          
          skipstringnonblanks(line);
          positionend := line.current - 1;
          if debug then write(output,'|':line.current,' ',line.current:1);
          if debug then write(output,'"',line.letters[positionend],'"');
          if debug then writeln(output);

          columnnumber := succ(columnnumber);
       end;
   end;
   clearstring(datum);
   if columnnumber = column then begin
      datum.length := positionend - positionbeg + 1;
      for c := 1 to datum.length do begin
          if debug then write(output,line.letters[positionbeg + c - 1]);
          datum.letters[c] := line.letters[positionbeg + c - 1];
      end;
   end;
   if debug then writeln(output);
{
   if debug then halt;
}
end;

(* end module skipstringblanks version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 representing a place value in the number *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module stringtoreal *)
procedure stringtoreal(datum: string; var rea: real; var gotten: boolean);
(* convert the string to a real number if possible
integer values are also accepted. *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
begin (* stringtoreal *)
      if datum.length = 0
      then begin
         gotten:=false;
         rea := maxint;
      end
      else begin
         datum.current := 1; (* directs figurestring to the string *)
         figurestring(datum,first,last,what,cha,int,rea);
         if not (what in ['r', 'i'])
         then begin
            gotten:=false
         end
         else begin
            gotten:=true;
            if what = 'i' then rea := int (* handle integers *)
         end;
      end;
end; (* stringtoreal *)
(* end module stringtoreal version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module datarange.themain *)
procedure themain(var datain, datarangep, dataout: text);
(* the main procedure of the program *)
const
   wid = 10; (* width of output reals *)
   dec = 5; (* decimals of output reals *)
var
   column: integer; (* column to use to cut data *)
   datum: string; (* a data value from the column *)
   delete: boolean; (* delete lines in the range *)
   deleteinclude: char; (* d or i for delete or include data *)
   gotten: boolean; (* was the line gotten? *)
   include: boolean; (* include lines in the range *)
   mincut, maxcut: real; (* minimum and maximum to cut from the data *)
   line: string; (* a line of data *)
   linenumber: integer; (* line number of the line string in the file *)
   parameterversion: real; (* parameter version number *)
   thedatum: real; (* the datum string as a number *)

begin
   writeln(output,'datarange ',version:4:2);
   reset(datarangep);
   readln(datarangep, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(datarangep, column);
   readln(datarangep, mincut);
   readln(datarangep, maxcut);
   readln(datarangep, deleteinclude);

   writeln(output, column:wid, ' column');
   writeln(output, mincut:wid:dec, ' mincut');
   writeln(output, maxcut:wid:dec, ' maxcut');
   writeln(output, deleteinclude:wid, ' deleteinclude');

   if deleteinclude = 'd' then begin
      delete  := true;
      include := false;
   end
   else begin
      if deleteinclude = 'i' then begin
         delete  := false;
         include := true;
      end
      else begin
         writeln(output,'deleteinclude must be one of ''di''');
         halt;
      end
   end;

   reset(datain);
   rewrite(dataout);

   while not eof(datain) do begin
      if eoln(datain) then begin
         readln(datain);
         writeln(dataout);
         linenumber := succ(linenumber);
      end
      else begin
         getstring(datain, line, gotten);
         linenumber := succ(linenumber);
         if gotten then begin
            (* detect the comment or blank lines *)
            if (line.length > 0) and (line.letters[1] <> '*') then begin
               stringcolumn(line, column, datum);
               if (datum.length > 0) then begin
                  (* convert datum from string to number thedatum *)
                  thedatum := 5;
                  stringtoreal(datum, thedatum, gotten);
                  if not gotten then begin
                     writeln(output,'ERROR: data in column ',column:1,
                             ' line ',linenumber:1,
                             ' is not an integer or real number:');
                     writestring(output,datum);
                     writeln(output);
                     halt;
                  end;

{
                  if (thedatum < mincut) or (thedatum > maxcut) then begin
}
                  if
                     (((thedatum <  mincut) or (thedatum >  maxcut)) and delete)
                     or
                     (((thedatum >= mincut) and (thedatum <= maxcut)) and include)
                  then begin
                     writestring(dataout,line);
                     writeln(dataout);
                  end
               end
            end
            else begin (* just copy the comment or blank lines *)
               writestring(dataout,line);
               writeln(dataout);
            end
         end;
      end;
   end;

end;
(* end module datarange.themain *)

begin
   themain(datain, datarangep, dataout);
1: end.
