program delinst(title, names,
 organism, chromosome, piece,
 coordinates, coordinates2,
 directions, delinstp, inst, output);
(* delinst: create delila instructions from lists

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.19; (* of delinst.p 2013 Aug 26
2013 Aug 26, 1.19: undo chr1 -> chr01
2013 Aug 23, 1.18: chr1 -> chr01 for sorting.
2013 Aug 23, 1.17: not m blocks writeln after chr/pie
2013 Aug 23, 1.16: correct number of instructions written
2013 Aug 19, 1.15: second coordinate set: implement
2013 Aug 19, 1.14: second coordinate set: parameter reading
2013 May 03, 1.13: identify bug
2013 May 02, 1.12: skip bad instructions
2013 May 02, 1.11: check that coordinate is a number
2013 Apr 09, 1.10: remove extra quote marks on org and chr commands!
2013 Feb 14, 1.09: remove blanks from ends of inst when multiline
2013 Feb 13, 1.08: clean up and add blank before new chr
2013 Feb 12, 1.07: compare org, chr and pie to previous; don't write if param
2013 Feb 12, 1.06: negative value for max instructions parameter means do all
2013 Feb 12, 1.05: max instructions parameter
2013 Feb 12, 1.04: add organism, chromosome, piece files
2013 Feb 08, 1.03: improve documentation
2013 Feb 08, 1.02: fix names
2013 Feb 07, 1.01: partially documented
2013 Feb 07, 1.00: origin
*)
updateversion = 1.14; (* defines lowest acceptable current parameter file *)
versionupperbound = 20.00; (* highest version that will be accepted *)
(* end module version *)

(* begin module describe.delinst *)
(*
name
   delinst: create delila instructions from lists

synopsis
   delinst(title: in,
           organism: in, chromosome: in, piece: in, names: in,
           coordinates: in,
           coordinates2: in,
           directions: in, delinstp: in,
           inst: out, output: out)

files

   title: Delila instructions for the title, organism, chromosome and
      piece along with any other controls a user wants.  The title
      file is copied to the top of the inst file.

   organism: a list of names, one per line, for the organisms to get.
   If the file is empty the program assumes that it has been specified
   in the title file.

   chromosome: a list of names, one per line, for the chromosomes to get.
   If the file is empty the program assumes that it has been specified
   in the title file.

   piece: a list of names, one per line, for the pieces to get.
   If the file is empty the program assumes that it has been specified
   in the title file.

   names: a list of names, one per line, for the pieces to get
   If the file is empty no name instructions are written.

   coordinates: a list of numbers, one per line, for the pieces to get

      NOTE: if the first character of a coordinate is not a digit, (it
      could be '-' or '+') and there are no digits in the remainder of
      the string, the entire get will be dropped.

      The Delila instruction format is:
       get from [coordinate] [from] to same [to] [direction];

   coordinates2: as with coordinates, the second coordinate of a get
      replacing the 'same'.

   directions:  list of directions (+ or -), one per line, for the
      pieces to get.   'f' (forward) and 'r' (reverse) is also
      acceptable.  The positive orientation is set for '+' or 'f'. 
      Otherwiswe the negative orientation is assumed.  No checking is
      done.

   inst:  The title file followed by name and get instructions according
      to the format:

      name "[NAMES]";
      get from [COO] [REVDIR][FROMRANGE]
            to same     [DIR][TORANGE] direction [DIR];

      where 

      [NAMES]     come from the 'names' file
      [REVDIR]    is generated by reversing the [DIR]
      [FROMRANGE] the from-range for the instructions from the 'delinstp' file
      [COO]       come from the 'coordinates' file
      [TORANGE]   the   to-range for the instructions from the 'delinstp' file
      [DIR]       come from the 'directions' file

      The name and get instruction are written to the inst file on one
      line; they are shown above on several lines for clarity.

   delinstp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      fromrange, torange (integer, integer): the range from and to for
      the Delila instructions.

      maxinstructions (integer): maximum number of instructions to write.
         A negative value means to write all of them.
         This is useful for testing before production runs.

      removeduplicates (char): If 'r' then remove duplicate Delila instructions
         for organism, chromosome and piece.

      multiline (char): If 'm' then put instructions on multiple lines.

      twocoordinates (char): If 't' then use both the coordinates and
         the coordinates2 files to get the coordinates.
         If the value is not 't', then the coordinates2 file
         is not looked at (and so need not exist).

description

   Often one needs to generate Delila instructions from data sets. 
   This program automates the task.

   The coordinates are read in and instructions are made corresponding
   to them.

   If the names file is empty or reaches the end of file, the number
   of the instruction is given.

   If the coordinate file is empty or reaches the end of file, '+' is
   assumed.

examples

documentation

see also

   {Program that uses the instructions:} delila.p

author

   Thomas Dana Schneider

bugs

technical notes

   This function is now NOT active since it interferes with Delila:
   NOTE: If the string for a chromosome name begins with 'chr' and is
   only 5 characters long, delinst will add a '0' after 'chr'.  Thus
   the name 'chr1' becomes 'chr01' and will sort correctly using the
   Unix sort.

*)
(* end module describe.delinst *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.48; (@ of prgmod.p 2011 May 03 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.48; (@ of prgmod.p 2011 May 03 *)

var
   title,
   organism, chromosome, piece, names,
   coordinates,
   coordinates2,
   directions, (* files used by this program *)
   delinstp, (* file used by this program *)
   inst: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module copyALLlines *)
procedure copyALLlines(var fin, fout: text);
(* copy ALL lines of file fin to file fout. *)
begin (* copyALLlines *)
      while (not eof(fin)) do begin
         copyaline(fin, fout);
      end;
end; (* copyALLlines *)
(* end module copyALLlines *)

(* begin module copyaword *)
procedure copyaword(var fin, fout: text);
(* copy a word from file fin to file fout *)
begin (* copyaword *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
end; (* copyaword *)
(* end module copyaword *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module readstringline *)
procedure readstringline(var afile: text; var buffer: string);
(* get a line (as a string) from a file. *)
var   index: integer; (* of buffer *)
begin (* getstring *)
   clearstring(buffer);
   index := 0;
   while (not eoln(afile)) and (index < maxstring) do begin
      index := succ(index);
      read(afile, buffer.letters[index])
   end;
   if not eoln(afile) then begin
      writeln(output, 'readstringline: a line exceeds maximum string size (',
                      maxstring:1,')');
      halt
   end;
   buffer.length := index;
   buffer.current := 1;
   readln(afile);
end; (* readstringline *)
(* end module readstringline version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* Test for equality between two strings at current positions.
NOTE:  A compiler bug results if one directly tests this way:
   if thedefinition^.nametag = aname
The reason is completely not clear!  I showed that the parts of the
strings were identical, but the whole was not by this test.
For this reason it is *critical to test strings with equalstring.
*)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module copystring *)
procedure copystring(a: string; var b: string);
(* copy string a to b.  The length, characters and
where the string points to are all copied. *)
var l: integer; (* index to the string *)
begin
   b.length := a.length;
   for l := 1 to a.length do b.letters[l] := a.letters[l];
   b.current := a.current;
   b.next := a.next;
end;
(* end module copystring *)

(* begin module copyfile *)
procedure copyfile(var fin, fout: text);
(* copy the rest of file fin to fout *)
begin
   while not eof(fin) do copyaline(fin, fout);
end;
(* end module copyfile version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module fixchrname *)
procedure fixchrname(var chr: string);
(* If the string begins with 'chr' and is only 5 characters
long, add a '0'.  Thus the name 'chr1' becomes 'chr01' and will sort
correctly using the Unix sort. *)
begin
  with chr do begin
     if length < 5 then begin
        if letters[1] = 'c' then begin
           if letters[2] = 'h' then begin
              if letters[3] = 'r' then begin
                 letters[5] := letters[4];
                 letters[4] := '0';
                 length := 5;
              end
           end
        end;
{zq
write(output,'fixed: "');
writestring(output,chr);
writeln(output,'"');
 halt;
}
     end
  end
end;
(* end module fixchrname *)

(* code stolen from lister: *)

(* begin module delinst.upgradeto114 *)
procedure upgradeto114(var delinstp: text);
(* upgrade the delinstp file to version 1.14:
*)
const
   copylines = 4; (* the number of copied lines *)
var
   internal: text; (* a place to hold the old delinstp *)
   line: integer; (* a line to be worked with *)
   parameterversion: real; (* parameter version number *)
begin
   parameterversion := 1.14;
   writeln(output, 'upgrading to version ',parameterversion:4:2,' ...');

   (* copy alist to internal *)
   reset(delinstp);
   readln(delinstp); (* skip old parameter version line *)
   rewrite(internal);
   for line := 1 to copylines
   do copyaline(delinstp, internal);

   (* write the NEW PARAMETER LINE *)
   writeln(internal, 'f       twocoordinates (char): If "t" use coordinates2');

   (* finish the copy of alist to internal *)
   copyfile(delinstp, internal);

   (* copy internal to alist *)
   reset(internal);
   rewrite(delinstp);
   writeln(delinstp,parameterversion:4:2,'        version of',
         ' delinstp that this parameter file is designed for.');
   copyfile(internal, delinstp);

   (* add the new material at the end: *)
   (* none to add *)

   reset(delinstp); (* ready to start reading again *)
   readln(delinstp); (* skip the new version number *)

end;
(* end module delinst.upgradeto114 *)

(* begin module delinst.upgradeparameters *)
procedure upgradeparameters(var delinstp: text);
(* make sure that the parameters are the latest spiffy version *)
var
   parameterversion: real; (* parameter version number *)

(* Since parameters are real numbers, there can be round off errors in
the 15th decimal place that prevent comparisons.  However, I only use
parameters to 2 decimal places, so we can multipy by 100 and round to
do reliable comparisons. *)

function r100(a: real): integer;
(* simple multiply and round function *)
begin
 r100 := round(100*a);
end;

function equalversion(a,b: real): boolean;
(* Are versions a and b equal? *)
begin
   if r100(a) = r100(b)
   then equalversion := true
   else equalversion := false
end; (* equalversion *)

function lessthan(a,b: real): boolean;
(* Is version a less than b? *)
begin
   if r100(a) < r100(b)
   then lessthan := true
   else lessthan := false
end; (* lessthan *)

function greaterthan(a,b: real): boolean;
(* Is version a greater than b? *)
begin
   if r100(a) > r100(b)
   then greaterthan := true
   else greaterthan := false
end; (* greaterthan *)

begin

   reset(delinstp);
   readln(delinstp, parameterversion);

   if    lessthan(parameterversion, updateversion) or
      greaterthan(parameterversion, versionupperbound)
   then begin
      writeln(output, 'You have an old parameter file!, version ',
              parameterversion:4:2,'!');
      writeln(output, '          version = ',          version:4:2);
      writeln(output, '    updateversion = ',    updateversion:4:2);
      writeln(output, ' parameterversion = ', parameterversion:4:2);
      writeln(output, 'versionupperbound = ',versionupperbound:4:2);

      if equalversion(parameterversion,  1.07) then upgradeto114(delinstp);

   end;

end;
(* end module delinst.upgradeparameters *)

(* begin module delinst.themain *)
procedure themain(var title,
   organism, chromosome, piece,
   names, coordinates, coordinates2, directions, delinstp, inst: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   fromrange: integer; (* from range *)
     torange: integer; (*   to range *)
   removeduplicates: char; (* r: remove duplicate org, chr, pie instructions *)
   keepduplicates: boolean; (* true if removeduplicates is not r *)
   multiline: char; (* If 'm' then put instructions on multiple lines. *)
   twocoordinates: char; (* use two coordinates from files coordinates
                            and coordinates2 *)
   dotwocoordinates: boolean; (* do two coordinates *)

   coo:  integer; (* a coordinate from the coordinates  file *)
   coo2: integer; (* a coordinate from the coordinates2 file *)
   coostring: string; (* string representing a coordinate *)
   coofile: text; (* file for converting coostring to coo *)
   coocounter: integer; (* a counter for coostring *)
   checkcoordinates: boolean; (* expensive check of the coordinates *)
   coodigits: integer; (* count of coordinate digits *)
   skipinstruction: boolean; (* skip writing the instruction *)
   skipstring: string; (* string used for skipping *)

   coodir: integer; (* a direction expressed as -1 or +1 *)
   dir: char; (* a direction + or - *)
   rid: char; (* the reverse direction - or + *)
   maxinstructions: integer; (* maximum instructions to write *)
   instructions: integer; (* number of instructions written so far *)
   skipped: integer; (* number of instructions skipped *)
   dataline: integer; (* line of data being read in *)
   neworgname, newchrname, newpiename: string; (* names of new specifications *)
   oldorgname, oldchrname, oldpiename: string; (* names of old specifications *)
   gotten: boolean; (* was the string read? *)
procedure givecoo(var afile: text; v: integer);
(* write a + or - to afile depending on the value v *)
begin
   if v < 0
   then write(afile,'-')
   else write(afile,'+')
end;
procedure endinst(quoteline: char);
(* write to the inst file the end depending on multiline *)
(* If quoteline is'"' then put " at the end of the instruction. *)
begin
         if quoteline = '"' then write  (inst,'"');
         write(inst,';');
         if multiline = 'm' then writeln(inst)
                            else write  (inst,' ');
end;
begin
   writeln(output,'delinst ',version:4:2);
   reset(delinstp);
   readln(delinstp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      (* upgrade from version 107 *)
      upgradeparameters(delinstp);
{
      writeln(output, 'You have an old parameter file!');
      halt
}
   end;

   readln(delinstp, fromrange, torange);
   readln(delinstp, maxinstructions);
   readln(delinstp, removeduplicates);
   readln(delinstp, multiline);
   readln(delinstp, twocoordinates);
   if twocoordinates = 't'
   then dotwocoordinates := true
   else dotwocoordinates := false;

   keepduplicates := (removeduplicates <> 'r');

   reset(title);
   reset(organism);
   reset(chromosome);
   reset(piece);
   reset(names);
   reset(coordinates);
   if dotwocoordinates then reset(coordinates2);
   reset(directions);

   rewrite(inst);
   copyALLlines(title, inst);

   clearstring(neworgname);
   clearstring(newchrname);
   clearstring(newpiename);

   clearstring(oldorgname);
   clearstring(oldchrname);
   clearstring(oldpiename);

   oldorgname.letters[1] := '.';
   oldorgname.length := 1;
   clearstring(oldchrname);
   clearstring(oldpiename);

   instructions := 0;
   skipped := 0;
   dataline := 0;
   checkcoordinates := false;

   while (not eoln(coordinates))
         and (   (instructions < maxinstructions)
              or (maxinstructions < 0)  )
   do begin
      dataline := succ(dataline);

      skipinstruction := false;
      if checkcoordinates or
            (not (coordinates^ in
            ['0','1','2','3','4','5','6','7','8','9']))
         (* This is a very time-expensive test of the coordinate so it
            should only be used to locate the problem
            if the program would otherwise crash *)
      then begin
         getstring(coordinates, coostring, gotten);
         {writestring(output,coostring); writeln(output);}
         if not gotten then halt;
         coodigits := 0;
         for coocounter := 1 to coostring.length do begin
            if (coostring.letters[coocounter] in
            ['0','1','2','3','4','5','6','7','8','9'])
            then coodigits := succ(coodigits);
         end;
         if coodigits = 0 then begin
            writeln(output);
            writeln(output,'** Bad coordinate at line ',
                           dataline:1,' of the coordinates file **');
            write(output,'The bad coordinate string is: "');
            writestring(output,coostring);
            writeln(output,'"');
            writeln(output,'THE INSTRUCTION WILL NOT BE GENERATED');
            skipinstruction := true;
            skipped := succ(skipped);
         end;
         if not skipinstruction then begin
            rewrite(coofile);
            writestring(coofile, coostring);
            reset(coofile);
            (* Without testing if the coordinates contain '-' the   
            program would crash at this read. *)
            read(coofile, coo);
         end
      end
      else begin
           readln(coordinates, coo);
           if dotwocoordinates
           then readln(coordinates2, coo2);
      end;

      if not skipinstruction then begin
         (* Example Delila instruction:
            name "thrL"; get from 190 -100 to same +100 direction +;
         *)
         if not eoln(organism) then begin
            getstring(organism,neworgname, gotten);
            if not gotten then halt;
            if keepduplicates or (not equalstring(oldorgname, neworgname))
            then begin
               write(inst,'org ');
               writestring(inst,neworgname);
               endinst(' ');
            end;
            if not equalstring(oldorgname, neworgname) then begin
               copystring(neworgname, oldorgname);
            end;
         end;

         if not eoln(chromosome) then begin
            getstring(chromosome,newchrname, gotten);
(* 2013 Aug 26 Block changing the chromosome name
            fixchrname(newchrname);
*)
{
writeln(output,'at chromosome ');
writestring(output,newchrname);
writeln(output);
halt;
}
{zq}
            if not gotten then halt;
            if not equalstring(oldchrname, newchrname) then begin
               if multiline = 'm' then writeln(inst);
            end;
            if keepduplicates or (not equalstring(oldchrname, newchrname))
            then begin
               write(inst,'chr ');
               writestring(inst,newchrname);
               endinst(' ');
            end;
            if not equalstring(oldchrname, newchrname) then begin
               copystring(newchrname, oldchrname);
            end;
         end;

         if not eoln(piece) then begin
            getstring(piece,newpiename, gotten);
            if not gotten then halt;
            if keepduplicates or (not equalstring(oldpiename, newpiename))
            then begin
               write(inst,'pie ');
               writestring(inst,newpiename);
               endinst(' ');
            end;
            if not equalstring(oldpiename, newpiename) then begin
               copystring(newpiename, oldpiename);
               (* better format for names: *)
               if not keepduplicates then begin
                  if multiline = 'm' then writeln(inst);
               end
            end
         end;

         if not eoln(names) then begin
            write(inst,'name "');
            copyaword(names,inst);
            endinst('"');
         end;

         if eoln(directions) then begin
         end
         else begin
            readln(directions,dir);
            if dir in ['f', 'F']
            then dir := '+';

            if dir = '+'
            then rid := '-'
            else rid := '+';

            if dir = '+'
            then coodir := +1
            else coodir := -1

         end;

         write(inst, 'get from ', coo:1,' ');
         givecoo(inst, coodir*fromrange);
         write(inst, abs(fromrange):1);
         write(inst, ' to ');

         if dotwocoordinates
         then write(inst, coo2:1)
         else write(inst, 'same');

         write(inst, ' ');
         givecoo(inst, coodir*torange);
         write(inst, abs(torange):1);
         write(inst, ' direction ',dir);
         write(inst, ';');

         writeln(inst);
         instructions := succ(instructions);
      end (* writing an instruction *)
      else begin (* skipping an instruction *)
         (* get each string and write it to output for the user *)

         getstring(organism, skipstring, gotten);
         write(output,'  organism: ');
         writestring(output,skipstring); writeln(output);

         getstring(chromosome, skipstring, gotten);
         write(output,'chromosome: ');
         writestring(output,skipstring); writeln(output);

         getstring(piece, skipstring, gotten);
         write(output,'     piece: ');
         writestring(output,skipstring); writeln(output);

         getstring(names, skipstring, gotten);
         write(output,'      name: ');
         writestring(output,skipstring); writeln(output);

{
         getstring(coordinates, coostring, gotten);
done already above!
}
         write(output,'coordinate: ');
         writestring(output,coostring); writeln(output, ' (BAD)');

         getstring(directions, skipstring, gotten);
         write(output,' direction: ');
         writestring(output,skipstring); writeln(output);

      end (* skipping an instruction *)
   end; (* reading data *)

   writeln(output);
   writeln(output,instructions:1, ' instructions generated');
   writeln(output,skipped:1, ' instructions skipped');

end;
(* end module delinst.themain *)

begin
   themain(title,
   organism, chromosome, piece,
   names, coordinates, coordinates2, directions, delinstp, inst);
1: end.
