program encfrq(encseq, cmp, fout, output);
(* encfrq: encoded sequence frequency analysis
   thomas schneider copyright (c) 1990

   module libraries required: delman, matmods, prgmods, auxmods
*)

label 1; (* end of the program *)

const
(* begin module version *)
version = 1.53; (* of encfrq.p 1994 sep 5
origin prior to 1983 feb 26 *)
(* end module version *)

(* begin module describe.encfrq *)
(*
name
      encfrq: encoded sequence frequency analysis

synopsis
      encfrq(encseq: in, cmp: in, fout: out, output: out)

files
      encseq: the output of the encode program
      cmp: a composition from the comp program.
      fout: frequency tables for each parameter set.  these are followed
         by z values for each frequency.  if cmp is empty, then equal
         frequencies are assumed.
      output: messages to the user.

description
      the frequency of each n-tide (mono- or di- or etc) is displayed in
      fout.  the actual number of sequences passing through a particular
      n-tide and position (ie, a parameter window) is taken into account.
         a second set of tables of z values are also presented.
      these are calculated from the composition provided in comp (p, the
      probability of obtaining the n-tide), the actual number of
      occurences (b) and the number of sequences at that position (n).
      the distribution of b can be described as a binomial distribution,
      with mean (m) np and standard deviation (s) sqrt(npq).  b is then
      normalized to obtain z: z=(b-m)/s.  if n is large, then z is
      normally distributed, and the probabilities can be found on any
      table for the normal distribution (use a two tailed test).  a rule
      of thumb for when the normal distribution can be used is that
      both np and n(1-p) should be greater than 5.  locations that violate
      this rule are marked with a '*'.
         locations of the z table that contain z values of 3 or greater are
      displayed to the right of the z table.  since these look somewhat
      like a dna footprint, they are called z-footprints.  the output
      for dinucleotide z-footprints is very wide, so one must split
      it up using the split program.  recommended values for splitp are
      p/14/112/4, where the slash means "start a new line".

see also
      encode.p, comp.p, split.p

author
      thomas d. schneider

bugs
      none known

*)
(* end module describe.encfrq *)

      (* more constants *)
      vectormax = 1616; (* vector maximum size *)

type
(* begin module encode.type.param *)
(* the following types allow the user to specify parameters which will
   be used to encode the sequences in the book. *)
(* spaces are allowed between the encoded bases, and the number of
   bases to be skipped between each encoded pair of bases are kept
   in this linked list of integers *)
      spaceptr = ^spacelist;
      spacelist = record
         skips : integer; (* bases skipped to next coded base *)
         next : spaceptr; (* points to next spacing number *)
      end;

(* the encoding parameters for each region are stored in these records,
   and the records for each region are connected into a linked list of
   all the encoding parameters for the entire sequences *)
      endpoints = (start,stop);  (* end points of a coding region *)
      paramptr = ^parameter;
      parameter = record  (* these are the instructions for doing the coding *)
         range : array[endpoints] of integer; (* the bases to be coded by these
                                 instructions, relative to the alignedbase *)
         window : integer; (* the number of bases included in a coding vector *)
         wshift : integer;  (* movement of the window to the next site *)
         coding : integer; (* the 'level' at which the coding is being done,
                             i.e., 1: mononucleotides; 2: dinucleotides; ... *)
         spaces : spaceptr;  (* the spacing between the encoded bases *)
         cshift : integer;  (* shift to the next coding unit in the window *)

         (* values calculated at read-in time *)
         wvlength: integer; (* length of a window vector.  this is coding
            raised to the 4th power *)
         pvlength: integer; (* parameter vector length.  the vector made up
            of a series of window vectors. *)
         (* note: pvlength/wvlength is the number of windows in the
            parameter range *)

         next : paramptr;  (* set of instructions for coding another region *)
      end;

(* end module encode.type.param version = 'matmod 1.54  82 dec 7 tds/gds'; *)

      vector = packed array[1..vectormax] of integer; (* define a vector *)

(* begin frozen module base.type *)
      base = (a,c,g,t);
(* end frozen module base.type *)

(* begin module comp.type *)
            (* the composition is stored in a tree of these nodes *)
      compnodeptr = ^compnode; (* points to a node of the tree *)
      compnode = record    (* a node of the composition tree *)
         count : integer; (* the number of oligos for this node *)
         son : array[base] of compnodeptr; (* the pointers to 'descendants'
                                           of this node of the tree *)
      end;

            (* spiders are used to make the composition tree *)
      spiderptr = ^spider; (* points to a 'spider' *)
      spider = record  (* a spider climbs the composition tree, its path
                        determined by the sequences, and increments 'count' at
                        all the nodes it passes, thereby determining the
                        composition *)
         depth : integer; (* the level of the node now at *)
         place : compnodeptr; (* a pointer to the current node *)
         next : spiderptr; (* the next spider in the collection *)
      end;

            (* the total number of composition entries at a given level
               is stored in the linked list of type comptotal *)
      compptr = ^comptotal;
      comptotal = record
         count: integer;  (* the number at a given level *)
         next: compptr;  (* pointer to the next level totals *)
      end;

           (* 'path' is used in printing the tree *)
      pathptr = ^path; (* pointer into a path on the tree *)
      path = record  (* the path of bases to get to a particular node *)
         bas : base;
         next : pathptr;
      end;
(* end module comp.type version = 'auxmod 1.32  85 feb 17 gds/tds'; *)

var
      encseq, cmp, fout: text; (* encoded sequences, composition and
          output files *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'auxmod 1.32  85 feb 17 gds/tds'; *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'prgmod 3.88  82 nov  3 tds'; *)

(* begin module package.encode *)
(* ************************************************************************ *)
(* begin module encode.readencpar *)
(* read encode parameters *)

function paramsize(param: paramptr): integer;
(* determine the size of the vector generated by one parameter record *)
var   rangesize,  (* the number of bases in the range *)
      numwindows: integer; (* the number of windows which start in the
                  range, taking into account the shifting *)
begin
      with param^ do begin
         rangesize := range[stop] - range[start] + 1;
         numwindows := trunc((rangesize -1)/wshift + 1);
         paramsize := numwindows * wvlength
      end
end; (* paramsize *)


procedure readencpar(var thefile: text; var param: paramptr;
                    var regions, endseq: integer);
(* read the parameter list from 'thefile' and put the information into the
   linked parameter list, beginning with 'param'; the number of parameter
   records is returned in 'regions'; the symbol that ends the sequence
   is 'endseq' *)
var   aparam,  (* from the parameter list *)
      newparam: paramptr;  (* the new parameter record from 'thefile' *)
      firstspaces,  (* in the spaces list *)
      newspaces: spaceptr; (* for the spaces list *)
      i,j: integer;  (* indices *)
      ch: char;  (* to check for proper positioning *)
begin
      reset(thefile);
      if eof(thefile) then begin
         writeln(output,' encoded sequence file is empty');
         halt
      end;
      new(aparam);
      if param = nil then new(param);
      aparam := param;
            (* skip the header *)
      for i := 1 to 2 do readln(thefile);

            (* read the end of sequence symbol *)
      readln(thefile,endseq);

            (* skip a blank line in the header *)
      readln(thefile);

            (* read the number of regions *)
      readln(thefile,regions);

      for i := 1 to regions do begin
         with aparam^ do begin
               (* read the range *)
            read(thefile,range[start]);
            repeat read(thefile,ch) until (ch = 'o');
            readln(thefile,range[stop]);
               (* read the window size and shift *)
            readln(thefile,window);
            readln(thefile,wshift);
               (* read the coding level and arrangement *)
            read(thefile,coding);
            new(spaces);
            if (coding > 1) then begin
               repeat read(thefile,ch) until (ch = ':');
               new(firstspaces);
               spaces := firstspaces;
               read(thefile,spaces^.skips);
               for j := 2 to coding -1 do begin
                  new(newspaces);
                  spaces^.next := newspaces;
                  spaces := newspaces;
                  read(thefile,spaces^.skips);
               end;
               spaces^.next := nil;
               spaces := firstspaces
            end
            else spaces := nil;
            readln(thefile);
            readln(thefile,cshift);

            (* set up wvlength *)
            wvlength := round(exp(coding*ln(4)));
         end;

         (* set up pvlength *)
         aparam ^.pvlength := paramsize(aparam);

         if (i < regions) then begin
            new(newparam);
            aparam^.next := newparam;
            aparam := newparam;
         end;
      end;
      aparam^.next := nil;
end;

(* end module encode.readencpar version = 'matmod 1.54  82 dec 7 tds/gds'; *)

(* begin module encode.functions *)
(* these functions are quite handy to help one know where one is
in an encoded vector.  they all take a parameter list as read by
readencpar, and the number of the element of the vector one is interested
in.  this element value runs from 0 to one less than the length of the
parameter vector.  examples are in evprint. *)

function beginpv(aparam: paramptr; element: integer): boolean;
(* are we at the beginning of a parameter vector? *)
begin (* beginpv *)
      if aparam <> nil (* is there a vector? *)
      then beginpv := (element = 0)
      else beginpv := true (* note circular use of aparam in evprint... *)
end; (* beginpv *)

function endpv(aparam: paramptr; element: integer): boolean;
(* are we at the end of a parameter vector?
the zero element is the first element of the parameter vector. *)
begin (* endpv *)
      if aparam <> nil
      then endpv := (element >= aparam^.pvlength)
      else endpv := true (* past the end of all vectors *)
end; (* endpv *)

function beginwv(aparam: paramptr; element: integer): boolean;
(* are we at the beginning of a window vector in the current parameter
vector?  the zero element is the first element of the parameter vector. *)
begin (* beginwv *)
      if aparam <> nil
      then with aparam^ do beginwv := ((element mod wvlength) = 0)
      else beginwv := false (* no further parameters *)
end; (* beginwv *)

function endwv(aparam: paramptr; element: integer): boolean;
(* are we at the end of a window vector in the current parameter vector?
the zero element is the first element of the parameter vector. *)
begin (* endwv *)
      if aparam <> nil
      then with aparam^ do endwv := (((element + 1) mod wvlength) = 0)
      else endwv := true (* no further parameters *)
(*;writeln(output,'element=',element:1,'   wvlength=',aparam^.wvlength:1); *)
end; (* endwv *)

function encposition(aparam: paramptr; element: integer): integer;
(* this returns the aligned position corresponding to the element in
a parmaeter list (aparam).
the zero element is the first element of the parameter vector. *)
begin (* encposition *)
      with aparam^
         do encposition := range[start] + (element div wvlength) * wshift
end; (* encposition *)
(* end module encode.functions version = 'matmod 1.54  82 dec 7 tds/gds'; *)

(* begin module encode.encread *)
procedure encread(var afile: text;  (* the file to read from *)
                  var val: integer; (* the value to read *)
                  var firstparam,   (* the first parameter *)
                      aparam:       (* the current parameter *)
                         paramptr;
                  var element: integer; (* the position in the vector *)
                  var wvend,        (* true = end of window vector *)
                      pvend,        (* true = end of parameter vector *)
                      vbegin,       (* true = begin of vector *)
                      seqend:       (* true = no more sequence *)
                         boolean;
                  var wvsum: integer; (* sum of window vector elements *)
                  var wvempty: boolean); (* true = the window vector just
                         read contains all zeros and so the window was off
                         one end of the sequence *)
(* encode read.  one vector value is returned from the encoded
file afile.  encread keeps track of what element of the vector is
read.  if you do not need this information, then the much simpler:
      read(afile, val);
      if val < 0 then begin
         readln(afile);
         if eof(afile) then ...(handle end of sequences)
                       else read(afile, val)
      end
will do.
      notes
1) to initialize:
      set aparam := nil; firstparam := nil;
and everything else (including parameter reading) will be taken care of.
alternatively, one can read the parameters first into firstparam and then
set aparam to nil.
2) wvempty is only meaningful when wvend is true.
*)
var
      regions, endseq: integer; (* for readencpar *)
begin (* encread *)
      vbegin := false; (* try to disprove this... *)
      if aparam = nil then begin
         aparam := firstparam;
         if aparam = nil (* still... *)  then begin
            readencpar(afile,firstparam,regions,endseq);
            if endseq >= 0 then begin
               writeln(output,' encread: endseq must be negative.');
               halt
            end;
            aparam := firstparam
         end;
         element := 0; (* assure zero on first read *)
         seqend := false;
         vbegin := true (* vector begins at start of reading *)
      end
      else begin
         element := succ(element);
         if endpv(aparam,element) then begin
            aparam := aparam^.next;
            if aparam = nil then aparam := firstparam; (* re-use the list *)
            element := 0 (* assure it is zero at the first read *)
         end
      end;

      read(afile, val);

      if val < 0
      then begin
         if element <> 0 then begin
            writeln(output,' encread: end of vector detected but not',
                           ' time to begin: element = ',element:1);
            halt
         end;
         readln(afile);

         if eof(afile)
         then seqend := true
         else begin
            read(afile, val);
            vbegin := true (* the next vector begins *)
         end
      end;

      if not seqend then with aparam^ do begin
         if beginwv(aparam, element) then wvsum := 0;
         wvsum := wvsum + val;

         wvend := endwv(aparam, element);
         pvend := endpv(aparam, element);

         if wvend
         then wvempty := (wvsum = 0)
         else wvempty := false
      end
      else begin (* end of sequence case *)
         wvend := true;
         pvend := true;
         vbegin := false;
         wvempty := false
      end
end; (* encread *)
(* end module encode.encread version = 'matmod 1.54  82 dec 7 tds/gds'; *)

(* begin module encode.evprint *)
(* procedures for encoded vector printing *)

procedure evln(var afile: text; element: integer; curparam: paramptr);
(* see evprint for definitions.  this procedure is used to provide
writelns to afile when the user wants to control the display better. *)
begin (* evln *)
      (* for end of windows: *)
      if beginwv(curparam,element)
      then writeln(afile)
      else if curparam = nil then writeln(afile); (* end of vector *)

      (* for end of parameters: *)
      if beginpv(curparam,element) then writeln(afile)
end; (* evln *)

procedure evheader(var afile: text;
                       posfield, valfield: integer;
                       firstparam: paramptr;
                   var curparam: paramptr);
(* see evprint for definitions of the variables and how to use evheader.
evheader must be called if you want to provide your own writelns
to the evprint display, ie, when ln is false *)
const
      codingmax = 4; (* the maximum coding level allowed *)
var
      bases: array[1..codingmax] of char; (* the bases *)
      i,j,n: integer; (* indices for loops *)
      size: integer; (* the number of characters in the encoded object.
                        a=1, aa=2, axa=3, etc *)
      s: spaceptr; (* for finding size *)
procedure space(i:integer);
(* put out i spaces to afile *)
var
      index: integer; (* for i *)
begin (* space *)
      for index := 1 to i do write(afile,' ');
end; (* space *)
procedure advance(level: integer);
(* advance the bases at the level.  aa -> ac -> ag -> ... -> tt *)
begin (* advance *)
      case bases[level] of
         'a': bases[level] := 'c';
         'c': bases[level] := 'g';
         'g': bases[level] := 't';
         't': begin
            bases[level] := 'a';
            advance(level - 1) (* aha.. *)
         end
      end
end; (* advance *)
begin (* evheader *)
      if curparam = nil then curparam := firstparam; (* start or recycle *)

      with curparam^ do begin
         if coding > codingmax then begin
            writeln(output,' evheader: coding levels must not be higher',
                           ' than ', codingmax:1);
            halt
         end;

         (* find the size of the encoded object *)
         size := coding;
         s := curparam^.spaces;
         while s <> nil do begin
            size := size + s^.skips;
            s := s^.next
         end;

         (* see if there is room for printing *)
         if abs(valfield) < size then begin
            writeln(output,' evheader: valfield (',valfield:1,
                           ') can not be less than coding level and spaces (',
                           size:1,')');
            halt
         end;

         (* clear bases *)
         for i := 1 to coding do bases[i] := 'a';

         (* print the header *)
         (* room for space, ' to ' and two numbers *)
         space(1 + 4 + 2*posfield);
         for i := 1 to wvlength do begin
            space(1 + abs(valfield) - size);
            s := spaces;
            for n := 1 to coding do begin
               write(afile, bases[n]);
               if s <> nil then begin
                  for j := 1 to s^.skips
                     do write(afile,'x');
                  s := s^.next
               end
            end;

            (* advance the bases except for the last time *)
            if i <> wvlength then advance(coding)
         end
      end
end; (* evheader *)

procedure evprint(var afile: text;         (* the file to print to *)
                      posfield: integer;   (* the number of characters in
                                              which to print the position *)
                      val: real;           (* the value to print *)
                      valfield,            (* the number of characters in
                                              in which to print the value *)
                      valdecimal: integer; (* the number of characters
                                              devoted to the digits below the
                                              decimal point of val.  if <= 0
                                              val is printed as an integer *)
                      ln: boolean;         (* control writelns (see below) *)
                      firstparam: paramptr; (* the first parameter *)
                  var curparam: paramptr;  (* the current encoding parameter *)

                  var element: integer);   (* the position in the vector *)
(* evprint is an encoded vector printing routine.  it is designed to make
generation of display of encoded vectors or calculated vectors based on
encoded vectors easy.  the basic idea is that evprint will print all the
display associated with one vector value.  so at the beginning it prints
an appropriate header from the vector parameters, and for each line it
prints the beginning and end positions of the window.
1) all fields printed out are separated by one space.
2) the value arguments are to be set by the user before calling evprint.
evprint is then called sequentially with all the vector values placed in val.
this allows the data to come from any data structure; it is not forced
to come from a particular array or file type.
3) to initialize, read in firstparam, set the element to 0 and curparam to nil.
evprint will restart at firstparam after each pass across the vector.
4) once the printing of the display has begun,
do not modify the element or curparam yourself.  it is also not wise to
modify any value parameters except val and possibly valdecimal.  the element
argument is used internally by evprint to keep track of the display.
the value arguments, after being set by the user initially, control the
form of the display.

control of writelns:  if ln is true, then evprint will take care of
writelns.  otherwise the user must do the writeln, as:
      if beginpv(curparam,element) then begin
         evheader(afile,posfield,valfield,firstparam,curparam);
         writeln(afile,[your headers for each column]);
      end;
      evprint(afile,...false...)
      if beginwv(curparam,element)
         then write(afile,[your own material on the same line]);
      evln(afile,curparam,element);
note that evprint had moved the element to the next position, so the
test is for begin of wv, not end of wv.

if valfield is negative (but abs(valfield) is a reasonable size) then
val will not be printed, allowing the user to do so after evprint.
this allows more flexible displays.
note: ln must be false to use this feature.
 *)
procedure sider;
(* produce the side of the display *)
var
      position: integer; (* the location that a window begins
                            from the aligned base *)
begin (* sider *)
      with curparam^ do begin
         position :=encposition(curparam, element);
         write(afile,' ',position: posfield, ' to ',
                 (position + window - 1):posfield)
      end
end; (* sider *)
procedure middler;
(* produce the middle of the display *)
begin (* middler *)
      write(afile,' ');
      if valfield > 0  (* to display or not to display... *)
      then if valdecimal <= 0 (* chose display *)
           then write(afile, round(val):valfield)
           else write(afile, val:valfield:valdecimal)
      else if ln then begin
         writeln(output,' evprint: if valfield is negative,',
                        ' then ln must be false');
         halt
      end
end; (* middler *)
procedure step;
(* step to the next position *)
begin (* step *)
      element := succ(element); (* move along the vector *)

      if endpv(curparam, element) (* are we off the end of the vector? *)
      then begin (* yes *)
         curparam := curparam^.next;
         element := 0;
      end
end; (* step *)
begin (* evprint *)
      if ln then if beginpv(curparam,element) then begin
         evheader(afile, posfield, valfield, firstparam, curparam);
         writeln(afile)
      end;

      if beginwv(curparam, element) then sider;

      middler;

      step;

      if ln then evln(afile,element,curparam)
end; (* evprint *)
(* end module encode.evprint version = 'matmod 1.54  82 dec 7 tds/gds'; *)

(* ************************************************************************ *)
(* end module package.encode version = 'matmod 1.54  82 dec 7 tds/gds'; *)

(* begin module comp.readcomp *)

(* begin module skipoligo *)
procedure skipoligo(var thefile: text);
(* this procedure is used to skip over an oligonucleotide string.
   it reads until it comes to a base, and then continues to read until
   it comes to a blank. no checking is done for non-base characters
   in between. *)
var ch : char;
begin
      repeat read(thefile,ch)
      until (ch in ['a','c','g','t']);
      repeat read(thefile,ch)
      until (ch = ' ');
end;
(* end module skipoligo version = 'auxmod 1.32  85 feb 17 gds/tds'; *)

procedure readcomp(var comp: text; var compmax: integer; readmax: integer;
                   var root: compnodeptr; var monocomptotal: compptr);
(* this procedure requires modules: comp.type, skipoligo, halt;
   this procedure reads from file 'comp' a composition and puts it into
   the tree pointed to by the 'root' pointer.  'compmax' is the depth
   of the composition tree which is stored.  it is the minimum of
   'readmax', the requested depth, and 'detcomp', the depth for which
   the input file composition was determined.  'monocomptotal' points
   to the beginning (for the monos) of a linked list which gives the
   totals for each level of the composition. *)
type  (* for efficient reading  of the information into the tree, each
         node that has a successor is stored in a queue as well as put
         into the tree.  the variables of the type list are the queue
         elements. *)
    listptr = ^list;
    list = record
      item: compnodeptr;  (* points to the composition tree node *)
      next: listptr;  (* points to the next list item in the queue *)
    end;

var  freeitem,  (* the list of unused list pointers *)
     listitem, (* an item in the queue *)
     first,  (* the first item in the queue *)
     last: listptr;  (* the last item in the queue *)
     comptot,  (* the comp total for a given level *)
     newcomptot: compptr;  (* for adding to the string of comptot"s *)
     detcomp,  (* the level to which the input composition was determined *)
     level,  (* of the composition being read, i.e., monos, dis, ... *)
     number: integer;  (* read from the 'comp' file *)
     ch: char;  (* for reading from 'comp' *)
     ba: base;  (* an index *)

(* getitem and clearitem provide efficient use of linked list storage by
   keeping a list of unused pointers that can be allocated instead of
   always creating 'new' ones  *)
procedure getitem(var l: listptr);
(* obtain a listitem from the free list or by making a new one *)
begin
      if freeitem<>nil
         then begin
            l:=freeitem;
            freeitem:=freeitem^.next
         end
         else new(l);
      l^.next:=nil
end;


procedure clearitem(var l: listptr);
(* return a listitem to the free list *)
var   lptr: listptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeitem;
         freeitem:=lptr
      end
end;

begin  (* readcomp *)
      reset(comp);
      if eof(comp) then begin
         writeln(output,' error: no composition file provided');
         halt;
      end;
      readln(comp); (* skip the program identification *)
      readln(comp); (* skip the book identification *)
      readln(comp,detcomp); (* obtain the determined composition *)
      readln(comp); (* skip the blank line *)

            (* determine the level of composition to be stored *)
      if (readmax < 1) then begin
         writeln(output);
         writeln(output,' warning: 0 or negative oligo length requested');
         writeln(output,' composition used is depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else if (readmax > detcomp) then begin
         writeln(output);
         writeln(output,' warning: requested composition oligo',
                        ' length (',readmax:1,')');
         writeln(output,' is larger than the determined composition',
                        ' oligo length (',detcomp:1,').');
         writeln(output,' composition used is to depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else compmax := readmax;

      new(root);
      new(monocomptotal);
      new(comptot);
      new(first);
      new(last);
      first^.item := root;
      first^.next := nil;
      last := first;
      new(freeitem);
      freeitem^.next := nil;

         (* read in the total number of bases from the composition *)
      readln(comp); (* skip the * *)
      readln(comp); (* skip the information line *)
      readln(comp,number);
      root^.count := number;

      repeat
         read(comp,ch); (* skip the space *)
         read(comp,ch); (* the ch determines what to do next *)
         if (ch = '*') then begin  (* determine the level about to be read *)
            readln(comp);
            readln(comp,level);
            if (level = 1) then monocomptotal := comptot
            else begin
               new(newcomptot);
               comptot^.next := newcomptot;
               comptot := newcomptot;
            end;
            comptot^.count := 0;
            comptot^.next := nil
         end
         else begin  (* read the composition values *)
            for ba := a to t do begin
               skipoligo(comp);
               read(comp,number);
               if (number <> 0) then begin
                  new(first^.item^.son[ba]);
                  first^.item^.son[ba]^.count := number;
                  comptot^.count := comptot^.count + number;
                  getitem(listitem);
                  last^.next := listitem;
                  last := listitem;
                  last^.next := nil;
                  last^.item := first^.item^.son[ba]
               end
               else first^.item^.son[ba] := nil;
            end;
            clearitem(first);
            readln(comp);
         end;
      until level = compmax;
         (* for the last level of compositions to be read we don"t
            need to store the nodes in the queue because their successors
            will not be read in. *)
      repeat
         for ba := a to t do begin
            skipoligo(comp);
            read(comp,number);
            if (number <> 0) then begin
               new(first^.item^.son[ba]);
               first^.item^.son[ba]^.count := number;
               comptot^.count := comptot^.count + number;
            end
            else first^.item^.son[ba] := nil;
         end;
         clearitem(first);
         readln(comp);
      until (first = nil);
end;  (* readcomp *)

function getcount(root: compnodeptr; start: pathptr): integer;
(* this function follows the tree from the node 'root' (which
   may be the root of a subtree) along the path initiated with
   'start', and returns the count of the resulting node.  if the
   path through the tree ever hits a null node in the process
   the count is returned to be zero. *)
var place: compnodeptr;  (* a place in the composition tree *)
    point: pathptr;  (* a point in the path through the tree *)
begin
      point:= start;
      place := root;
      while ((place <> nil) and (point <> nil)) do begin
         place := place^.son[point^.bas];
         point := point^.next;
      end;
      if (place = nil) then getcount := 0
      else getcount := place^.count;
end;

(* end module comp.readcomp version = 'auxmod 1.32  85 feb 17 gds/tds'; *)

(* begin module encode.vectorsize *)
function vectorsize(param: paramptr): integer;
(* determine the size of the vector generated by a string of parameter
   records, begin with 'param' *)
var   size: integer;  (* of the vector, so far *)
begin
      size := 0;
      while (param <> nil) do begin
         size := size + param^.pvlength;
         param := param^.next;
      end;
      vectorsize := size;
end;

(* end module encode.vectorsize version = 'matmod 1.54  82 dec 7 tds/gds'; *)


procedure clearvector(var v: vector);
(* clear the vector *)
var
      index: integer; (* to v *)
begin (* clearvector *)
      for index := 1 to vectormax
         do v[index] := 0
end; (* clearvector *)

procedure themain(var encseq, cmp, fout: text);
(* the main procedure *)
const
      posfield = 4; (* field size for sequence positions *)
      valfield = 6; (* field for values *)
      decfield = 3; (* how many decimal places *)
var
      sumx, (* sum of x *)
      n: (* numbers of x *)
         vector;

      firstparam, (* first parameter link *)
      aparam: (* one of the parameters *)
         paramptr;
      element: integer; (* index of the elements in the vectors *)

procedure header;
(* set up and display the header of the output *)
var
      regions, endseq: integer; (* for readencpar *)
begin (* header *)
      writeln(output, ' encfrq ',version:4:2);
      rewrite(fout);
      writeln(fout,' encfrq ',version:4:2,
                   ' encoded sequence frequencies from:');
      reset(encseq);
      copyaline(encseq,fout);
      copyaline(encseq,fout);
      writeln(fout);

      aparam := nil;
      firstparam := nil;
      readencpar(encseq,firstparam,regions,endseq);

      if vectorsize(firstparam) > vectormax then begin
         writeln(output,' the vectors are ',vectorsize(firstparam):1,
                        ' long.  this exceeds the maximum',
                        ' which is ',vectormax:1);
         halt
      end;
end; (* header *)

procedure scan;
(* scan the encoded sequences and sum them up *)
var
      pos, (* position in the vector *)
      x,   (* a value of a vector *)
      index: (* an index to the vector *)
         integer;
      (* variables for encread *)
      wvsum: integer;
      wvend, pvend, vbegin, seqend, wvempty: boolean;
begin (* scan *)
      clearvector(sumx);
      clearvector(n);

      repeat
         encread(encseq,x,firstparam,aparam,element,
                 wvend,pvend,vbegin,seqend,wvsum,wvempty);

         if not seqend then begin
            if vbegin then pos := 1
                      else pos := succ(pos);

            sumx[pos] := sumx[pos] + x;

            if wvend
               then if not wvempty
                  then with aparam^
                     do for index := pos downto pos - wvlength + 1
                        do n[index] := n[index] + wvsum;
         end
      until seqend
end; (* scan *)

procedure makefreq;
(* display the frequencies *)
var
      pos: integer; (* position in the vector *)
      freq: real; (* a frequency *)
begin (* makefreq *)
      element := 0;
      aparam := nil;

      for pos := 1 to vectorsize(firstparam) do begin
         if n[pos] = 0
         then begin (* make a blank spot *)
            if beginpv(aparam, element) then begin
               evheader(fout,posfield,valfield,firstparam,aparam);
               writeln(fout)
            end;
            evprint(fout,posfield,0,-1,0,false,firstparam,aparam,element);
            write(fout,' ':valfield);
            evln(fout,element,aparam)
         end
         else begin
            freq := sumx[pos]/n[pos];

            evprint(fout,posfield,freq,valfield,decfield,true,
                 firstparam,aparam,element)
         end
      end
end; (* makefreq *)

procedure makez;
(* make normalized z values *)
const
      maxcode = 2; (* only up to dinucleotides *)
      zflag = 3; (* smallest z value to flag in the z footprint *)
var
      curcoding: integer; (* for holding the coding after aparam
         becomes nil at the end of the vector *)
      (* mono and di-nucleotide probabilities *)
      pmono: array[0..3] of real; (* a to t *)
      pdi: array[0..15] of real; (* aa to tt *)

      pos: integer; (* position in a vector *)
      p, (* a probability from the composition *)
      q, (* 1-p *)
      mean, (* binomial mean *)
      stdev, (* binomial standard deviation *)
      z: (* a normalized b = frequency * n *)
         real;
      warn: integer; (* 1 means z estimate is poor, 0 means it is ok *)
      (* stored z values for printing z footprint *)
      ztable:   array [0..15] of integer; (* for dinucleotides *)
procedure makeprobabilities;
(* fill the probability arrays *)
var
      (* variables for composition procedures readcomp and getcount *)
      compmax, readmax: integer;
      root: compnodeptr;
      monocomptotal: compptr;
      start: pathptr;
      b1,b2: base; (* bases for comp *)
      total: real; (* the number of bases used to make the comp *)
      place: integer; (* the place in a probability array *)
procedure equalmono;
(* set pmono to all equal probabilities *)
var
      e: real; (* the probability *)
      p:integer; (* position in pmono *)
begin (* equalmono *)
      e := 1/4;
      for p := 0 to 3 do pmono[p] := e;
      writeln(fout,' equal mononucleotides assumed')
end; (* equalmono *)
procedure equaldi;
(* set pdi to all equal probabilities *)
var
      e: real; (* the probability *)
      p: integer; (* position in pdi *)
begin (* equaldi *)
      e := 1/16;
      for p := 0 to 15 do pdi[p] := e;
      writeln(fout,' equal   dinucleotides assumed')
end; (* equal *)
begin (* makeprobabilities *)
      reset(cmp);
      if eof(cmp) then begin
         equalmono;
         equaldi
      end
      else begin
         writeln(fout,' probabilities from:');
         copyaline(cmp,fout);
         copyaline(cmp,fout);

         readmax := maxcode;
         readcomp(cmp,compmax,readmax,root,monocomptotal);

         (* obtain the mono composition *)
         new(start);
         start^.next := nil;
         total := 0;
         for b1 := a to t do begin
            start^.bas := b1;
            pmono[ord(b1)] := getcount(root,start);
            total := total + pmono[ord(b1)]
         end;
         (* divide by total to get probabilities *)
         for place := 0 to 3
            do pmono[place] := pmono[place]/total;

         if compmax >= 2 then begin
            (* obtain the di composition *)
            new(start^.next);
            start^.next^.next := nil;
            total := 0;
            for b1 := a to t do begin
               start^.bas := b1;
               for b2 := a to t do begin
                  start^.next^.bas := b2;
                  place := 4*ord(b1) + ord(b2);
                  pdi[place] := getcount(root, start);
                  total := total + pdi[place]
               end
            end;
            (* divide by total to get probabilities *)
            for place := 0 to 15
               do pdi[place] := pdi[place]/total
         end
         else equaldi
      end
end; (* makeprobabilities *)
procedure makeheader;
(* make the header for a parameter *)
begin (* makeheader *)
      evheader(fout,posfield,valfield,firstparam,aparam);
      writeln(fout,'  z footprint')
end; (* makeheader *)
procedure zfootprint;
(* display the z values in footprint format *)
const
      symbol = '.'; (* the symbol to use for blank areas of the print *)
var
      column: integer; (* index to zmono and zdi *)
      size: integer; (* maximum for column *)
      z: integer; (* a truncated z value.  0 means do not print *)
procedure letter(i: integer);
(* convert a number 0 to 3 to a letter a to t *)
begin (* letter *)
      case i of
         0: write(fout,'a');
         1: write(fout,'c');
         2: write(fout,'g');
         3: write(fout,'t')
      end
end; (* letter *)
begin (* zfootprint *)
      case curcoding of
         1: size := 3;
         2: size := 15;
      end;
      write(fout,' ');
      for column := 0 to size do begin
         write(fout,' ');
         z := ztable[column];
         if abs(z) > 0 then begin
            if z > 0
            then write(fout,'+')
            else write(fout,'-');

            if abs(z) > 9 (* can not print directly *)
            then write(fout,'+')
            else write(fout,abs(z):1);

            if curcoding = 2
            then letter(column div 4);

            letter(column mod 4)
         end
         else begin (* put out blanks *)
            write(fout,symbol,symbol,symbol);
            if curcoding = 2 then write(fout,symbol)
         end
      end
end; (* zfootprint *)
begin (* makez *)
      page(fout);
      writeln(fout,' z values for frequencies: z footprint');
      writeln(fout);
      writeln(fout,' z  p(x >= z)');
      writeln(fout,' 1    0.3174');
      writeln(fout,' 2    0.0456');
      writeln(fout,' 3    0.0026');
      writeln(fout);
      makeprobabilities;
      writeln(fout);
      writeln(fout,' * means insufficient data for z estimate');
      writeln(fout);

      element := 0;
      aparam := firstparam; (* so that p can be found before first evprint *)
      for pos := 1 to vectorsize(firstparam) do begin
         if n[pos] = 0 then begin
            if beginpv(aparam, element) then makeheader;
            evprint(fout,posfield,0,-1,0,false,firstparam,aparam,element);
            write(fout,' ':valfield);
            evln(fout,element,aparam);
         end
         else begin
            with aparam^ do begin
               if coding > maxcode then begin
                  writeln(output,' maximum coding is ',maxcode:1,
                                 ' for this program');
                  halt
               end;
               curcoding := coding; (* capture the current value *)
               case curcoding of
                  1: p := pmono[element mod wvlength];
                  2: p := pdi[element mod wvlength]
               end
            end;

            (* calculate statistics *)
            q := 1 - p;
            mean := n[pos] * p;
            stdev := sqrt(mean * q);
            (* calculate z with continuity correction of 0.5: *)
            z := (sumx[pos] - 0.5 - mean)/stdev;

            if (mean < 5.0) or ((n[pos] * q) < 5.0)
            then warn := 1
            else warn := 0;

            (* record data for z footprint *)
            with aparam^ do
            if (* (warn = 0) and *) (abs(z) >= zflag)
            then ztable[element mod wvlength] := trunc(z)
            else ztable[element mod wvlength] := 0;

            if beginpv(aparam, element) then makeheader;
            evprint(fout,posfield,z,valfield-warn,decfield-warn,false,
                    firstparam,aparam,element);
            if warn = 1 then write(fout,'*');

            if beginwv(aparam, element) or endpv(aparam, element) then begin
               (* display z footprint at end of the line *)
               zfootprint;
            end;

            evln(fout,element,aparam);
         end;
         if beginpv(aparam, element) then begin
            (* end previous vector before start of next parameter vector *)
            writeln(fout);
         end
      end
end; (* makez *)

begin (* themain *)
      header;
      scan;
      makefreq;
      makez
end; (* themain *)
begin (* encfrq *)
      themain(encseq,cmp,fout);
1: end. (* encfrq *)
