program gedo(gedop, result, output);
(* gedo: gene dosage

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.08; (* of gedo.p 2015 May 12
2015 May 12, 1.08: upgrade documentation
2014 Nov 18, 1.07: upgrade documentation
2011 Apr 11, 1.06: upgrade documentation
2010 Sep 09, 1.05: finished
2010 Sep 09, 1.04: first part of results protected by *
2010 Sep 09, 1.03: fix commenting
2010 Sep 09, 1.02: upgrade, use result file
2010 Sep 09, 1.01: upgrade
2010 Sep 09, 1.00: origin (functional!)
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.gedo *)
(*
name
   gedo: gene dosage

synopsis
   gedo(gedop: in, result: out, output: out)

files

   result:  results in a table computed according to 
      Condon.Squires1993

   gedop:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      line 2: width and decimal places of output

      line 3: doubling time in minutes

      line 4: OriC minutes of a locus on the E. coli map

      remaining lines(real, string):
        minutes of locus
        name
      If a line begins with '*' then it is a comment line
      and is skipped.
      If a line begins with 'n' then it is calculated and
      shown but not incuded in the sum calculation.

   output: messages to the user

description

   Compute gene dosage in E. coli based on chromosome location of a
   locus, location of the origin and doubling time.

examples

In Condon.Squires1993
page 1411 is the example:

    "A wildtype E. coli strain with a doubling time of 24.1 min has an
    equivalent of 38 rRNA operons, due to the high replication rate on
    rich media (for calculation, see Materials and methods).

There is only a minor a discrepancy between this result (38) and the
following (36).

Example gedop from the Condon.Squires1993:

1.00 version of gedo that this parameter file is designed for.
6 3  width and decimal places of output
24.1 doubling time in minutes
  84.0 OriC minutes of a locus on the E. coli map
  84.5 rrnC
  86.5 rrnA
  89.8 rrnB
  90.5 rrnE
  05.1 rrnH
  56.1 rrnG
  72.1 rrnD
n 17.5 lambda att

result file:

gedo 1.02
24.1 doubling time (min)
84.0 OriC minutes of a locus on the E. coli map
 1.73 the specific growth rate in $h^{-1}$
   min      f      D      C      x  name
84.500  0.010 23.318 42.011  6.468  rrnC
86.500  0.050 23.318 42.011  6.163  rrnA
89.800  0.116 23.318 42.011  5.690  rrnB
90.500  0.130 23.318 42.011  5.595  rrnE
 5.100  0.422 23.318 42.011  3.932  rrnH
56.100  0.558 23.318 42.011  3.336  rrnG
72.100  0.238 23.318 42.011  4.910  rrnD
17.500  0.670 23.318 42.011  2.914  lambda att - skipped in sum
36.094 sum of x

documentation

See pages 4311 and 4314 of:

@article{Condon.Squires1993,
author = "C. Condon
 and S. French
 and C. Squires
 and C. L. Squires",
title = "{Depletion of functional ribosomal RNA operons in
\emph{Escherichia coli} causes increased expression of the remaining
intact copies}",
journal = "EMBO J",
volume = "12",
pages = "4305--4315",
pmid = "8223440",
pmcid = "PMC413727",
year = "1993"}

The equations come from Table 5, page 1538:

Bremer,H. and Dennis,P.P. (1987) In Neidhardt,F.C, Ingraham,J.L.,
Low,K.B., Magasanik,B., Schaechter,M. and Umbarger,H.E. (eds) ,
Escherichia coli and Salmonella typhimuriurn: Cellular and Molecular
Biology. American Society for Microbiology, Washington, DC, pp. 
1527-1542.

see also

  {Example parameter file: } gedop

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.gedo *)

var
   gedop, (* file used by this program *)
   result: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.72; {of delmod.p 2007 Jul 23} *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 7.72; {of delmod.p 2007 Jul 23} *)

(* begin module copytoeolnwidth *)
procedure copytoeolnwidth(var fin, fout: text; thewidth: integer);
(* copy a line from file fin to file fout,
set to total width using spaces at the end. *)
var
   c: integer; (* count of characters so far *)
begin (* copyaline *)
   c := 0;
   while not eoln(fin) do begin
      fout^ := fin^;
      put(fout);
      get(fin);
      c := c + 1;
   end;
   while c < thewidth do begin
      write(fout, ' ');
      c := c + 1;
   end;
end; (* copyalinewidth *)
(* end module copytoeolnwidth *)

(* begin module gedo.themain *)
procedure themain(var gedop, result: text);
(* the main procedure of the program *)
const
   namewidth = 11; (* width provided for names *)
var
   C: real; (* time taken to replicate the chromosome *)
   D: real; (* the time between completion of chromosomal replication and cell
               division *)
   dec: integer; (* decimal places of output numbers *)
   f: real; (* fractional distance from the origin of replication
               i.e. the distance in either direction from oriC divided by 50 *)
   fprime: real; (* f calculated according to 
       Bremer,H. and Dennis,P.P. (1987) In Neidhardt,F.C,
       Ingraham,J.L., Low,K.B., Magasanik,B., Schaechter,M. and
       Umbarger,H.E. (eds) , Escherichia coli and Salmonella
       typhimuriurn: Cellular and Molecular Biology. American Society
       for Microbiology, Washington, DC, pp.  1527-1542. *)
   g: real; (* doubling time in minutes *)
   k: real; (* the specific growth rate in $h^{-1}$ *)
   minutes: real; (* minutes around the chromosome of a locus (distance) *)
   oriC: real; (* minutes around the chromosome of oriC *)
   parameterversion: real; (* parameter version number *)
   skipsum: boolean; (* skip adding this value into the sum *)
   x: real; (* number of copies of a gene per cell *)
   xsum: real; (* sum of x values *)
   wid: integer; (* width of output numbers *)
begin
   writeln(output,'gedo ',version:4:2);
   reset(gedop);
   readln(gedop, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   rewrite(result);
   writeln(result,'* gedo ',version:4:2);

   (* read other parameters *)
   readln(gedop, wid, dec);
   readln(gedop, g);
   writeln(result, '* ', g:4:1,' doubling time (min)');
   write(result, '* ');
   read(gedop, oriC);
   write(result, oriC:4:1);
   copyaline(gedop, result);

   (* compute k from kminutes *)
   k := 60 * ln(2) / g;
   writeln(result, '*', k:5:2, ' the specific growth rate in $h^{-1}$');
   writeln(result,'*');
   writeln(result,'* min, minutes on the E. coli map');
   writeln(result,'* x, the number of copies of a gene per cell');
   writeln(result,
     '* f, is the fractional distance from the origin of replication;');
   writeln(result,
     '  i.e. the distance in either direction from oriC divided by 50.');

   writeln(result,
     '* C, the time taken to replicate the chromosome');
   writeln(result,
     '* D, the time between completion of chromosomal',
     ' replication and cell division');
   writeln(result,'*');

   xsum := 0.0;
   writeln(result, '*','min':(wid-1),
                   ' ','x':wid,
                   '  ','name':namewidth,
{
                   ' ':namewidth-3,'name':4,
}

                   ' ','f':wid,
                   ' ','C':wid,
                   ' ','D':wid);

   while not eof(gedop) do begin
      if gedop^ = '*' then begin
        readln(gedop);
      end
      else begin
         if gedop^ = 'n'
         then begin
            skipsum := true;
            get(gedop);
         end
         else skipsum := false;

         read(gedop, minutes);
         write(result, minutes:wid:dec);

{
oriC := 84;
}
         (* calculate f *)
         if oriC - minutes > 50
         then f := ((100 - oriC) + minutes)/50
         else f := abs(oriC - minutes)/50;

         (* original method matches the above: *)
         if (minutes >=  0) and (minutes <   36)
                                  then fprime := (minutes + 16)/50;
         if (minutes >= 36) and (minutes < oriC)
                                  then fprime := (oriC - minutes)/50;
         if (minutes >= oriC) and (minutes <= 100)
                                  then fprime := (minutes - oriC)/50;
         if f <> fprime then begin
            writeln(output, 'Warning: f <> fprime');
            writeln(output, 'Warning: ',f:wid:dec,' <> ',fprime:wid:dec);
            writeln(output, 'Warning: ',minutes:wid:dec,' minutes ');
         end;

         D := 21 + 4/k;
         C := 40.5 + 4.5/(k*k);
         (* compute 2^ ((C * (1-f) + D)/g) *)
         x := exp( ((C * (1-f) + D)/g) * ln(2));
         if not skipsum then xsum := x + xsum;

         write(result, ' ', x:wid:dec);
         write(result, ' '); (* extra space in front of name *)
         copytoeolnwidth(gedop, result,namewidth);
         readln(gedop);

         write(result, ' ', f:wid:dec);
         write(result, ' ', C:wid:dec);
         write(result, ' ', D:wid:dec);

         if skipsum then write(result, ' - skipped in sum');
         writeln(result);
      end;
   end;
   writeln(result, '* ', xsum:wid:dec,' sum of x')

end;
(* end module gedo.themain *)

begin
   themain(gedop, result);
1: end.
