program lig(input, list, output);
(* lig: DNA ligation theory
  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/
*)

const
(* begin module version *)
version = 1.28; (* of lig 1999 Aug 20
1999 Aug 20, 1.28: added example
previous changes: 1988 Jan 5
origin 1983 June 26 as basic lig.4
converted to pascal 1986 December 26 *)
(* end module version *)

(* begin module describe.lig *)
(*
name
   lig: DNA ligation theory

synopsis
   lig(input: in, list: out, output: out)

files
   input: user commands
   list: a 'hard copy' of the inputs and the outputs
   output: ligation predictions

description
   This program computes the results of a ligation reaction
   for insertion of a linker onto both ends of a linearized plasmid. 
   The user gives 
      the size of a plasmid in KB,
      the pico moles or micrograms of plasmid (you get to chose)
      the size of an insert in KB,
      the pico moles of or micrograms of insert
      the volume of the reaction in micro liters
   The program calculates whether circular or linear molecules are favored
   for the plasmid alone and with the insert.  The logic is:
   1.  The plasmid alone should circularize.
   2.  The plasmid with the insert should be linear.
   Thus, as the ligation proceeds, the first thing that happens is that
   the plasmid and insert ligate together (2 above).  Then the concentration of
   ends is lower, and circularization will be favored (1 above).
   Obviously this is all really rule of thumb, but it does seem to work
   in my experience.

documentation
   A. Dugaiczyk, H. W. Boyer and H. M. Goodman
   J. Mol. Biol. 96: 171-184 (1975)
   'Ligation of EcoRI Endonuclease-generated DNA fragements
    into Linear and Circular Structures'

author
   Thomas Dana Schneider

examples

| % lig
| lig 1.27
| plasmid size in kb: 4
| for the input plasmid, do you want to give ug or pMOL (u or p):u
| ug input plasmid: 1
| that is    0.37500 pMol
| 
| for the input linker, do you want to give ug or pMOL (u or p):u
| ug linker: 1
| linker size in kb: 0.025
| that is   60.00000 pMol
| volume in microliters: 60
| 
| for the plasmid alone
| critical [dna] = 0.0323 (g/l = micrograms/microliter)
| above this concentration: linears below: circles of the plasmid
| actual concentration of 1.000 micrograms of dna: 0.0167 (g/l)
| CIRCULAR MOLECULES FAVORED  (j/i =  1.939E+00)
| 
| for the plasmid mixed with linkers:
| ratio of linker molecules to plasmid:  1.600E+02
| j =  1.521E+13 (ends / mil)  effective concentration of plasmid ends
| i =  6.097E+17 (ends / mil)  total ends
| LINEAR MOLECULES FAVORED  (j/i =  2.494E-05)

This example shows that when the plasmid is mixed with linkers one would get
them to ligate together to form linears.  Then the concentration drops so
that the plasmid is effectively alone, and one would form circles.  This
would insert the linker into the plasmid - exactly what one wants!

bugs
   none known

*)
(* end module describe.lig *)

(* more constants *)
(* begin module lig.const *)
   width = 10; (* width of some numbers output *)
   decimal = 5; (* number of decimal places of some numbers output *)

   (* size of lambda phage in kilo base pairs: *)
   kblambda = 48.514;

   (* avagadro's number (molecules/mole): *)
   n0 = 6.022e+23;
(* end module lig.const *)

var
(* begin module lig.var *)
   list: text; (* copy of the terminal input and output *)

   kbplasmid, (* size of plasmid in kb *)
   kblinker, (* size of linker in kb *)
   pmolplasmid, (* pmol of plasmid *)
   pmollinker, (* pmol of linker *)
   vol, (* volume of the reaction in microliters *)
   mw, (* dna molecular weight *)
   dnaconc, (* dna concentration *)
   micrograms, (* number of micrograms of material *)
   aconc, (* actual concentration in ug/ul *)
   pmol, (* number of pico moles of material *)
   i, (* see documentation on i/j ratio *)
   j: (* see documentation on i/j ratio *)
      real;

   kind: char; (* the kind of  input desired *)
(* end module lig.var *)

begin
   writeln(output,'lig ',version:4:2);
   rewrite(list);
   writeln(list  ,'lig ',version:4:2);

   write(output, 'plasmid size in kb: ');
   write(list  , 'plasmid size in kb: ');
   readln(input, kbplasmid);
   writeln(list, kbplasmid:width:decimal);

   write(output, 'for the input plasmid, do you want to give',
                 ' ug or pMOL (u or p):');
   write(list  , 'for the input plasmid, do you want to give',
                 ' ug or pMOL (u or p):');
   readln(input,kind);
   writeln(list,kind);
   if kind = 'u' then begin
      write(output, 'ug input plasmid: ');
      write(list  , 'ug input plasmid: ');
      readln(input, micrograms);
      writeln(list, micrograms:width:decimal);
      pmolplasmid := 1.5 * micrograms / kbplasmid;
      writeln(output,'that is ',pmolplasmid:width:decimal,' pMol');
      writeln(list  ,'that is ',pmolplasmid:width:decimal,' pMol');
   end
   else begin
      write(output, 'pMOL input plasmid: ');
      write(list  , 'pMOL input plasmid: ');
      readln(input, pmolplasmid);
      writeln(list, pmolplasmid:width:decimal);
   end;

   writeln(output);
   writeln(list  );

   write(output, 'for the input linker, do you want to give',
                 ' ug or pMOL (u or p):');
   write(list  , 'for the input linker, do you want to give',
                 ' ug or pMOL (u or p):');
   readln(input,kind);
   writeln(list,kind);
   if kind = 'u' then begin
      write(output, 'ug linker: ');
      write(list  , 'ug linker: ');
      readln(input, micrograms);
      writeln(list, micrograms:width:decimal);

      write(output, 'linker size in kb: ');
      write(list  , 'linker size in kb: ');
      readln(input, kblinker);
      writeln(list, kblinker:width:decimal);
      pmollinker := 1.5 * micrograms / kblinker;
      writeln(output,'that is ',pmollinker:width:decimal,' pMol');
      writeln(list  ,'that is ',pmollinker:width:decimal,' pMol');
   end
   else begin
      write(output, 'pMOL input linker: ');
      write(list  , 'pMOL input linker: ');
      readln(input, pmollinker);
      writeln(list, pmollinker:width:decimal);
   end;

   write(output, 'volume in microliters: ');
   write(list  , 'volume in microliters: ');
   readln(input, vol);
   writeln(list, round(vol):width);

   mw := kbplasmid * 625000;
   (* from equation (5) of the paper: *)
   dnaconc := 51.1/sqrt(mw);
   writeln(output);
   writeln(list  );
   writeln(output, 'for the plasmid alone');
   writeln(list  , 'for the plasmid alone');
   writeln(output,'critical [dna] = ',dnaconc:6:4,
                  ' (g/l = micrograms/microliter)');
   writeln(list  ,'critical [dna] = ',dnaconc:6:4,
                  ' (g/l = micrograms/microliter)');
   writeln(output, 'above this concentration: linears',
                   ' below: circles of the plasmid');
   writeln(list  , 'above this concentration: linears',
                   ' below: circles of the plasmid');
   micrograms := pmolplasmid * kbplasmid /1.5;
   aconc := micrograms/vol;
   writeln(output, 'actual concentration of ',micrograms:5:3, ' micrograms',
                   ' of dna: ',aconc:6:4,' (g/l)');
   writeln(list  , 'actual concentration of ',micrograms:5:3, ' micrograms',
                   ' of dna: ',aconc:6:4,' (g/l)');
   if aconc>dnaconc then write(output, 'LINEAR');
   if aconc>dnaconc then write(list  , 'LINEAR');
   if aconc<dnaconc then write(output, 'CIRCULAR');
   if aconc<dnaconc then write(list  , 'CIRCULAR');
   write(output, ' MOLECULES FAVORED');
   write(list  , ' MOLECULES FAVORED');
   writeln(output, '  (j/i = ',51.1/(aconc*sqrt(mw)):10,')');
   writeln(list  , '  (j/i = ',51.1/(aconc*sqrt(mw)):10,')');
  
   (* from equation (2) of the paper: *)
   j:=3.6e+11 * exp((3/2) * ln(kblambda / kbplasmid));
   (* from equation (3) of the paper (modified for this case): *)
   (* pmol is the total moles of ends in the reaction *)
   (* 1e9 is the number of pmols per mol *)
   (* 1000 is the number of microliters per milliliter *)
   pmol := pmollinker + 2*pmolplasmid;
   i := n0 * pmol * (1/1e+09) * (1/vol) * 1000;
   writeln(output);
   writeln(list  );
   writeln(output, 'for the plasmid mixed with linkers:');
   writeln(list  , 'for the plasmid mixed with linkers:');
   writeln(output, 'ratio of linker molecules to plasmid: ',
                   pmollinker/pmolplasmid:10);
   writeln(list  , 'ratio of linker molecules to plasmid: ',
                   pmollinker/pmolplasmid:10);
   writeln(output, 'j = ',j:10,
         ' (ends / mil)  effective concentration of plasmid ends');
   writeln(list  , 'j = ',j:10,
         ' (ends / mil)  effective concentration of plasmid ends');
   writeln(output, 'i = ',i:10,' (ends / mil)  total ends');
   writeln(list  , 'i = ',i:10,' (ends / mil)  total ends');
   if i>j then write(output, 'LINEAR');
   if i>j then write(list  , 'LINEAR');
   if i<j then write(output, 'CIRCULAR');
   if i<j then write(list  , 'CIRCULAR');
   if i=j then write(output, 'LINEARS AND CIRCULAR');
   if i=j then write(list  , 'LINEARS AND CIRCULAR');
   write(output, ' MOLECULES FAVORED');
   write(list  , ' MOLECULES FAVORED');
   writeln(output, '  (j/i = ', j/i:10,')');
   writeln(list  , '  (j/i = ', j/i:10,')');
end.
