program migrate(migratep, output);
(* migrate: selective migration of drunkard's walk

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of migrate.p 1999 Dec 22
1999 Dec 22: 1.02 minor edits
previous changes: 1997 Feb 15
origin 1997 Feb 14 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.migrate *)
(*
name
   migrate: selective migration of drunkard's walk

synopsis
   migrate(migratep: in, output: out)

files
   migratep:  parameters to control the program.  The file must contain the
        following parameters, one per line:

         The version number of the program.  This allows the user to be
         warned if an old parameter file is used.

         range: (integer) +/- range of the world
	    This is the region that will be displayed.  A constant in the
	    program (see technical notes) defines the range of the array.

         displaymax: (integer) vertical display maximum
            How many vertical lines are devoted to the display.

         pagejump: 'l' means produce a page jump

         maxgens: (integer) maximum number of generations to run

         initpop: (integer) initial population placed at zero

         survival rate: (two integers)
             probability of survival for negative and positive steps

         iseed: (real) initial random seed

	 delaycount: (integer) number of times to do a futile addition to
	     slow the display.  Start with a small value and work up to a
	     reasonable delay.  3000000 gives near a 1 second delay on a Sun
	     Sparcstation 20/61.

	 touchhalt:  stop the program when this location is touched This
	     allows one to determine how many steps of selective evolution
	     are needed to get a particular distance from the origin.

   output: messages to the user

description

   The migrate program shows evolutionary selection in the context of the
   classical drunkard's walk.  In the drunkard's walk, a population of drunks
   comes out of a bar and starts stepping to the left or the right randomly.
   The drunks don't bump into each other, and so are really ghosts ...
   Statistics text books demonstrate that the distribution will approach a
   Gaussian distribution.

   The program adds only two new factors.  Each drunk really doesn't move.
   Instead "it" divides in two like a bacterium, and the two progeny are born
   one step to the left or the right of the parent.  The other new factor is
   that there are survival probabilities for the two directions.  By playing
   with these probabilities, the population can be made to migrate to the
   left or right.

   A natural example if this is the mechanism that bacteria use to find
   food.  They have two modes of movement, either tumbling or swimming.
   Which happens depends on the direction that they rotate their flagella.
   (Yes, it is a rotating part!  See Eisenbach1990.) The trouble for the
   bacterium is that it is too small to sense chemical gradients.  So it
   swims for a while and compares the new situation to the previous
   situation.  If things are getting better, it keeps going.  If not, it
   tumbles and then tries to swim again.

   This program demonstrates natural selection, and it also makes clear the
   falsity of the following argument.  Creationists often want to argue that
   the "probability" of forming a biological system is impossibly low.  For
   example, the E. coli genome is 4638858 basepairs long in GenBank entry
   U00096.  The probability of generating this sequence by equally probable
   choices for the 4 bases is

   1/(4^4638858) = 1/340282346638528859811704183484516925440 = 3.4x10^38

   (assuming that my computer really did the computation correctly!) This is
   then cited as evidence against evolution and for creationist views of how
   living things came into existance.

   The program here adds the missing factor to this argument, which is that
   there is selection at every generation.  This makes the probability
   computation given above invalid, BECAUSE THE PROBABILITIES ARE NOT
   INDEPENDENT.  That is, it is only valid to multiply probabilities if they
   are independent.  But selection at each generation means that the
   condition of THIS generation is not independent of what happened in the
   past.

examples

   Example migratep.1  BROWNIAN DRIFT

1.00     version of migrate that this parameter file is designed for.
27       range: +/- range of the world
40       displaymax: vertical display maximum
l        pagejump: 'l' means produce a page jump
100      maxgenss: maximum number of generations to run
10000    initpop: initial population placed at zero
0.5 0.5  survival rate: probability of survival for negative and positive steps
5        iseed: initial random seed
0        delaycount: number of times to do a futile addition to slow the display
25       touchhalt:  stop the program when this location is touched

At the first generation, the population looks like this:

maximum population density = 10000 is at location = 0
 ------------------                   ++++++++++++++++++
 222222221111111111--------- +++++++++111111111122222222
 7654321098765432109876543210123456789012345678901234567
|...........................*...........................|  40
|...........................*...........................|  39
|...........................*...........................|  38
|...........................*...........................|  37
|...........................*...........................|  36
|...........................*...........................|  35
|...........................*...........................|  34
|...........................*...........................|  33
|...........................*...........................|  32
|...........................*...........................|  31
|...........................*...........................|  30
|...........................*...........................|  29
|...........................*...........................|  28
|...........................*...........................|  27
|...........................*...........................|  26
|...........................*...........................|  25
|...........................*...........................|  24
|...........................*...........................|  23
|...........................*...........................|  22
|...........................*...........................|  21
|...........................*...........................|  20
|...........................*...........................|  19
|...........................*...........................|  18
|...........................*...........................|  17
|...........................*...........................|  16
|...........................*...........................|  15
|...........................*...........................|  14
|...........................*...........................|  13
|...........................*...........................|  12
|...........................*...........................|  11
|...........................*...........................|  10
|...........................*...........................|   9
|...........................*...........................|   8
|...........................*...........................|   7
|...........................*...........................|   6
|...........................*...........................|   5
|...........................*...........................|   4
|...........................*...........................|   3
|...........................*...........................|   2
|...........................*...........................|   1
|*******************************************************|   0
generation 1

This example reaches 25 from the origin in 44 steps.
The final population looks like:

maximum population density = 1264 is at location = -1
 ------------------                   ++++++++++++++++++
 222222221111111111--------- +++++++++111111111122222222
 7654321098765432109876543210123456789012345678901234567
|..........................*............................|  40
|..........................*.*..........................|  39
|..........................*.*..........................|  38
|..........................*.*..........................|  37
|..........................*.*..........................|  36
|........................*.*.*.*........................|  35
|........................*.*.*.*........................|  34
|........................*.*.*.*........................|  33
|........................*.*.*.*........................|  32
|........................*.*.*.*........................|  31
|........................*.*.*.*........................|  30
|......................*.*.*.*.*.*......................|  29
|......................*.*.*.*.*.*......................|  28
|......................*.*.*.*.*.*......................|  27
|......................*.*.*.*.*.*......................|  26
|......................*.*.*.*.*.*......................|  25
|......................*.*.*.*.*.*......................|  24
|....................*.*.*.*.*.*.*.*....................|  23
|....................*.*.*.*.*.*.*.*....................|  22
|....................*.*.*.*.*.*.*.*....................|  21
|....................*.*.*.*.*.*.*.*....................|  20
|....................*.*.*.*.*.*.*.*....................|  19
|....................*.*.*.*.*.*.*.*....................|  18
|....................*.*.*.*.*.*.*.*....................|  17
|....................*.*.*.*.*.*.*.*.*..................|  16
|..................*.*.*.*.*.*.*.*.*.*..................|  15
|..................*.*.*.*.*.*.*.*.*.*..................|  14
|..................*.*.*.*.*.*.*.*.*.*..................|  13
|..................*.*.*.*.*.*.*.*.*.*..................|  12
|..................*.*.*.*.*.*.*.*.*.*..................|  11
|..................*.*.*.*.*.*.*.*.*.*.*................|  10
|................*.*.*.*.*.*.*.*.*.*.*.*................|   9
|................*.*.*.*.*.*.*.*.*.*.*.*................|   8
|................*.*.*.*.*.*.*.*.*.*.*.*................|   7
|................*.*.*.*.*.*.*.*.*.*.*.*.*..............|   6
|..............*.*.*.*.*.*.*.*.*.*.*.*.*.*..............|   5
|..............*.*.*.*.*.*.*.*.*.*.*.*.*.*..............|   4
|..............*.*.*.*.*.*.*.*.*.*.*.*.*.*.*............|   3
|............*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*............|   2
|..........*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*..........|   1
|*******************************************************|   0
generation 44
the population touched 25
 program halt.

The population has spread out and reached 25 that way.  The average is still
zero.  If there were other factors keeping the population homogenous, the
variants would be trimmed away.

   Example migratep.2  SELECTION

1.00     version of migrate that this parameter file is designed for.
27       range: +/- range of the world
40       displaymax: vertical display maximum
l        pagejump: 'l' means produce a page jump
100      maxgenss: maximum number of generations to run
10000    initpop: initial population placed at zero
0.1 0.9  survival rate: probability of survival for negative and positive steps
5        iseed: initial random seed
0        delaycount: number of times to do a futile addition to slow the display
25       touchhalt:  stop the program when this location is touched

This example reaches 25 from the origin in 26 steps.
The final population looks like:

maximum population density = 2674 is at location = 21
 ------------------                   ++++++++++++++++++
 222222221111111111--------- +++++++++111111111122222222
 7654321098765432109876543210123456789012345678901234567
|................................................*......|  40
|................................................*......|  39
|................................................*......|  38
|................................................*......|  37
|................................................*......|  36
|................................................*......|  35
|................................................*......|  34
|..............................................*.*......|  33
|..............................................*.*......|  32
|..............................................*.*......|  31
|..............................................*.*.*....|  30
|..............................................*.*.*....|  29
|..............................................*.*.*....|  28
|..............................................*.*.*....|  27
|..............................................*.*.*....|  26
|..............................................*.*.*....|  25
|..............................................*.*.*....|  24
|..............................................*.*.*....|  23
|..............................................*.*.*....|  22
|............................................*.*.*.*....|  21
|............................................*.*.*.*....|  20
|............................................*.*.*.*....|  19
|............................................*.*.*.*....|  18
|............................................*.*.*.*....|  17
|............................................*.*.*.*....|  16
|............................................*.*.*.*....|  15
|............................................*.*.*.*....|  14
|............................................*.*.*.*....|  13
|............................................*.*.*.*....|  12
|............................................*.*.*.*....|  11
|............................................*.*.*.*.*..|  10
|..........................................*.*.*.*.*.*..|   9
|..........................................*.*.*.*.*.*..|   8
|..........................................*.*.*.*.*.*..|   7
|..........................................*.*.*.*.*.*..|   6
|..........................................*.*.*.*.*.*..|   5
|..........................................*.*.*.*.*.*..|   4
|........................................*.*.*.*.*.*.*..|   3
|........................................*.*.*.*.*.*.*..|   2
|......................................*.*.*.*.*.*.*.*..|   1
|*******************************************************|   0
generation 26
the population touched 25
 program halt.

The population reached 25 steps at almost one per generation.  More
importantly, the ENTIRE population migrated, and the average changed
significantly.

documentation

@article{Eisenbach1990,
author = "M. Eisenbach",
title = "Functions of the flagellar modes of rotation in bacterial
motility and chemotaxis",
journal = "Molec. Microb.",
volume = "4",
pages = "161-167",
year = "1990"}

  atchange program:
  https://alum.mit.edu/www/toms/atchange.html

  E. coli genome:
  ftp://ncbi.nlm.nih.gov/genbank/genomes/bacteria/ecoli.gbk

see also

  migratep
  migratep.1
  migratep.2

author
   Thomas Dana Schneider

bugs

technical notes

   The easiest way to run the program is under a Unix operating system.  Put
   the migratep file into your favorite editor.  Set up an atchange program
   running in a different window that will run the program whenever you write
   the migratep file out.  This allows one to play with the variables and
   watch the results without mousing around (which wastes an enor-mousE
   amount of time ;-).

   The populations are not stable.  They can grow or shrink exponentially
   depending the selection values.  The reason for the instability is that
   the population is not part of a balanced ecology.  This could be added to
   the model, but it would reduce the clarity of the results.  (Fortunately
   balanced ecologies obviously exist or the creationists would probably
   argue that this program demonstrates their impossibility!)

   There is a global constant (worldsize) that determines the maximum width
   of the world.

*)
(* end module describe.migrate *)

(* begin module migrate.const *)
  worldsize = 100; (* maximum range of the world *)
(* end module migrate.const *)

type
  (* pointer to a 'world' to allow them to be switched around easily: *)
  worldptr = ^world;
  world = record (* a world containing numbers of organisms
                    at different positions *)
     population: array[-worldsize..worldsize] of integer;
  end;

var
   migratep: text; (* parameter file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 2.38; (@ of malign.p 1996 April 19 *)

(* begin module package.numbar *)
(* ************************************************************************ *)
(* begin module numberdigit *)
function numberdigit(number, logplace:integer): char;
(* return the digit at the place value ('logplace') position of number.
example:
      numberdigit(13625, 3) = 3
      numberdigit(13625, 4) = 1
*)
var
      place: integer; (* the exponent of logplace *)
      count: integer; (* used to make place *)
      absolute: integer; (* the absolute value of number *)
      acharacter: char; (* the character to be returned *)
procedure digit;
(* extract a digit at the place position *)
var
      tenplace: integer; (* ten times place *)
      z: integer; (* an intermediate value *)
      d: integer; (* the digit extracted *)
begin (* digit *)
      tenplace:=10*place;
      z:=absolute-((absolute div tenplace)*tenplace);
      if place = 1
         then d:=z
         else d:= z div place;
      case d of
         0: acharacter:='0';
         1: acharacter:='1';
         2: acharacter:='2';
         3: acharacter:='3';
         4: acharacter:='4';
         5: acharacter:='5';
         6: acharacter:='6';
         7: acharacter:='7';
         8: acharacter:='8';
         9: acharacter:='9';
      end
end; (* digit *)
procedure sign;
(* put a negative sign out or a positive sign *)
begin (* sign *)
      if number <0 then acharacter:='-'
                   else acharacter:='+'
end; (* sign *)
begin (* numberdigit *)
      place:=1;
      for count:=1 to logplace do place:=10*place;

      if number=0 then begin
         if place=1 then acharacter:='0'
                    else acharacter:=' '
      end
      else begin
         absolute:=abs(number);
         if absolute < (place div 10)
            then acharacter:=' '
            else if absolute >= place
                    then digit
                    else sign
      end;
      numberdigit:=acharacter
end; (* numberdigit *)
(* end module numberdigit version = 4.19; (@ of prgmod.p 1996 Oct 16 *)

(* begin module numbersize *)
function numbersize(n: integer):integer;
(* calculate amount of space to be reserved for the integer n *)
const ln10 = 2.30259; (* natural log of 10 - for conversion to log base 10 *)
      epsilon = 0.00001; (* a small number to correct log base 10 errors *)
begin (* numbersize *)
      if n = 0
         then numbersize:=1
         else numbersize:=trunc(ln(abs(n))/ln10 + epsilon) + 2;
(* the epsilon assures that we do not lose a place due to
   roundoff.  eg, sometimes log base 10 of 10 would be
   0.9999 instead of 1, and we would not do it right...
   note: this will fail for very large numbers on the order of 1/epsilon. *)
(* the 2 is for the sign and last digit *)
end; (* numbersize *)
(* end module numbersize version = 4.19; (@ of prgmod.p 1996 Oct 16 *)

(* begin module numberbar *)
procedure numberbar(var afile: text; spaces, firstnumber, lastnumber: integer;
                    var linesused: integer);
(* write a bar of numbers to a file, with several spaces before.
   the number of lines used is returned *)
var
      logplace: integer; (* the log of the digit being looked at *)
      spacecount: integer; (* count of spaces *)
      number: integer; (* the current number being written *)
begin
      if abs(firstnumber) > abs(lastnumber)
         then linesused:= numbersize(firstnumber)
         else linesused:= numbersize(lastnumber);

      for logplace:=linesused-1 downto 0 do begin
         for spacecount:=1 to spaces do write(afile,' ');
         for number:=firstnumber to lastnumber
            do write(afile,numberdigit(number,logplace));
         writeln(afile)
      end
end;
(* end module numberbar version = 4.19; (@ of prgmod.p 1996 Oct 16 *)
(* ************************************************************************ *)
(* end module package.numbar version = 4.19; (@ of prgmod.p 1996 Oct 16 *)

procedure clearworld(var w: worldptr);
(* clear the world to 0. *)
var i: integer; (* index to the world w *)
begin
  for i := -worldsize to worldsize do w^.population[i] := 0;
end;

procedure startworld(var w: worldptr; initpop: integer);
(* set up and clear the world with initial population initpop at 0. *)
begin
  new(w);
  clearworld(w);
  w^.population[0] := initpop;
end;

procedure switchworlds(var a, b: worldptr);
(* switch the worlds a and b *)
var
   h: worldptr; (* holder for reversing world arrays *)
begin
  h := a;
  a := b;
  b := h;
end;

procedure showworld(var f: text; var w: worldptr;
                   worldrange: integer;
                   displaymax: integer;
                   pagejump: char;
                   generation: integer;
                   touchhalt: integer);
(* set up and clear the world with initial population initpop at 0. *)
var
   i: integer; (* index to the world w *)
   mpd: integer; (* maximum population density *)
   location: integer; (* location of mpd *)
   mindex: integer; (* index to mpd *)
   linesused: integer; (* lines use for numbar *)
   touched: boolean; (* has any part of the population touched touchhalt? *)
begin
  with w^ do begin
    mpd := 0;
    touched := false;
    for i := -worldsize to worldsize do begin
      if mpd < population[i] then begin
          mpd := population[i];
          location := i;
      end;
      if i = touchhalt
      then if population[i] > 0
      then touched := true
   end;

    if mpd = 0 then begin
       writeln(output,'population died out');
       halt
    end;

    if pagejump = 'l'
    then begin page(f); writeln(f) end
    else writeln(f);
    writeln(output,'maximum population density = ',mpd:1,
                   ' is at location = ',location:1);
    numberbar(f, 1, -worldrange, +worldrange, linesused);

    for mindex := displaymax downto 0 do begin
       write(f,'|');
       for i := -worldrange to worldrange do begin
         if (population[i]/mpd) < (mindex/displaymax)
         then write(f,'.')
         else write(f,'*');
       end;
       writeln(f,'| ', mindex:3);
    end;
  end;
  writeln(f,'generation ',generation:1);

  if touched then begin
     writeln(f,'the population touched ',touchhalt:1);
     halt
  end;
end;

(* begin module shift.random.31 *)
procedure random(var rand: real; var iseed: integer);
(* This random number generator is based on a shift register with a single bit
of feedback, as described in Electronics for Neurobiologists, by Brown, Maxfield
and Moraff, MIT press 1973, referencing Random Process Simulation and
Measurement by Korn, McGraw-Hill 1966.  A 31-bit shift register is maintained in
the integer iseed, which should initially be given any positive value.  This
31-bit number is shifted right one bit and the output of the last (31st) bit and
the 13th bit are added modulo 2 (exclusive orred) and fed back into the new
first bit.  This is done between 4 and 11 times, depending on the last 3 bits of
the original number.  The result is converted to a real number between 0 and 1
and returned as the value of rand.  The 31-bit shift register goes through all
2**31-1 values before repeating; the repetition frequency of this algorithm
could be less or greater depending on the seed, because of the random number of
multiple shifts per call. *)
const pow18=262144; pow19=524288; pow30=1073741824; (* powers of 2 *)
      pow31m=2147483647.0;  (* 2**31-1 *)
var i, nrep: integer;  (* index, number of times to do shift *)
begin (* random *)  
      nrep := 4 + iseed mod 8;  (* do it 4 to 11 times based on last 3 bits *)
      for i:= 1 to nrep do
         (* if last bit and 9th bit are equal, feed back a 0, otherwise a 1 *)
      if( odd(iseed) = ((iseed mod pow19) >= pow18))
         then iseed := iseed div 2
         else iseed := (iseed div 2) + pow30;
      rand := iseed/pow31m;
end; (* random *) 
(* end module shift.random.31 version = 2.38; (@ of malign.p 1996 April 19 *)

function chose(survival: real; var iseed: integer): integer;
(* given the survival probability, chose whether
the organism lives (add 1) or dies (add nothing = 0).
Carry the random number generator in rand and iseed. *)
var
  rand: real; (* the current random number corresponding to the new iseed *)
begin
  random(rand,iseed);
  if rand < survival 
  then chose := 1
  else chose := 0;
end;

procedure stepselect(var a, b: worldptr; nsurvival, psurvival: real;
                     var iseed: integer);
(* replicate every organism in world a, steping to the left or
right and putting the new organisms in world b.
Select against those that move negatively by probability p.
Carry the random number generator in rand and iseed. *)
var
   i: integer; (* index to the world w *)
   c: integer; (* index to the creatures at i *)
begin
   clearworld(b);
   for i := -worldsize+1 to worldsize-1 do begin
      for c := 1 to a^.population[i] do begin
          b^.population[i-1] := chose(nsurvival,iseed) + b^.population[i-1];
          b^.population[i+1] := chose(psurvival,iseed) + b^.population[i+1];
      end;
   end;
end;

procedure sleep(delaycount: integer);
(* sleep for a while *)
var j, k: integer;
begin
   for j := 1 to delaycount
   do k := succ(k);
end;

(* begin module migrate.themain *)
procedure themain(var migratep: text);
(* the main procedure of the program *)
var
   c: worldptr; (* current world *)
   delaycount: integer;  (* number of times to do a futile addition to slow
      the display *)
   displaymax: integer;  (* vertical display maximum *)
   initpop: integer; (* initial population placed at zero *)
   iseed: integer; (* current random number *)
   p: worldptr; (* previous world *)
   parameterversion: real; (* parameter version number *)
   generation: integer; (* index for maxgens *)
   maxgens: integer; (* number of generations to run *)
   pagejump: char; (* l means produce a page jump *)
   psurvival, nsurvival: real; (* probability of selection
                              for negative and positive steps *)
   touchhalt: integer; (* stop the program when this location is touched *)
   worldrange: integer; (* range to use in the world *)

procedure readparameters;
(* read the parameters *)
begin
   reset(migratep);
   readln(migratep, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(migratep, worldrange);
   readln(migratep, displaymax);
   readln(migratep, pagejump);
   readln(migratep, maxgens);
   readln(migratep, initpop);
   readln(migratep, nsurvival, psurvival);
   readln(migratep, iseed);
   readln(migratep, delaycount);
   readln(migratep, touchhalt);

   if worldrange <= 0 then begin
      writeln(output,'worldrange must be positive');
      halt
   end;

   if worldrange >= worldsize then begin
      writeln(output,'worldrange must be less than worldsize =',worldsize:1);
      halt
   end;

   if displaymax <= 0 then begin
      writeln(output,'displaymax must be positive');
      halt
   end;

   if maxgens <= 0 then begin
      writeln(output,'maxgens must be positive');
      halt
   end;

   if initpop <= 0 then begin
      writeln(output,'initpop must be positive');
      halt
   end;

   if (nsurvival < 0.0) or (nsurvival > 1.0) or
      (psurvival < 0.0) or (psurvival > 1.0) then begin
      writeln(output,'selection must be between 0 and 1');
      halt
   end;

   if iseed < 1 then begin
      writeln(output,'iseed must be positive');
      halt
   end;

   (* delaycount can be any value *)

end;

begin
   writeln(output,'migrate ',version:4:2);
   readparameters;
   startworld(p, 0);
   startworld(c, initpop);

   for generation := 1 to maxgens do begin
      showworld(output, c, worldrange, displaymax, pagejump,
                        generation, touchhalt);
      switchworlds(c, p);
      stepselect(p, c, nsurvival, psurvival, iseed);
      sleep(delaycount);
   end;

   writeln(output, 'done');
end;
(* end module migrate.themain *)

begin
   themain(migratep);
1: end.
