program mnomial(mnomialp, list, output);
(* produce the multinomial distribution for base probabilities
   Thomas Dana Schneider *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.19; (* of mnomial, 1994 sep 54
origin 1988 feb 12 from binplo *)
(* end module version *)

(* begin module describe.mnomial *)
(*
name
   mnomial: produce the multinomial distribution for base probabilities

synopsis
   mnomial(mnomialp: in, list: out, output: out)

files
   mnomialp: parameters to control the program, as pairs of lines
      first line: na,nc,ng,nt
      second line: pa,pc,pg,pt
        this may repeat.
      When ng=nt=0 and pg=pt=0, the binomial distribution is calculated.
      The mean (n*pa) and standard deviation (sqrt(n*pa*pc)) are given.
   list: results
   output: messages to the user

description
   This program calculates the multinomial distribution:
                                 (na+nc+ng+nt)!   na   nc   ng   nt
    p(na,nc,ng,nt|pa,pc,pg,pt) = -------------- pa   pc   pg   pt
                                 na!nc!ng!nt!

see also
   binplo.p

author
   Thomas Dana Schneider

bugs
   none known

*)
(* end module describe.mnomial *)

(* begin module mnomial.const *)
   tolerance = 0.001; (* sum of p's must be 1.0 to this tolerance *)
   maxtable = 100; (* largest number that the program can handle *)
   wid = 10; (* number of digits to print *)
   dec = 8 ; (* number of decimal places to print *)
(* end module mnomial.const *)

type
(* begin module mnomial.type *)
   lnfactarray = array[0..maxtable] of real; (* precalculated logs *)
(* end module mnomial.type *)

var
   mnomialp, (* control parameters for mnomial *)
   list: (* results of the program *)
      text;

(* begin module halt *) 
procedure halt; 
(* stop the program.  the procedure performs a goto to the end of the   
   program.  you must have a label: 
      label 1;
   declared, and also the end of the program must have this label:  
      1: end. 
   examples are in the module libraries.
   this is the only goto in the delila system. *) 
begin 
      writeln(output,' program halt.'); 
      goto 1
end;  
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module pmultinomial *)
function pmultinomial(var n, na,nc,ng,nt: integer;
                      var p, pa,pc,pg,pt: real;
                          lnfact: lnfactarray): real;
(* calculate the probability of obtaining the set of nucleotites
na,nc,ng,nt given the probabilities pa,pc,pg,pt.
Return the sums n and p as side effects.

This is a multinomial distribution:
                                 (na+nc+ng+nt)!   na   nc   ng   nt
    p(na,nc,ng,nt|pa,pc,pg,pt) = -------------- pa   pc   pg   pt
                                 na!nc!ng!nt!
This is calulated by taking the log of the whole thing, figuring
out the sum of the parts and exponentiating. *)
function lnpower(p: real; n: integer): real;
(* calculate p raised to the n,
give the result as the log of the power *)
begin (* lnpower *)
   if p > 0
   then lnpower := n * ln(p)
   else lnpower := 0.0
end; (* lnpower *)

begin
   p := pa + pc + pg + pt;
   if (p < 1.0-tolerance) or (p > 1.0+tolerance) then begin
      writeln(output,'p is not 1.0 within the tolerance');
      halt
   end;

   n := na + nc + ng + nt;
   if n > maxtable then begin
      writeln(output, ' n too big for program, increase maxtable');
      halt
   end;

   pmultinomial := exp(  lnfact[n]
                    - lnfact[na]
                    - lnfact[nc]
                    - lnfact[ng]
                    - lnfact[nt]
                    + lnpower(pa,na)
                    + lnpower(pc,nc)
                    + lnpower(pg,ng)
                    + lnpower(pt,nt) )
end;
(* end module pmultinomial *)

(* begin module makelnfact *)
procedure makelnfact(var lnfact: lnfactarray; n: integer);
(* make fill the global table of the logs of i! for i=0 to n *)
var
   i: integer; (* index for making the table *)
begin
   lnfact[0] := 0;
   for i := 1 to n do begin
      lnfact[i] := lnfact[i-1] + ln(i)
   end
end; (* makelnfact *)
(* begin module makelnfact *)


(* begin module mnomial.themain *)
procedure themain(var mnomialp, list: text);
(* the main procedure of the mnomial program *)
var
   lnfact: lnfactarray; (* precalculated logs *)
   n: integer; (* sum of na,nc,ng,nt *)
   na,nc,ng,nt: integer; (* number of bases in an experiment *)
   p:    real; (* sum of pa,pc,pg,pt *)
   pa,pc,pg,pt: real; (* probabilities of the bases *)
   pmul: real; (* the multinomial probability *)
   sum: real; (* sum of all the pmuls *)
begin
   writeln(output,'mnomial ',version:4:2);
   rewrite(list);
   writeln(list,'mnomial ',version:4:2);

   makelnfact(lnfact,maxtable);

   sum := 0.0;
   reset(mnomialp);
   while not eof(mnomialp) do begin
      readln(mnomialp,na,nc,ng,nt);
      readln(mnomialp,pa,pc,pg,pt);

      pmul := pmultinomial(n, na,nc,ng,nt, p, pa,pc,pg,pt, lnfact);
      sum := sum + pmul;

      writeln(list,'-----');
      if (ng=0) and (nt=0) then begin
         writeln(list,' Binomial: ');
         writeln(list,'    n  = ',n:wid);
         writeln(list,'    pa = ',pa:wid:dec);
         writeln(list,'    mean = n*pa = ',n*pa:wid:dec);
         writeln(list,'    sigma = sqrt(n*pa*(1-pa)) = ',
                                   sqrt(n*pa*(1-pa)):wid:dec);
         writeln(list);
      end;
      writeln(list,
                   ' na= ',na:wid,
                   ' nc= ',nc:wid,
                   ' ng= ',ng:wid,
                   ' nt= ',nt:wid);
      writeln(list,
                   ' pa= ',pa:wid:dec,
                   ' pc= ',pc:wid:dec,
                   ' pg= ',pg:wid:dec,
                   ' pt= ',pt:wid:dec);
      writeln(list);
      writeln(list, ' p(na,nc,ng,nt|pa,pc,pg,pt)=',pmul: wid: dec);
      writeln(list, '              running sum = ',sum: wid: dec);
   end
end;
(* end module mnomial.themain *)

begin (* mnomial *)
   themain(mnomialp, list);
1: end.
