program modlen(fin, fout, modlenp, output);
(* modlen: determine module lengths
   by tom schneider
   module libraries required: delman, prgmods *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.37; (* of modlen 94 sep 5 *)
(* end module version *)

(* begin module describe.modlen *)
(*
name
      modlen: determine module lengths

synopsis
      modlen(fin: in, fout: out, modlenp: in, output: out)

files
      fin: a text file containing modules
      fout: a list of the module names and their lengths
      modlenp: parameters to control modlen.
         if the file is empty, modlen gives a complete list of
         modules and their lengths to fout, and notes those
         longer than 'a certain number of lines' lines to output.
         otherwise the file is expected to contain two integers,
         each at the start of a line.  these are the shortest and
         longest lengths to print to output.
      output: messages to the user and modules with lengths
         determined by modlenp.

description
      the delila manual consists of module pages.  this tool allows
      you to find out if the pages will fit onto your printer.

see also
      module.p, show.p, break.p

author
      thomas d. schneider

bugs
      none known

technical notes
      'a certain number of lines' is set by the constant defshort.
*)
(* end module describe.modlen *)

(* begin module modlen.const *)
      (* constants that determine the lengths of modules
      to be announced to the output file *)
      defshort = 62; (* default for shortest length module *)
      deflong = maxint; (* default for longest length module *)
(* end module modlen.const version = 1.10; (@ of modlen 85 mar 18 *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module modlen.filler.const *)
      fillermax = 16; (* the size of the filler array for a string *)
(* end module modlen.filler.const version = 1.10; (@ of modlen 85 mar 18 *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 'prgmod 3.96  85 mar 18 tds'; *)

var
(* begin module modlen.var *)
      fin, fout, modlenp: text;
(* end module modlen.var version = 1.10; (@ of modlen 85 mar 18 *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 that a number is *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.token *)
procedure token(var buffer, atoken: string; var gotten: boolean);
(* get a token from the buffer *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
      rea: real;
      index: integer; (* to the buffer *)
begin
      figurestring(buffer,first,last,what,cha,int,rea);

      if what = ' '
      then gotten := false
      else begin
         clearstring(atoken);
         for index := first to last
            do atoken.letters[index-first+1] := buffer.letters[index];
         atoken.length := last - first + 1;
         atoken.current := 1;

         gotten:=true
      end
end;
(* end module interact.token version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
(*         if debugging then begin
            writestring(list,seek);
            writeln(list,'testfortrigger seek.letters[',state:1,']:',
                           seek.letters[state],' ch:',ch);
         end;*)
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin (* reset trigger *)
            state := 0;
            skip := true;
            found := false
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module modlen.themain *)
procedure themain(var fin, fout, modlenp: text);
(* the main procedure of the program *)
var
      bmo, emo: trigger; (* the begin and end module strings *)

      short, long: integer; (* the shortest and longest page lengths
         to report to the output *)

      aline, (* a line of text from fin *)
      aname: (* a name of a module *)
         string;
      bmoline: integer; (* line number where current module begins *)
      position: integer; (* the position in a string *)
      length: integer; (* the length of the current module *)
      line: integer; (* the line number of aline *)
      gotten: boolean; (* true when a trigger is found *)
begin
      writeln(output,'modlen ',version:4:2);
      reset(fin);
      reset(modlenp);
      rewrite(fout);
      writeln(fout,'* modlen ',version:4:2,' module lengths');
      writeln(fout,'* length begin   end  name');

      if eof(modlenp)
      then begin
         short := defshort;
         long  := deflong
      end
      else begin
         readln(modlenp,short);
         if eof(modlenp) then begin
            writeln(output,' missing second parameter');
            halt
         end;
         readln(modlenp,long)
      end;

      filltrigger(bmo,'(* begin module ');
      resettrigger(bmo);

      writeln(output,'lengths reported: ',short:1,' to ',long:1);

      line := 0;
      while not eof(fin) do begin
         getstring(fin,aline,gotten);
         line := line + 1;

         with aline
         do while (current<length) and
                  (not bmo.skip) and
                  (not bmo.found)
            do begin
               testfortrigger(letters[current],bmo);
               current := current + 1
            end;

         if bmo.found then begin
            bmoline := line;
            filltrigger(emo,'(* end module   ');
            token(aline,aname,gotten);

            (* fill the emo trigger with the name *)
            for position := 1 to aname.length
            do emo.seek.letters[14+position]
               := aname.letters[position];
            (* set emo.seek to the new length, do not forget the spaces
               before and after the name *)
            emo.seek.length := emo.seek.length + aname.length + 2;

            (* seek the end of the module *)
            resettrigger(emo);
            while (not eof(fin)) and (not emo.found) do begin
               getstring(fin,aline,gotten);
               line := line + 1;
               resettrigger(emo);

               with aline
               do while (current<length) and
                  (not emo.skip) and
                  (not emo.found)
                  do begin
                     testfortrigger(letters[current],emo);
                     current := current + 1
                  end;
            end;

            if eof(fin) and (not emo.found) then begin
               write(output, ' module ');
               writestring(output,aname);
               writeln(output,' has no end');
               halt
            end;

            length := line-bmoline+1;
            write(fout,' ',length:7, (* length of module *)
                       ' ',bmoline:5, (* begin line of module *)
                       ' ',line:5,'  '); (* end line of module *)
            writestring(fout,aname);
            writeln(fout);

            if (short <= length) and (length <= long) then begin
               write(output, length:1,' ');
               writestring(output,aname);
               write(output,' from ',bmoline:1,
                            ' to ',line:1);
               writeln(output)
            end;
         end;
         resettrigger(bmo); (* prepare for next round *)
      end
end;
(* end module modlen.themain version = 1.10; (@ of modlen 85 mar 18 *)

begin
      themain(fin,fout,modlenp);
1: end.
