program oligosetc(sequ, oligosetcp, order, output);
(* oligosetc: raw sequence to Oligo's Etc format

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.03; (* of oligosetc.p 2000 Oct 2
2000 Oct  2, 1.03: allow fasta format
1997 Dec 19, 1.03: previous version
origin 1997 May 8 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.oligosetc *)
(*
name
   oligosetc: raw sequence to Oligo's Etc format

synopsis
   oligosetc(sequ: in, oligosetcp: in, order: out, output: out)

files
   sequ:  raw sequences with a, c, g, t and ending in "."

      If the first character of the first line of the file is
      a ">", then fasta format is assumed.  In this format
      the line following the ">" is the name of the next sequence.
      Sequences do not end in periods but rather with another ">"
      at the start of the line.
      The names will be printed to the order file.

   order:  multiple line detailed description of file 2, etc

   oligosetcp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      modification (character):
         "5" means put an x on the 5' end to indicate 5' biotin
         "3" means put an x on the 3' end to indicate 3' biotin
         "8" means put an x on both the ends

      maxlinelength (integer):  maximum length of output lines.

   output: messages to the user

description
   Convert simple DNA sequence to the format that the Oligo's Etc company
   likes it in.  It is broken into sets of 3 bases with spaces between.  All
   bases are capitalized except g.  Biotins are indicated by X on the end.

examples

   sequ:

g gaattc aagctt
actcaaaaactgaacaac
cgggtcag
aagctt
g
c
gcgaagc
g
c
aagctt
ctgacccg
gttgttcagtttttgagt
aagctt gaattc c.

    gives:

oligo number 1

 5' ggA ATT CAA gCT TAC TCA AAA ACT gAA CAA CCg
ggT CAg AAg CTT gCg CgA AgC gCA AgC TTC TgA CCC
ggT TgT TCA gTT TTT gAg TAA gCT TgA ATT CC 3'

101 bases long

documentation

see also
   makebk.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.oligosetc *)

(* begin module datetime.const *)
   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789 
         1         2
*)
(* end module datetime.const version = 7.40; {of delmod.p 2000 Feb 18} *)

type

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.40; {of delmod.p 2000 Feb 18} *)

var
   sequ, (* file used by this program *)
   oligosetcp, (* file used by this program *)
   order: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline prgmod (@ of prgmod.p 2000 Sep 2 *)

(* begin module getdatetime *)
procedure getdatetime(var adatetime: datetimearray);
(* get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.
*)
begin
   date(adatetime,'%Y/%m/%d %H:%M:%S');
end;
(* end module getdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module writedatetime *)
procedure writedatetime(var thefile: text; adatetime: datetimearray);
(* expand the date and time out and print in the file *)
var
      index: integer; (* index of datetime *)
begin
      for index:=1 to datetimearraylength do write(thefile,adatetime[index])
end;
(* end module writedatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module capitalize *)
function capitalize(c: char): char; 
(* convert the character c to upper case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 

   if (n >= ord('a')) and 
      (n <= ord('z'))   
   then c := chr( n - ord('a') + ord('A'));

   capitalize := c
end; 
(* end module capitalize version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module decapitalize *)
function decapitalize(c: char): char; 
(* convert the character c to lower case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 

   if (n >= ord('A')) and 
      (n <= ord('Z'))   
   then c := chr( n - ord('A') + ord('a'))
   else c := chr(n);

   decapitalize := c
end; 
(* end module decapitalize version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module oligosetc.startoligo *)
procedure startoligo(var f: text; biotins: char;
                     var linelength, sequlength: integer);
(* Start an oligo output.  Put a biotin on if needed.  Increase
linelength *)
begin
   linelength := 1; (* blank in front of 5' *)
   sequlength := 0;
   write(f,' 5'' ');
   if (biotins = '5') or
      (biotins = '8')
   then begin
      write(f, 'X');
      linelength := linelength + 1;
      sequlength := sequlength + 1
   end;
   linelength := linelength + 3
end;
(* end module oligosetc.startoligo *)

(* begin module oligosetc.endoligo *)
procedure endoligo(var f: text; biotins: char;
                   var linelength, sequlength: integer);
(* end an oligo output.  Put a biotin on if needed.  Increase linelength *)
begin
   if (biotins = '3') or
      (biotins = '8')
   then begin
      write(f, 'X');
      linelength := linelength + 1;
      sequlength := sequlength + 1
   end;

   writeln(f,' 3''');
   writeln(f);

   writeln(f, sequlength:1, ' bases long');
   linelength := linelength + 3;

   if (biotins = '5') or
      (biotins = '8')
   then write(f, '5'' biotin');

   if (biotins = '8')
   then write(f, ' and ');

   if (biotins = '3') or
      (biotins = '8')
   then write(f, '3'' biotin');

   if (biotins = '3') or
      (biotins = '5') or
      (biotins = '8')
   then write(f, ' indicated by "X"');

   writeln(f);
end;
(* end module oligosetc.endoligo *)

(* begin module oligosetc.themain *)
procedure themain(var sequ, oligosetcp, order: text);
(* the main procedure of the program *)
var
   c: char; (* a character in sequ *)
   biotins: char; (* 5 or 3 or 8 for both ends biotinylated *)
   fasta: boolean; (* if true then we are reading fasta format *)
   linelength: integer; (* current length of output line *)
   maxlinelength: integer; (* maximum of length of output lines *)
   number: integer; (* number of oligo *)
   outofoligo: boolean; (* true when outside of oligo *)
   parameterversion: real; (* parameter version number *)
   sequlength: integer; (* current length of sequence *)
   thedate: datetimearray; (* current date and time *)
begin
   writeln(output,'oligosetc ',version:4:2);
   reset(oligosetcp);
   readln(oligosetcp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(oligosetcp, biotins);
   readln(oligosetcp, maxlinelength);

   reset(sequ);
   fasta := false;
   if not eof(sequ)
   then if not eoln(sequ)
        then if sequ^ = '>'
             then begin
                fasta := true;
                writeln(output,'using fasta format');
             end;

   rewrite(order);

   writeln(order,'oligosetc ',version:4:2);
   getdatetime(thedate);
   writedatetime(order,thedate);
   writeln(order);
   writeln(order);
   writeln(order,'Order for Oligos Etc.');
   writeln(order);

   outofoligo := true;
   number := 0;
   while not eof(sequ) do begin
      if eoln(sequ)
      then begin
         readln(sequ);
         if (not eof(sequ)) and fasta then begin
            if sequ^ = '>' then begin
               endoligo(order,biotins,linelength, sequlength);
               outofoligo := true;
            end;
         end;
      end
      else begin

         if outofoligo then begin
            number := succ(number);
            for linelength := 0 to maxlinelength do write(order,'-');
            writeln(order);
            writeln(order,'oligo number ',number:1);
            writeln(order);
    if fasta then begin (* write the name out *)
    copyaline(sequ, order);
{zzz}
    end;

            startoligo(order,biotins,linelength,sequlength);
            outofoligo := false;
         end;

         read(sequ, c);

         if c in ['A','C','G','T']
         then c := decapitalize(c);

         if c in ['a','c','g','t']
         then begin
            if c in ['a','c','t']
            then c := capitalize(c);

            sequlength := succ(sequlength);
{
write(order,sequlength:1);
}

            write(order, c);
        
            linelength := succ(linelength);
            if linelength > maxlinelength then begin
               writeln(order);
               linelength := 0;
            end
            else begin
               if (sequlength mod 3) = 0 then begin
                  write(order,' ');
                  linelength := succ(linelength);
               end;
            end
         end
         else if c = '.'
              then begin
                 endoligo(order,biotins,linelength, sequlength);
                 outofoligo := true;
              end
              else if c <> ' ' then begin
                       writeln(order,'WARNING: character "',c,'" ignored!');
                   end
      end
   end;
   if fasta then begin
      endoligo(order,biotins,linelength, sequlength);
   end
   else if (c <> '.') then begin
     if linelength <> 0 then writeln(order);
     writeln(output, 'sequence did not end with "."');
   end;

end;
(* end module oligosetc.themain *)

begin
   themain(sequ, oligosetcp, order);
1: end.
































