program patana(pattern,anal,output);
(* analysis of a pattern matrix
   by gary d stormo (modified by tom schneider),  copyright 1986
module libraries needed: delman, delmods, auxmods *)

label 1;  (* end of program *)

const
(* begin module version *)
version = 2.19; (* of patana 1994 sep 5
origin august 14, 1982 *)
(* end module version *)

(* begin module describe.patana *)
(*
name
      patana: pattern analysis

synopsis
      patana(pattern: in, anal: out, output: out)

files
      pattern: a pattern matrix, the output of the pattern learning
         program patlrn;
      anal: the analysis of the pattern matrix;
      output: for messages to the user;

description
      patana does some simple analyses of a pattern matrix.  for each
      position (i.e., row) of the matrix it calculates the:
         sum;
         average;
         variance;
         maximum;
         minimum.
      it also calculates the sum of each of those measures.

      the sum of sums is used in other calculations.  if all training
      sequences are the same length, this is the difference between
      the number of + class sequences and - class sequences added
      together to make the w matrix.

      the sum of the average is an estimate of the mean response to
      random sequences

      the sum of the variance is a variance that estimates the
      spread of responses to random sequences.  take the
      square root to obtain the standard deviation.

      the sum of the maxima is the largest response possible.

      the sum of the minima is the smallest response possible.

see also
      patlrn.p

author
      gary d. stormo (modified by tom schneider)

bugs
      none known
*)
(* end module describe.patana *)

(* begin module matrix.const *)
      maxmatrix = 120; (* the maximum size of a pattern matrix *)
(* end module matrix.const version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

type
      base = (a,c,g,t);  (* this is declared so we do not need book.type *)
(* begin module matrix.type *)
      matrix = array[base,1..maxmatrix] of integer; (* a pattern matrix *)
(* end module matrix.type version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

var   anal,   (* output analysis *)
      pattern : text;  (* input pattern matrix *)
      wmatrix : matrix; (* where the pattern is stored *)
      i,
      width, (* actual width of the matrix *)
      beginning, (* of the matrix *)
      sumsum, (* sum of the position sums *)
      maxsum, (* sum of the position maxima *)
      minsum : integer; (* sum of the position minima *)
      varsum,  (* sum of the position variances *)
      avesum : real; (* sum of the position averages *)
      funcmin, nfuncmax: integer; (* functional range of the pattern *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module matrix.readmatrix *)
procedure findlearnend(var pattern: text);
(* locate the end of the cyclic learning part of the pattern file *)
const
      learnmax = 16; (* the size of learnend *)
var
      (* this array contains the pattern that defines the end *)
      learnend: packed array[1..learnmax] of char;
      state: integer; (* how close we are to finding learnend *)
begin (* findlearnend *)
      reset(pattern);
      learnend := 'end of learning.';
      state := 1;
      while not eof(pattern) and (state < learnmax) do begin
         if learnend[state] = pattern^
         then state := succ(state)
         else state :=1;
         if eoln(pattern) then readln(pattern)
                          else get(pattern)
      end;
      if eof(pattern) then begin
         write  (output,'pattern matrix does not contain "');
         writeln(output,learnend,'" signal');
         halt
      end
end; (* findlearnend *)
procedure getcolon(var f: text);
(* move to the next colon in file f *)
begin (* getcolon *)
      while f^<>':' do begin
         get(f);
         if eof(f) then begin
            writeln(output,'pattern is missing colons');
            halt
         end
      end;
      get(f); (* move past the colon *)
end; (* getcolon *)

procedure readmatrix(var thefile: text; var wmatrix: matrix;
                     var beginning, width: integer);
(* read a pattern matrix wmatrix from the file thefile, returning
the aligned position for the first position (beginning) and the width *)
var   i : integer;  (* an index *)
      ba : base;  (* an index *)
begin (* readmatrix *)
      findlearnend(thefile);

      (* skip min and max values *)
      getcolon(thefile); getcolon(thefile);

      getcolon(thefile); (* skip to range *)
      read(thefile,beginning); (* read first position of the matrix *)

      getcolon(thefile); (* skip to width *)
      read(thefile,width); (* read width of the matrix *)

      if (width > maxmatrix) then begin
         writeln(output,'input matrix too large');
         halt;
      end;

      for i := 1 to width do   (* read in matrix *)
      begin
         getcolon(thefile); (* skip position label *)
         for ba := a to t do read(thefile,wmatrix[ba,i]);
         readln(thefile);
      end;
            (* set all the unused elements of wmatrix to 0 *)
      for i := (width + 1) to maxmatrix do
         for ba := a to t do wmatrix[ba,i] := 0;
end; (* readmatrix *)
(* end module matrix.readmatrix version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module matrix.readthresholds *)
procedure readthresholds(var pattern: text; var funcmin, nfuncmax: integer);
(* find the funcmin and nfuncmax of the pattern *)
begin (* readthresholds *)
      reset(pattern);
      findlearnend(pattern);

      getcolon(pattern);
      readln(pattern,funcmin);
      getcolon(pattern);
      readln(pattern,nfuncmax)
end; (* readthresholds *)
(* end module matrix.readthresholds version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure positsum(i:integer);
(* this calculates for each matrix position the sum, average and
   variance of the elements. the variance is calculated as:
      vari = (sum(elements)**2)/4 - ave**2   *)
var  ba : base;
    vari,  (* the variance of the elements of a matrix position *)
    ave : real;  (* the average of the elements of a matrix position *)
    sum : integer;  (* the sum of the elements of a matrix position *)
begin
      vari := 0;
      sum := 0;
      ave := 0;
      for ba := a to t do begin
         sum := sum + wmatrix[ba,i];
         vari := vari + (wmatrix[ba,i] * wmatrix[ba,i]);
      end;
      ave := sum/4;
      vari := vari/4 - (ave * ave);
      sumsum := sumsum + sum;
      varsum := varsum + vari;
      avesum := avesum + ave;
      write (anal,sum:8,ave:8:2,vari:10:2);
end;

procedure positmax(i:integer);
(* this calculates the maximum element of each position of the matrix *)
var   max : integer;
begin
      max := wmatrix[a,i];
      if max < wmatrix[c,i] then max := wmatrix[c,i];
      if max < wmatrix[g,i] then max := wmatrix[g,i];
      if max < wmatrix[t,i] then max := wmatrix[t,i];
      maxsum := maxsum + max;
      write (anal,max:6);
end;

procedure positmin(i:integer);
(* this calculates the minimum element of each position of the matrix *)
var   min : integer;
begin
      min := wmatrix[a,i];
      if min > wmatrix[c,i] then min := wmatrix[c,i];
      if min > wmatrix[g,i] then min := wmatrix[g,i];
      if min > wmatrix[t,i] then min := wmatrix[t,i];
      minsum := minsum + min;
      writeln (anal,min:6);
end;

procedure header;
var ch : char;
begin
      writeln(anal,' patana version ',version:4:2,
                   '; analysis of pattern matrix of ');
      readln(pattern);  (* skip first line of pattern file *)
      copyaline(pattern,anal);
      writeln(anal);
end;

procedure analyze;
(* this procedure calls each of the calculating procedures and prints
   out both the matrix being analyzed and the results of the analysis *)
var ba: base;
    i: integer; (* index *)
begin (* analyze *)
      write(anal,' minimum value for    functional sequences: ');
      writeln(anal,funcmin:4);
      write(anal,' maximum value for nonfunctional sequences: ');
      writeln(anal,nfuncmax:4);
      writeln(anal);
      writeln(anal,' range: ',beginning:1,' to ',
            (beginning + width -1):1,'  width: ',width:1);
      writeln(anal);

            (* label the columns *)
      write(anal,'  pos. ','a':5,'c':5,'g':5,'t':5);
      writeln(anal,'sum':10,'ave':8,'var':10,'max':6,'min':6);
            (* make the upper matrix-border *)
      write(anal,' ':6);
      for i := 1 to 23 do write(anal,'-');
      writeln(anal);
            (* write the matrix positions and elements *)
      for i := 1 to width do begin
         write(anal,' ',(beginning + i - 1):4,' :');
         for ba := a to t do write(anal,' ',wmatrix[ba,i]:4);
         write(anal,' :');
            (* calculate and write the analysis *)
         positsum(i);
         positmax(i);
         positmin(i);
      end;
            (* make the lower matrix-border *)
      write(anal,' ':6);
      for i := 1 to 23 do write(anal,'-');
      writeln(anal,'    ----  ------    ------  ----  ----');
      for i := 1 to 29 do write(anal,' ');
      writeln(anal,sumsum:8,avesum:8:2,varsum:10:2,maxsum:6,minsum:6);
end;

procedure initialize;
begin
      writeln(output,'patana ',version:4:2);
      reset(pattern);
      if eof(pattern) then begin
         writeln(output,'no pattern file');
         halt;
      end;
      rewrite(anal);
      maxsum := 0;
      minsum := 0;
      avesum := 0;
      sumsum := 0;
      varsum := 0;
end;

begin (* patana *)
      initialize;
      header;
      readmatrix(pattern,wmatrix,beginning,width);
      readthresholds(pattern, funcmin, nfuncmax);
      analyze;
1: end. (* patana *)
