program piechart(piechartp, pienumbers, pieps, output);
(* piechart: create a pie chart

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.07; (* of piechart.p 2011 Feb 14
2011 Feb 14, 1.07: H control
2010 Dec 06, 1.06: document font control
2010 Dec 05, 1.05: page size control!
2010 Dec 05, 1.04: document pages of PostScript manual
2010 Dec 05, 1.03: title
2010 Dec 05, 1.02: clean up
2010 Dec 05, 1.01: tie up loose ends - functional!
2010 Dec 05, 1.00: origin - partially functional!
*)
updateversion = 1.03; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.piechart *)
(*
name
   piechart: create a pie chart

synopsis
   piechart(piechartp: in, pienumbers: in, pieps: out, output: out)

files

   piechartp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      diameter of the chart (cm)

      angle to start (degrees).  This is standard polar coordinates,
      where 0 is along the positive x axis.

      direction +1 or -1 to draw.  As in standard polar coordinates,
      positive is counter clockwise, negative is clockwise.

      title line.  The title is centered on the piechart.  A nice
      thing to have is the name of the computer using `uname -n`.

      font size for the title.

   pienumbers: a set of pienumbers for the pie chart.  Each row is a slice.

      The first column is the pienumbers.  These are summed to get the
      total and then each number is divided by that total so they
      don't have to be fractions in the pienumbers file.

      The second column is either an 'h' for hsb or 'r' for rgb, the
      color model.  If the column contains 'H', it stands for hsb but
      the hsb color is created and columns 3 to 5 are skipped.

      columns 3 to 5 are:
      hue, saturation brightness (hsb model)
      or
      red, green, blue (rgb colormodel)

   pieps:  A postscript program that produces the pie chart.

   output: messages to the user

description

   Make a pie chart as a postscript program.

examples

documentation

see also

   memory.p

   {PostScript Language Reference, third edition}
   http://www.adobe.com/products/postscript/pdfs/PLRM.pdf
   page 531 describes arc and arcn.
   page 405 describes page size control.

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.piechart *)

var
   pienumbers, (* file used by this program *)
   piechartp, (* file used by this program *)
   pieps: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

(* begin module copyalinenocr *)
procedure copyalinenocr(var fin, fout: text);
(* copy a line from file fin to file fout, no cr  *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
end; (* copyalinenocr *)
(* end module copyalinenocr *)

function filelength(var title: text):integer;
(* determine the length of the first file line. *)
var
   n: integer; (* number of charaters *)
begin
   reset(title);
   n := 0;
   while not eoln(title) do begin
      get (title);
      n := n + 1;
   end;
   filelength := n;
end;

(* begin module piechart.themain *)
procedure themain(var piechartp, pienumbers, pieps: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)

   diametercm: real; (* diameter of the piechart, cm *)
   angle: real; (* starting angle of the piechart *)
   direction: integer; (* direction to dra piechart *)
   fontsize: integer; (* font size for title *)
   title: text; (* title line *)

   total: real; (* sum of numbers for computing fractions and then
                   angles *)
   anumber: real; (* a numbers to use for the piechart *)
   anumbermax: real; (* highest anumber *)
   sections: integer; (* number of numbers = number of pie sections *)
   section: integer; (* one of the pie sections *)
   wid: integer; (* width of numbers in piechart *)
   dec: integer; (* decimals of numbers in piechart *)

   diameterpts: integer; (* diameter in points *)
   radiuspts: integer; (* radius in points *)
   edgepts: integer; (* extra edge for the graph in points *)
   squareedge: integer; (* the side of the graphic square *)

   colortype: char; (* rgb or hsb identifier *)
   redhue, greensaturation, bluebrightness: real; 
     (* rgb or hsb color values *)

   anglecurrent: real; (* current angle of the piechart *)
   anglenext: real; (* the next angle of the piechart *)

begin
   writeln(output,'piechart ',version:4:2);
   reset(piechartp);
   readln(piechartp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   readln(piechartp, diametercm);
   readln(piechartp, angle);
   readln(piechartp, direction);
   rewrite(title);
   copyaline(piechartp, title);
   readln(piechartp, fontsize);

   (* First Pass through the numbers, find total and number sections *)
   reset(pienumbers);
   total := 0;
   anumbermax := -maxint;
   sections := 0;
   while not eof (pienumbers) do begin
      readln(pienumbers, anumber);
      total := total + anumber;
      sections := sections + 1;
      if anumbermax < anumber then anumbermax := anumber;
   end;

   wid := 10; (* characters *)
   dec := 8; (* characters *)
   edgepts := 3; (* points extra around the edge *)

   diameterpts := round((diametercm / 2.54) * 72);
   radiuspts := round(diameterpts/2);
   squareedge := diameterpts+2*edgepts;

   rewrite(pieps);
writeln(pieps, '%!PS-Adobe-2.0 EPSF-2.0');
writeln(pieps, '%%Title: piechart ',version:4:2);
writeln(pieps, '%%Creator: Tom Schneider, toms@alum.mit.edu',
                  ' https://alum.mit.edu/www/toms');
writeln(pieps, '%%BoundingBox:    0 0 ', squareedge:1,
                                    ' ', squareedge:1);
writeln(pieps, '%%Pages: 1');
writeln(pieps, '%%DocumentFonts:');
writeln(pieps, '%%EndComments');
writeln(pieps, '/cmfactor 72 2.54 div def % defines points -> cm conversion');
writeln(pieps, '/cm {cmfactor mul} bind def % defines centimeters');
writeln(pieps, '<< /PageSize [', (* page size control! *)
               squareedge:1,' ',squareedge:1,'] >> setpagedevice');

   writeln(pieps, '%');
   writeln(pieps, '% Parameters:');
   writeln(pieps, '% ',diametercm:wid:dec, ' diameter (cm)');
   writeln(pieps, '% ',angle:wid:dec, ' start angle, degrees');
   writeln(pieps, '% ',direction:1, ' direction');
   writeln(pieps, '% ');
   writeln(pieps, '% ',sections:1,' sections');
   writeln(pieps, '% ',total:wid:dec,' total');

writeln(pieps);
writeln(pieps, radiuspts+edgepts:1, ' dup translate');

(* useful postscript page:
http://homepage.mac.com/andykopra/pdm/tutorials/an_introduction_to_postscript.html
*)
(* x y r ang1 ang2  arc - *)

{ arc demo:
writeln(pieps, 'newpath 0 0 moveto');
writeln(pieps, '0 0 ',radiuspts:1,' 10 45 arc closepath fill');
}

writeln(pieps, '/startarc {newpath 0 0 moveto 0 0 ',radiuspts:1,'} def');
writeln(pieps, '/positiveendarc {arc  closepath fill} def');
writeln(pieps, '/negativeendarc {arcn closepath fill} def');

{demo:
writeln(pieps, 'startarc 35 90 positiveendarc');
}

   (* Second Pass through the numbers, this time plot *)
   anglecurrent := angle;
   section := 0;
   reset(pienumbers);
   while not eof (pienumbers) do begin
      read(pienumbers, anumber);
      skipblanks(pienumbers);
      read(pienumbers, colortype);
      if colortype = 'H' then begin
         (* 
         redhue := anumber/total;
         This doesn't work when all numbers are similar.
         Then they all have the same red/orange color.
         Instead, normalize to the maximum, which spreads
         the values out:
         *)
         redhue := (anumber/anumbermax)*0.84 + 0.16;
{
write  (output,' ',anumber:10:5);
write  (output,' ',anumbermax:10:5);
writeln(output,' ',redhue:10:5);
}
         greensaturation := 1.0;
         bluebrightness := 1.0;
      end
      else begin
         read(pienumbers, redhue, greensaturation, bluebrightness );
      end;
      readln(pienumbers);

      writeln(pieps, '% ',anumber:wid:dec, ' a number');
      writeln(pieps, '% ',anumber/total:wid:dec, ' the number as a fraction');
      writeln(pieps, '% ',colortype:wid, ' % color type');
      writeln(pieps, ' ',redhue:wid:dec, ' % redhue');
      writeln(pieps, ' ',greensaturation:wid:dec, ' % greensaturation');
      writeln(pieps, ' ',bluebrightness:wid:dec, ' % bluebrightness');

      if (colortype = 'h') or (colortype = 'H')
      then writeln(pieps, 'sethsbcolor ')
      else writeln(pieps, 'setrgbcolor ');

      if direction > 0
      then begin
         anglenext := anglecurrent + 360*anumber/total;
         if anglenext < 360 then anglenext := anglenext - 360;
         writeln(pieps, 'startarc');
         writeln(pieps, anglecurrent:wid:dec);
         writeln(pieps, anglenext:wid:dec);
         writeln(pieps, 'positiveendarc');
      end
      else begin
         anglenext := anglecurrent - 360*anumber/total;
         if anglenext < 0 then anglenext := anglenext + 360;
         writeln(pieps, 'startarc');
         writeln(pieps, anglecurrent:wid:dec);
         writeln(pieps, anglenext:wid:dec);
         writeln(pieps, 'negativeendarc');
      end;
      anglecurrent := anglenext;
      section := section + 1;
      
   end;

   (* generate title line in center *)
   writeln(pieps);

   (* google:
      aspect ratio courier font
      http://www.sqrug.org/sqr-users/sqrusers_02q2/msg00227.html
      'Courier font characters have an "aspect ratio" of 3:5'
      Use this fact and font size to do horizontal scale computation.
   *)
   
   writeln(pieps, '0 0 moveto');
   (* center the title *)
   writeln(pieps,((-(3/5)*fontsize*filelength(title))/2):wid:dec);
   writeln(pieps,(-fontsize/4):wid:dec); (* computer name is lower case *)
   writeln(pieps, ' moveto');
   writeln(pieps, '/Courier-Bold findfont');
   writeln(pieps, fontsize:1,' scalefont');
   writeln(pieps, 'setfont');
   write(pieps, '0 setgray');
   write(pieps, '(');
   reset(title);
   copyalinenocr(title,pieps);
   writeln(pieps, ') show');

end;
(* end module piechart.themain *)

begin
   themain(piechartp, pienumbers, pieps);
1: end.
