program rsconv(rsdata1, rsdata2, rsconvp, xyin, rsconvmax, output);
(* rsconv: Compute the convolution between two rsdata files.

  Tom Schneider
  permanent email: toms@alum.mit.edu
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.19; (* of rsconv.p 1998 June 10
origin 1998 June 4 *)
updateversion = 1.08; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.rsconv *)
(*
name
   rsconv: Compute the convolution between two rsdata files.

synopsis
   rsconv(rsdata1, rsdata2: in, rsconvp: in,
          xyin: out, rsconvmax: out, output: out)

files
   rsdata1, rsdata2:  rsdata files from the rseq program

   rsconvp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      convolutionmethod: A single character that determines
         how the convolution is performed:
         m(ultiplication) or c(orrelation)

   xyin:  input to the xyplo program for graphing.  columns:
            1: point definition: d(ata) p(eak)
            2: coordinate
            3: convolution value

   rsconvmax:  maximum convolution value (one real)

   output: messages to the user

description

   Rsconv will convolve two information curves by one of two methods:

      multiplication: The two Rsequence(l) vectors are set at various
      relative positions.  Each corresponding pair of positions is then
      multiplied together and summed, to get total bits squared for a
      particular offset.  This is the dot product of the two information
      curves represented as vectors.  The cosine of the angle between them is
      then found by dividing by the magnitude of each vector.

      correlation:  The standard correlation coefficient is computed for the
      overlapping region

  The convolution method can't use the actual bases since they differ.
Individual information can be used instead for comparisons based on sequence
matching.

  When looking for a relationship between human acceptor and E. coli ribosome
binding site, we don't want to get large hits by doing the acceptor to the T7
promoters.  Either normalization takes care of this.

examples

   Example files are listed below.
   The file rsconv.xyplop is used for standard runs.  Change the zzz to the
   appropriate pair.  This can be done conveniently in Unix with sed:

         cat rsconv.xyplop |  sed -e "s/zzz/name/" | cat > xyplop

documentation

see also
   rseq.p,
   xyplo.p,
   rsconvp,
   rbs1055.rsdata,
   spliceA.rsdata,
   rsconv.xyplop

author
   Thomas Dana Schneider

bugs
   The largest combined size of the two Rs(l) vectors is given by constant
   maxposition.

technical notes

*)
(* end module describe.rsconv *)

(* begin module rsconv.const *)
   infodec = 5; (* number of decimal places for printing information *)
   infowid = 10; (* width of reported information *)

   maxposition = 1000; (* largest range allowed *)
(* end module rsconv.const *)

var
   rsdata1, rsdata2, (* file used by this program *)
   rsconvp, (* file used by this program *)
   rsconvmax, (* file used by this program *)
   xyin: text; (* file used by this program *)

type
   (* data array for storing Rs curves *)
   dataarray = record
(*
      l: array[1..maxposition] of integer;
      a: array[1..maxposition] of integer;
      c: array[1..maxposition] of integer;
      g: array[1..maxposition] of integer;
      t: array[1..maxposition] of integer;
*)
      rangefrom: integer; (* the 'from' part of the range *)
      rangeto: integer; (* the 'to' part of the range *)
      rangesize: integer; (* size of the entire range *)
      rsl: array[1..maxposition] of real;
      rs: real; (* sum of rsl, computed from read data to avoid
         roundoff discrepancies *)
   end;

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'matmod 2.06 94 sep 5 tds/gds'; *)

(* begin module linear.regression *)
(* linear regression procedure
origin: 1980 august 5
      by thomas schneider
      copyright 1986 *)

procedure regress ( control: char; (* described below *)
                    x, y: real;    (* the data pairs *)
                var sumx, sumy, sumxsqd, sumysqd, sumxy, (* internal records *)
(* mean, variance and covariance of x and y: *)
                    ex, ey, varx, vary, covxy,
(* correlation coefficient: *)
                    r,
(* m = slope, b = y intercept as in y = mx + b *)
                    m, b
                        : real;
(* number of data pairs entered: *)
                var n: integer);
(* linear regression for variables x and y.
   the control variable has only three acceptable states:
      c clear all variables to zero
        (note that r is set to 2.0, so that a program can check whether
         or not the results have been calculated yet. the recommended test
         is not for equality, since round-off could cause problems.  use
              if r > 1.5 then (no calculations yet)
                         else (calculations done)    )
      e enter the x and y values into the internal sums
      r results are calculated from the sums
*)
begin
      if not( control in ['c','e','r']) then begin
         writeln(output,' linear regression control variable value, "',control,
                        '" is not acceptable.');
         writeln(output,' it must be in ["c","e","r"]');
         halt
      end

      else case control of
         'c': begin (* clear *)
            x:=0.0; y:=0.0;
            sumx:=0.0; sumy:=0.0;
            sumxsqd:=0.0; sumysqd:=0.0;
            sumxy:=0.0;
            ex:=0.0; ey:=0.0;
            varx:=0.0; vary:=0.0;
            covxy:=0.0;
            r:=2.0; (* this (impossible) value can act as a flag that
                       shows that the calculations have not been made. *)

            m:=0.0; b:=0.0;
            n:=0
         end;

         'e': begin (* enter data *)
            sumx:=sumx + x;
            sumy:=sumy + y;
            sumxsqd:=sumxsqd + x*x;
            sumysqd:=sumysqd + y*y;
            sumxy:=sumxy + x*y;
            n:=succ(n)
         end;

         'r': begin (* calculate results *)
            (* check for conditions that would bomb the program *)
            if n = 0 then begin
               writeln(output,'regress: n is 0; no samples recorded');
               halt
            end;
            if n = 1 then begin
               writeln(output,'regress: n is 1; regression impossible');
               halt
            end;

            ex:=sumx/n;
            ey:=sumy/n;

            varx:=(sumxsqd/n) - ex*ex;
            vary:=(sumysqd/n) - ey*ey;
            if varx=0.0 then begin
               writeln(output,'regress: variance of x is zero;',
                              ' regression impossible');
               halt
            end;
            if vary=0.0 then begin
               writeln(output,'regress: variance of y is zero;',
                              ' regression impossible');
               halt
            end;

            covxy:=(sumxy/n) - ex*ey;

            r:=covxy/(sqrt(varx * vary));

            m:=covxy/varx;
            b:=ey - m*ex
         end
      end
end;
(* end module linear.regression version = 'matmod 2.06 94 sep 5 tds/gds'; *)

(* begin module cleararray *)
procedure cleararray(var a: dataarray);
(* set data array values to zero *)
var
   position: integer; (* position in the array *)
begin
   with a do begin
      for position := 1 to maxposition
      do rsl[position] := -111111.0; (* a funny number to see! *)
   end;
end;
(* end module cleararray *)

(* begin module maxarray *)
function maxarray(a: dataarray): real;
(* find the maximum array value *)
var
   position: integer; (* position in the array *)
   max: real; (* maxiumu value in the array *)
begin
   with a do begin
      for position := 1 to maxposition
      do if max < rsl[position] then max := rsl[position]
   end;
   maxarray := max
end;
(* end module maxarray *)

(* begin module findpeak *)
procedure findpeak(a: dataarray;
                   var peaklocation: integer;
                   var peakvalue: real); (* maximum value in the array *)
(* find the maximum array value, return peak location and value.
Looking only at one position near by ends up finding a local peak.
So look at the width positions nearby. *)
const
   debug = false; (* turn on to see the stepping *)
   width = 20; (* turn on to see the stepping *)
var
   position: integer; (* position in the array *)
   test: integer; (* testing for a new position in the array *)
   done: boolean; (* found a peak *)
   changed: boolean; (* moved to a new position *)
begin
   with a do begin
      position := 1 - rangefrom; (* the zero coordinate *)
      peakvalue := rsl[position];
      done := false;

if debug then begin
   peaklocation := position + (rangefrom - 1);
   write  (output,'initial peak location: ',peaklocation:1);
   writeln(output,' initial peak value: ',peakvalue:infowid:infodec);
end;

      while not done do begin
         changed := false;

         for test := position - width to position + width do begin
if debug then begin
   write  (output,'   test position: ',test+rangefrom-1:1);
   writeln(output,'   test value: ',rsl[test]:infowid:infodec);
end;
            if rsl[test] > peakvalue then begin
               position := test;
               peakvalue := rsl[position];
               changed := true;
            end;
         end;

         if not changed then done := true;

if debug then begin
   peaklocation := position + (rangefrom - 1);
   write  (output,'peaklocation: ',peaklocation:1);
   writeln(output,'   peakvalue: ',peakvalue:infowid:infodec);
end
      end;

      peaklocation := position + (rangefrom - 1);
   end;
end;
(* end module findpeak *)

(* begin module readrsdata *)
procedure readrsdata(var r: text; var anarray: dataarray);
(* read the rsdata file.  Skip * comment lines and absorb data:
*   l    a    c    g    t    rs(l)       rs var(hnb)  sum-var n(l)   e(hnb) f
*)
var
   notstarted: boolean; (* to detect when we have read the first data *)
   l, a, c, g, t: integer; (* data being read *)
   position: integer; (* position in the rsdata file *)
begin
   reset(r);
   position := 0;
   notstarted := true;
   cleararray(anarray);
   while not eof(r) do begin
      if r^ = '*'
      then readln(r)
      else with anarray do begin
         position := succ(position);
         readln(r,
                   l, a, c, g, t,
                  (* l[position],
                   a[position], c[position], g[position], t[position], *)
                 rsl[position]);
         if notstarted then begin
            anarray.rangefrom := l;
            notstarted := false
         end;
         rs := rs + rsl[position];
      end;
   end;
   with anarray do begin
      rangeto := l;
      rangesize := position;
      if rangesize <> (rangeto-rangefrom+1) then begin
         writeln(output,'range size does not equal from-to size');
         writeln(output,'rangesize = ',rangesize:1);
         writeln(output,'rangefrom = ',rangefrom:1);
         writeln(output,'rangeto = ',rangeto:1);
      end;
   end;
end;
(* end module readrsdata *)

(* begin module writearray *)
procedure writearray(var r: text; startcharacter: char;
                     anarray: dataarray);
(* write the an array to file r.  Put the startcharacter at the
begining of the line for identificiation purposes in the graph *)
var
   position: integer; (* position in the array *)
   l: integer; (* position in site *)
begin
   with anarray do begin
      l := rangefrom;
      for position := 1 to rangesize do begin
         write(r,startcharacter,' ');
         writeln(r,l:1,' ',rsl[position]:infowid:infodec);
         l := l + 1
      end;
   end;
end;
(* end module writearray *)

(* begin module magnitude *)
function magnitude(anarray: dataarray): real;
(* magnitude of an array *)
var
   position: integer; (* position in the array *)
   sum: real; (* sum of squares *)
begin
   with anarray do begin
      sum := 0.0;
      for position := 1 to rangesize do begin
         sum := sum + sqr(rsl[position])
      end;
      magnitude := sqrt(sum)
   end;
end;
(* end module magnitude *)

(* begin module minimum *)
function minimum(a,b: integer): integer;
(* return the minimum of a and b *)
begin
   if a < b then minimum := a
            else minimum := b
end;
(* end module minimum *)

(* begin module maximum *)
function maximum(a,b: integer): integer;
(* return the maximum of a and b *)
begin
   if a > b then maximum := a
            else maximum := b
end;
(* end module maximum *)

(* begin module convolve *)
procedure convolve(convolutionmethod: char;
                   a, b: dataarray; var c: dataarray);
(* convolve array a and array b and put the result in array c.
Use the method:  m(ultiply) or c(orrelation) *)
const
   debug = false; (* turn on to see the stepping *)
var
   aposition: integer; (* position in a array *)
   bposition: integer; (* position in b array *)
   cposition: integer; (* position in c array *)
   offset: integer; (* the relative positions of the a and b dataarrays *)
   position: integer; (* position *)
   ars, brs: real; (* totals for a and b vectors only for overlap region.
                      This gives a more reasonable normalization *)

   (* variables for linear regression: *)
                    x, y: real;    (* the data pairs *)
                sumx, sumy, sumxsqd, sumysqd, sumxy, (* internal records *)
(* mean, variance and covariance of x and y: *)
                ex, ey, varx, vary, covxy,
(* correlation coefficient: *)
                r,
(* m = slope, b = y intercept as in y = mx + b *)
                m, bintercept
                        : real;
(* number of data pairs entered: *)
                n: integer;
begin
   (*
      a is fixed relative to c.  b slides past.
      b starts at negative position with respect to a.
    [--0-]       a range is -3 to +2
   [---0----]    b range is -4 to +5

            [--0-]          a is fixed relative to c
   [---0----]               b start
                 [---0----] b end
       [-------0-----]      c records data, range is: -8 to +6

   [---0----]               b start
    [---0----]
     [---0----]
      [---0----]
       [---0----]
        [---0----]
         [---0----]
          [---0----]
           [---0----]
            [---0----]
             [---0----]
              [---0----]
               [---0----]
                [---0----]
                 [---0----] b end at step 15

            [--0-]          a is fixed relative to c
    [---0----]              b at step 2
       [-------0-----]      c records data

            [--0-]          a is fixed relative to c
                [---0----]  b at step 14
       [-------0-----]      c records data

   *)

   c.rangefrom := a.rangefrom - b.rangeto;
   c.rangeto   := a.rangeto   - b.rangefrom;
   c.rangesize := c.rangeto - c.rangefrom + 1;

   if debug or (c.rangesize > maxposition) then begin
      writeln(output,'a.range from/to/size:',      a.rangefrom :10,
                                                   a.rangeto   :10,
                                                   a.rangesize :10);
      writeln(output,'b.range from/to/size:',      b.rangefrom :10,
                                                   b.rangeto   :10,
                                                   b.rangesize :10);
      writeln(output,'c.range from/to/size:',      c.rangefrom :10,
                                                   c.rangeto   :10,
                                                   c.rangesize :10);
      if c.rangesize > maxposition then begin
         writeln(output,'Convolution results exceed available space.');
         writeln(output,'Increase constant maxposition from');
         writeln(output,'maxposition = ',maxposition:1);
         writeln(output,'to at least ',c.rangesize:1);
         halt
      end;
   end;

   for offset := c.rangefrom to c.rangeto do begin
      cposition := offset - c.rangefrom+1;

      if convolutionmethod = 'c' then begin
         regress('c',
                   x, y,
                   sumx, sumy, sumxsqd, sumysqd, sumxy,
                   ex, ey, varx, vary, covxy,
                   r, m, bintercept, n);
         x := 0.0; (* not used but just for clarity. *)
         y := 0.0;
      end;

      if debug then write(output, 'offset = ',offset:5,' | ');
      c.rs := 0.0;
      ars := 0.0;
      brs := 0.0;
      for position := maximum(a.rangefrom, offset + b.rangefrom)
                      to
                      minimum(a.rangeto,   offset + b.rangeto)
      do begin
         if debug then write(output, position:3);
         aposition := position          - a.rangefrom + 1;
         bposition := position - offset - b.rangefrom + 1;

         c.rs := c.rs
               + a.rsl[aposition]
               * b.rsl[bposition];

         ars := ars
               + a.rsl[aposition]
               * a.rsl[aposition];

         brs := brs
               + b.rsl[bposition]
               * b.rsl[bposition];

         if convolutionmethod = 'c' then begin
            regress('e',
                      a.rsl[aposition], b.rsl[bposition],
                      sumx, sumy, sumxsqd, sumysqd, sumxy,
                      ex, ey, varx, vary, covxy,
                      r, m, bintercept, n);
         end;
      end;

      if convolutionmethod = 'c' then begin
         if n > 2
         then regress('r',
                   a.rsl[aposition], b.rsl[bposition],
                   sumx, sumy, sumxsqd, sumysqd, sumxy,
                   ex, ey, varx, vary, covxy,
                   r, m, bintercept, n)
         else r := 0;
         c.rsl[cposition] := r;
      end
      else c.rsl[cposition] := c.rs/sqrt(ars*brs); (* normalization *)

      if debug then writeln(output);

   end;
end;
(* end module convolve *)

(* begin module rsconv.themain *)
procedure themain(var rsdata1, rsdata2, rsconvp, xyin, rsconvmax: text);
(* the main procedure of the program *)
var
   convolutionmethod: char; (* method of doing the convolution *)
   parameterversion: real; (* parameter version number *)
   peaklocation: integer; (* location of the peak of an array *)
   peakvalue: real; (* value at the peak of an array *)
   arrayA: dataarray; (* an array of data from rsdata1 *)
   arrayB: dataarray; (* an array of data from rsdata2 *)
   arrayC: dataarray; (* convolution of array 1 and 2 *)
begin
   writeln(output,'rsconv ',version:4:2);
   reset(rsconvp);
   readln(rsconvp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(rsconvp, convolutionmethod);

   rewrite(xyin);
   writeln(xyin,'* rsconv ',version:4:2);
   writeln(xyin, '* ', convolutionmethod, ': convolution method');
   writeln(xyin, '* columns:');
   writeln(xyin, '* 1: point definition: d(ata) p(eak)');
   writeln(xyin, '* 2: coordinate');
   writeln(xyin, '* 3: convolution value');

   readrsdata(rsdata1, arrayA);
   readrsdata(rsdata2, arrayB);

   cleararray(arrayC);

   convolve(convolutionmethod, arrayA, arrayB, arrayC);

   writearray(xyin, 'd', arrayC);

   findpeak(arrayC, peaklocation, peakvalue);

   rewrite(rsconvmax);
   writeln(rsconvmax,     peakvalue:infowid:infodec,
                     ' ', peaklocation:infowid);
   (* original method was just maximum of the array, but there is a lot
      of noise near the outer edged.  so this was replaced by
      findpeak, which finds the peak nearest the zero coordinate. *)
   (* writeln(rsconvmax,maxarray(arrayC):infowid:infodec); *)

   writeln(xyin, 'p ',peaklocation:1, (* mark the peak on the graph *)
                 ' ', peakvalue:infowid:infodec);

end;
(* end module rsconv.themain *)

begin
   themain(rsdata1, rsdata2, rsconvp, xyin, rsconvmax);
1: end.
