program scale(fin, scalep, fout, output);
(* scale: scale a column of a table

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.07; (* of scale.p 2013 Feb 22
2013 Feb 22, 1.07: fix documentation
2013 Feb 22, 1.06: work on scaling. - functional
2013 Feb 21, 1.05: clean up
2013 Feb 21, 1.04: functional
2013 Feb 21, 1.03: tidy, clean up, fix logic
2013 Feb 21, 1.02: add string functions, column manipulation
2013 Feb 18, 1.01: add notes
2013 Feb 18, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.scale *)
(*
name
   scale: scale a column of a table

synopsis
   scale(fin: in, scalep: in, fout: out, output: out)

files

   fin:  A data table with space separated columns.
         Comment lines are started with '*', they are just copied
         to the fout.

   fout: The fin data table with a particular column altered or
         an additional column on the end created according to the
         parameters.

   scalep:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      inputcolumn: integer; the column to read in.
      lowerbound:  real;    the lower bound to accept
      upperbound:  real;    the upper bound to accept
      multiplier:  real;    a number to multipy the data by
      adder:       real;    a number to add to the data after multiplication
      colwid:      integer; output column width in characters
      coldec:      integer; output column decimals in characters
      newcolumn:   character; if 'n' put a new column at the end
         of the line, otherwise alter the inputcolumn.
         NB: newcolumn is not implemented yet

   output: messages to the user

description

   This program allows one to manipulate the data in a space delimited
   data table.  One can add a column at the end or one can alter a
   column in the middle of the table.  One can exclude values outside
   a given range (lowerbound to upperbound).

examples

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.scale *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.48; (@ of prgmod.p 2011 May 03 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.48; (@ of prgmod.p 2011 May 03 *)

var
   fin, (* file used by this program *)
   scalep, (* file used by this program *)
   fout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 representing a place value in the number *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module grabtoken *)
procedure grabtoken(var thefile: text; var thestring: string);
(* skip any blanks and then grab the next token from the file *)
var
   c: char; (* a character in thefile *)
   done: boolean; (* done finding the name *)
begin
   skipblanks(thefile);
   done := false;
   with thestring do begin
      length := 0;
      while not done do begin
         if eoln(thefile)
         then done := true
         else begin
            read(thefile,c);
            if c = ' '
            then done := true
            else begin
               length := succ(length);
               letters[length] := c;
            end
         end
      end
   end
end;
(* end module grabtoken version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module copyalineNOln *)
procedure copyalineNOln(var fin, fout: text);
(* copy a line from file fin to file fout.
Don't go to the next line. *)
begin (* copyalineNOln *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
end; (* copyalineNOln *)
(* end module copyalineNOln *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module skipstringblanks *)

procedure skipstringblanks(var line: string);
(* skip blanks in the line *)
begin
   with line do while (letters[current] = ' ') and (current < maxstring)
   do current := succ(current)
end; (* skipstringblanks *)

procedure skipstringnonblanks(var line: string);
(* skip nonblanks in the line *)
begin
   with line do while (letters[current] <> ' ') and (current < maxstring)
   do current := succ(current)
end; (* skipstringblanks *)

procedure skipstringcolumn(var line: string);
(* skipstring over a data column in the line *)
begin
   skipstringblanks(line); skipstringnonblanks(line);
end;

procedure stringcolumn(line: string; column: integer; var datum: string);
(* extract the column from the line and put into the datum string.
Ignore leading blanks in defining a column. *)
const
{
   debug = true;
}
   debug = false;
var
   c: integer; (* index to the datum characters *)
   columnnumber: integer; (* current column in the line *)
   done: boolean; (* did we find the datum column *)
   positionbeg: integer; (* position of the datum, begin *)
   positionend: integer; (* position of the datum, end *)
begin
   columnnumber := 0;
   done := false;
   if debug then begin writestring(output, line); writeln(output); end;
   while not done do begin
       if (columnnumber = column) or (line.current >= line.length) then begin
          done := true;
       end
       else begin
          skipstringblanks(line);
          positionbeg := line.current;
          if debug then write(output,'|':line.current,' ',line.current:1);
          if debug then write(output,'"',line.letters[positionbeg],'"');
          if debug then writeln(output);
          
          skipstringnonblanks(line);
          positionend := line.current - 1;
          if debug then write(output,'|':line.current,' ',line.current:1);
          if debug then write(output,'"',line.letters[positionend],'"');
          if debug then writeln(output);

          columnnumber := succ(columnnumber);
       end;
   end;
   clearstring(datum);
   if columnnumber = column then begin
      datum.length := positionend - positionbeg + 1;
      for c := 1 to datum.length do begin
          if debug then write(output,line.letters[positionbeg + c - 1]);
          datum.letters[c] := line.letters[positionbeg + c - 1];
      end;
   end;
   if debug then writeln(output);
{
   if debug then halt;
}
end;

(* end module skipstringblanks version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module scale.themain *)
procedure themain(var fin, scalep, fout: text);
(* the main procedure of the program *)
const
   verbose = false; (* set to true to get more details *)
var
   parameterversion: real; (* parameter version number *)

      inputcolumn: integer; (* the column to read in. *)
      lowerbound:  real;    (* the lower bound to accept *)
      upperbound:  real;    (* the upper bound to accept *)
      multiplier:  real;    (* a number to multipy the data by *)
      adder:       real;    (* a number to add to the data after multiplication *)
      colwid:      integer; (* output column width in characters *)
      coldec:      integer; (* output column decimals in characters *)
      newcolumn:   char;    (* if 'n' put a new column at the end *)

      datum:  real;    (* a data item read in from column inputcolumnt
                          to be manipulated *)

(* procedure figurestring( *)
      line: string; (* a string of characters to figure out *)
      first: integer; (* first found non-blank character in the line *)
      last:  integer; (* last character before a blank after first *)
      whzat: char; (* what the token is *)
      c:     char;  (* the first character of the token *)
      i:     integer; (* integer value of token if it is integer; or 0 *)
      r:     real; (* the real value if it is real; or 0.0 *)

   gotten: boolean; (* line was gotten ok? *)
   token: string; (* a string read in *)

   internalfile: text; (* file to put results into - and write
      out only if  the datum is within lowerbound to upperbound *)

   kept:    integer; (* number of data lines kept *)
   skipped: integer; (* number of data lines skipped *)

begin
   writeln(output,'scale ',version:4:2);
   reset(scalep);
   readln(scalep, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(scalep, inputcolumn);
   readln(scalep, lowerbound);
   readln(scalep, upperbound);
   readln(scalep, multiplier);
   readln(scalep, adder);
   readln(scalep, colwid);
   readln(scalep, coldec);
   readln(scalep, newcolumn);

   kept := 0;
   skipped := 0;

   reset(fin);
   rewrite(fout);

   while not eof(fin) do begin
      if fin^ = '*'
      then copyaline(fin,fout)
      else begin
         {zq}

         (* copy up to the desired column *)
         rewrite(internalfile);
         for i := 1 to inputcolumn -1 do begin
            { skipnonblanks(fin); }
            skipblanks(fin);
            grabtoken(fin, token);
            (* Avoid adding blank at line start: *)
            if i > 1 then write(internalfile,' ');
            writestring(internalfile,token);
         end;

         read(fin, datum); (* read the desired column datum *)

         if (lowerbound <= datum) and (datum <= upperbound) then begin
            kept := kept + 1;
            if verbose then writeln(output,'----- kept ', kept:1);

            (* manipulate the datum *)
{
            datum := multiplier * datum + adder;
}
            datum := multiplier * (datum + adder);

            write(internalfile, ' ',datum:colwid:coldec);

            copyalineNOln(fin, internalfile);
            readln(fin);
            writeln(internalfile);
            reset(internalfile);
            copyaline(internalfile, fout);
            reset(internalfile);
            if verbose then copyaline(internalfile, output);
         end
         else begin
            skipped := skipped + 1;
            if verbose then begin
               writeln(output,'----- skipped ', skipped:1);
               reset(internalfile);
               copyalineNOln(internalfile, output);
               write(output, ' [skipping ',datum:colwid:coldec,']');
               copyaline(fin, output); (* this does a readln *)
            end
            else readln(fin);
         end
      end
   end;

   writeln(output,'skipped: ', skipped:5);
   writeln(output,'   kept: ', kept:5);
   writeln(output,'  total: ', (skipped+kept):5);

end;
(* end module scale.themain *)

begin
   themain(fin, scalep, fout);
1: end.
