program tod(abi, ssb, thedate, todp, results, summary, db, output);
(* tod: to database format for sites program

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 3.01; (* of tod.p 2016 Jan 24
2016 Jan 24, 3.01 debug
  sed -i.bak "s|'9','0'|'9'|g"                      tod.p
1993 Jan 28, 3.00 functional
1992 Oct 14, 1.00 origin
origin 1992 October 14 *)
(* end module version *)

(* begin module describe.tod *)
(*
name
   tod: to database format for sites program

synopsis
   tod(abi: in, thedate: in, ssb: in, todp: in,
       results: out, summary: out, db: out, output: out)

files
   abi: Raw sequences from the ABI sequencing machine.  The
      files called *_??.Seq.txt are manipulated under Unix by:

      more *_??.Seq.txt | cat > abi
      echo "" >> abi

      The more program puts each name followed by the contents, and it is smart
      enough to pipe it to cat which joins the results together.  Thus the abi
      file contains the sample names followed by the sequences.  The echo puts
      a single carriage return at the end of the file so that it ends cleanly.

   thedate:  Date that the sequences were run, output of makedate program.

   ssb:  This must be a copy of the "Sample_Sheet.bin" file from the ABI
      machine.  It contains:
           lane number
           plasmid name
           primer name
      in a funny non-ASCII format which this program extracts from.  (The
      program extracts the data from their known rigid locations.)  The sample
      name column of the sample sheet must contain the plasmid name.  Any
      number of spaces, slashes (/) or null characters are then skipped and the
      next non null word (ending in null or space) is taken as the primer
      name.  Thus the format is "plasmid/primer".  For example:  pTS421/pTS37f1

      There is a bug in the ABI code which will replace the first letter
      of the 24th lane with a null character sometimes.  To get around
      the bug, we will try to rewrite the sample sheet if this appears.

   todp: parameters to control the program
      first line:  a string of characters, called R1, which represents
         a restriction site or other sequence.
         NOTE: it should be self complementary.
      second line:  a string of characters, called R2, which represents
         a restriction site or other sequence.
         NOTE: it should be self complementary.
      following lines: Editing commands for sequences.  There is one editing
	 command per line, and each consists of three integers (called N, P1
	 and P2) followed by a string (S).  N is the lane number to edit.
	 The P1 and P2 define two positions in the sequence.  The sequence
	 between these positions is deleted and replaced by the string S.  The
	 string must contain only the letters 'acgt' or the single letter 'd'.
	 This allows one to insert sequence (make the P1 = P2 + 1, string is
	 'acgt' form), to delete (P1 > P2 + 1, string is 'd') and to replace
	 (make P1 = P2 + 1 + length of string in 'acgt' form).  Lines that
	 begin with '*' are comments, copied to the results file.  Comments
         for each edit must be placed just below the edit command line.

   results: running commentary of the processing of the sequences.

      The following changes are made:
      1.  Each sequence is edited according to instructions in todp.
      2.  Each sequence is converted to lower case.
      3.  The letter 'n' is converted to 'x'.
      4.  When there is exactly one copy of R1 and one copy of R2,
          the region between R1 and R2 is printed (including R1 and R2).
          Otherwise, the entire original sequence is printed.
      5.  The sequence complement is printed if necessary to assure that R1
          is printed before R2.  The program will print the original
          sequence if R1 and R2 cannot be found on the complement.

      The sequence is then joined to the data from ssb, and the results printed.

   summary: summary of the results.

   db: The sequences from abi are reformed into the database format needed
      by the sites program.

   output: messages to the user

description
   Convert output sequence from ABI sequencing machine into format usable
   by the sites program.

examples

documentation

see also
   sites.p, makedate.p, dotod

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.tod *)

(* begin module tod.const *)
   maxseq = 1000; (* largest sequence the program can handle *)
   maxname = 30; (* largest name the program can handle *)
   maxfillname= 10; (* largest name the program can handle *)
   quicksequence = 25; (* number of bases to display on initial reading *)
(* end module tod.const *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* end module filler.const version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

type
(* begin module tod.type *)
   fillname = array[1..maxfillname] of char; (* name for filling names *)
   name = record
      letters: array[1..maxname] of char; (* name *)
      length: integer;  (* the length of the name *)
   end;
   seqptr = ^sequence;
   sequence = record (* a sequence *)
      bases: array[1..maxseq] of char; (* the sequence itself *)
      length: integer;  (* the length of the sequence *)
      lanenumber: integer; (* lane number *)
      plasmidname: name; (* plasmid name *)
      primername: name; (* primer used *)
      date: name; (* date of sequencing *)
      start: integer; (* first base to print *)
      stop: integer; (* last base to print *)
      editdeleted: boolean; (* true only if the sequence was deleted
         by editing commands *)
      next:  seqptr; (* next sequence in the data structure *)
   end;
   locptr = ^locations;
   locations = record (* a list of integer locations *)
      spot: integer; (* on spot *)
      next: locptr (* the next location *)
   end;
(* end module tod.type *)

(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

var
   abi: text; (* input sequences *)
   thedate: text; (* date of run *)
   ssb: text; (* sample sheet information table *)
   todp: text; (* parameters to control the program *)
   results: text; (* commentary on the results *)
   summary: text; (* summary of the results *)
   db: text; (* output database format *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
(*         if debugging then begin
            writestring(list,seek);
            writeln(list,'testfortrigger seek.letters[',state:1,']:',
                           seek.letters[state],' ch:',ch);
         end;*)
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin (* reset trigger *)
            state := 0;
            skip := true;
            found := false
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(***************************************************************************)
(***************************************************************************)
(***************************************************************************)
(***************************************************************************)

(* begin module tod.bar *)
procedure bar(var f: text; n: integer; c: char);
(* write a bar of n characters c to f *)
var i: integer; (* index to the characters *)
begin
  for i := 1 to n do write(f,c);
end;
(* end module tod.bar *)

(* begin module tod.clearname *)
procedure clearname(var n: name; f: fillname);
var i: integer;
begin
   with n do begin
      for i := 1 to maxfillname do letters[i] := f[i];
      length := maxfillname;
   end;
end;
(* end module tod.clearname *)

(* begin module tod.writename *)
procedure writename(var f: text; n: name);
(* write the name n to file f *)
var
   i: integer; (* index to the name *)
begin
   with n do begin
      for i := 1 to length
      do write(f, letters[i]);
   end;
end;
(* end module tod.writename *)

(* begin module tod.oldreadname *)
procedure oldreadname(var f: text; var n: name);
(* read the name n from file f.  NEW METHOD: names end in space
or slash *)
var
   c: char; (* a character in f *)
begin
   (* skip leading empty characters *)
   while (ord(f^) = 0) or (f^ in [' ','/']) do get(f);
   with n do begin
      length := 0;
      c := ' ';
      repeat
         read(f,c);
         if (c <> chr(0))
            and (c <> ' ') and (c <> '/') then begin
            length := succ(length);
            letters[length] := c;
         end
      until (ord(c) = 0) or eoln(f) or
            (c = ' ') or (c = '/')
   end;
end;
(* end module tod.oldreadname *)

(* begin module tod.readname *)
procedure readname(var f: text; var n: name; var location: integer);
(* read the name n from file f.  Names end in space, null or slash *)
var
   c: char; (* a character in f *)
begin
   n.length := 0;
   (* skip leading empty characters *)
{
   while (ord(f^) = 0) or (f^ in [' ','/']) do begin
}
   while (f^ in [' ','/']) do begin
{
      if ord(f^) = 0 then with n do begin (* VERY SPECIAL CASE *)
         length := succ(length);
         letters[length] := 'p';
         writeln(output,'CORRECTING NULL FIRST CHARACTER OF NAME TO "p"');
      end;
}
      get(f);
      location := succ(location);
   end;
   with n do begin
      repeat
         read(f,c);
         location := succ(location);
         if (c <> chr(0)) and (c <> ' ') and (c <> '/') then begin
            length := succ(length);
            letters[length] := c;
         end
      until (c = chr(0)) or (c = ' ') or (c = '/') or eoln(f)
   end;
end;
(* end module tod.readname *)

(* begin module tod.copyname *)
procedure copyname(n1: name; var n2: name);
(* copy name n1 into name n2 *)
var
   i: integer; (* index *)
begin
   n2.length := n1.length;
   for i := 1 to n1.length do n2.letters[i] := n1.letters[i]
end;
(* end module tod.copyname *)

(* begin module tod.clearseq *)
procedure clearseq(var seq: sequence);
(* clear the sequence *)
begin
   with seq do begin
      length := 0;
      lanenumber := 0;
      clearname(plasmidname,'plasmid?  ');
      clearname( primername,'primer?   ');
      clearname(       date,'date?     ');
      start := 0;
      stop := 0;
      editdeleted := false;
      next := nil;
   end;
end;
(* end module tod.clearseq *)

(* begin module tod.joinname *)
procedure joinname(var d: name; add: name);
(* add more to the name *)
var
   i: integer; (* index *)
begin
   if d.length + add.length + 1 > maxname then begin
      writeln(output,'name too long in joinname');
      halt
   end;
   d.length := d.length + 1;
   d.letters[d.length] := ' '; (* tack on a space *)
   for i := 1 to add.length do d.letters[i+d.length] := add.letters[i];
   d.length := d.length + add.length;
end;
(* end module tod.joinname *)

(* begin module tod.readdate *)
procedure readdate(var results: text; var date: text; var d: name);
(* read the date *)
var
   location: integer; (* location in the date file *)
   year, month, day: name;
begin
   reset(thedate);
   location := 0;
   if eof(date) then begin
      writeln(output,'The file "thedate" is empty - run makedate.');
      halt
(*
       clearname(d,'date?     ');
*)
   end
   else begin
(* original method was just to read the whole date; but readname
now breaks into spaces; reconstruct it.
       readname(date,d);
*)
      readname(date,year,location);
      readname(date,month,location);
      readname(date,day,location);
      copyname(year,d);
      joinname(d,month);
      joinname(d,day);
      readln(date);  (* allow for multiple dates *)
   end;
   writeln(results);
   write(results,'The date found in file thedate is: "');
   writename(results,d);
   writeln(results,'"');
end;
(* end module tod.readdate *)

(* begin module tod.readssblane *)
procedure readssblane(var ssb: text;
                          lanenumber: integer;
                      var plasmidname, primername: name;
                      var location: integer);
(* read sample data from the ssb file.
Lane number lanenumber will be read.  If the data do not
match, the program will halt.  The plasmid and primer names
are obtained.  The variable location keeps track of the
location in the file.  It must be initialized to zero, and
then only this routine is allowed to change it. *)
const
   lanebot = 131;  (* location of the start of the lane # data *)
{  lanetop = 162;}  (* location of the end of the lane # data *)
   namebot = 163;  (* location of the start of the name data *)
{  nametop = 418;} (* location of the end of the name data *)
   lanelength = 586; (* length of the lane blocks *)
var
   numbers: set of '0'..'9';
   number1,number2: char; (* the characters of the lane number *)
function spot(a: integer): integer;
begin
   spot := a + lanelength*(lanenumber-1);
end;
procedure tolocation(a: integer);
(* move to the location a *)
begin
  if location > a then begin
     reset(ssb);
     location := 1; (* ready to read position 1 *)
  end;
  while location < a do begin
     get(ssb);
     location := succ(location);
  end;
{writeln(output,'tolocation: ',a:1);}
end;
begin
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
   tolocation(spot(lanebot));
   while not (ssb^ in numbers) do begin
      get(ssb); location := succ(location);
   end;
   read(ssb,number1,number2); location := location + 2;
   if not (number2 in numbers) then begin
      writeln(output,'readssblane: bad number: "',number1,number2,'"');
      writeln(output,'(ords: ',ord(number1):1,' & ',ord(number2),')');
      writeln(output,'for lane number ',lanenumber:1);
      halt
   end;

   if (ord(number1) - ord('0')) * 10 + (ord(number2) - ord('0'))
      <> lanenumber then begin
      writeln(output,'readssblane: trying to read lane number ',lanenumber:1,
                     ' but found ',number1,number2);
      halt
   end;

   tolocation(spot(namebot));
   (* if the first character is nul, neither name exists *)
   if ord(ssb^) = 0
   then begin
      plasmidname.length := 0;
      primername.length := 0;
   end
   else begin
      readname(ssb, plasmidname, location);
      readname(ssb, primername, location);
   end;
end;
(* end module tod.readssblane *)

(* begin module tod.checkthedate *)
procedure checkthedate(var thedate, ssb: text);
(* check that the dates in the files thedate and ssb are the same *)
var
   done: boolean; (* finished reporting difference in the dates *)
   wasaspace: boolean; (* true when the previous character was a space.
   This allows the program to accept two spaces in the thedate file *)
procedure position;
begin
   reset(ssb);
   get(ssb);
   get(ssb);
   reset(thedate);
   get(thedate); (* 1 *)
   get(thedate); (* 9 *)
end;
procedure step;
begin
   get(ssb);
   wasaspace := (thedate^=' ');
   get(thedate);
   if wasaspace and (thedate^=' ') then get(thedate)
end;
begin
   position;
   while (ord(ssb^) <> 0) and (not eoln(thedate)) do begin
      if ssb^ <> thedate^ then begin
         writeln(output,'Sample Sheet name does not match date');
         position;
         writeln(output,   'thedate | ssb');
         done := false;
         repeat
{zzz}
            writeln(output,'      ',thedate^,' | ',ssb^);
            done := ssb^ <> thedate^;
            if not done then step
         until done;
         halt
      end;
      step
   end;
end;
(* end module tod.checkthedate *)

(* begin module tod.readssb *)
procedure readssb(var ssb: text; now: name; var data: seqptr);
(* read the ssb table into the data structure *)
var
   d: seqptr; (* the data currently being read in *)
   location: integer; (* location in the ssb file *)
   lane: integer; (* lane number *)
begin
   clearseq(data^);
   d := data;
   copyname(now,d^.date);
   reset(ssb);
   location := 1;
   for lane := 1 to 24 do begin
      if lane > 1 then begin (* open space for the next data item *)
         new(d^.next);
         d := d^.next;
         clearseq(d^);
         copyname(now,d^.date);
      end;
      with d^ do begin
         readssblane(ssb, lane, plasmidname, primername,location);
         d^.lanenumber := lane;
      end;
   end;
end;
(* end module tod.readssb *)

(* begin module tod.readaseq *)
procedure readaseq(var f: text; var seq: sequence);
(* Read the sequence from file f to the end of a line or when a colon appears.
Skip blanks, and move to the beginning of the next line. *)
var c: char;
begin
   with seq do begin
      if length > 0 then begin
         writeln(output,'* THERE IS A SECOND CASE OF SAMPLE ',
                        lanenumber:1,' *');
         rewrite(db);
         halt
      end;
      while (not eoln(f)) and (f^<> ':') do begin
         read(f, c);
         if c <> ' ' then begin
            length := length + 1;
            bases[length] := c;
         end
      end;
      start := 1;
      stop := length;
      readln(f);
   end;
end;
(* end module tod.readaseq *)

(* begin module tod.writeseq *)
procedure writeseq(var f: text; var seq: sequence);
(* write the sequence to file f, over the range start to stop *)
var
   l: integer; (* position in sequence *)
begin
   with seq do begin
      if (start <= stop) and (start > 0) and (stop > 0)
      then for l := start to stop do write(f, bases[l])
      else begin
         write(f,'NO-SEQUENCE');
         if editdeleted then write(f,' (DELETED BY EDITING)')
                        else write(f,' (NO SEQUENCE PROVIDED)')
      end
   end;
end;
(* end module tod.writeseq *)

(* begin module tod.reportlanes *)
procedure reportlanes(var results: text; data: seqptr;
                      expectsequence: boolean);
(* Report all the lanes, and report on the DNA sequence if expectsequence
is true. *)
var
   d: seqptr; (* pointer to the data structure *)
   l: integer; (* index to the sequence *)
begin
   writeln(results);
   bar(results,80,'*'); writeln(results); (* all bars are 80 wide *)
   write(results,'*** Sample Sheet *');
   bar(results,62,'*'); writeln(results); (* all bars are 80 wide *)
   writeln(results);
   d := data; 
   while d <> nil do with d^ do begin
      write(results,'lane number: ',lanenumber:2);
      if plasmidname.length > 0 then begin
         write(results,'  plasmid: ');
         writename(results,plasmidname);
         write(results,'  primer: ');
         writename(results, primername);
      end;

      if expectsequence then if length > 0 then begin
         if plasmidname.length <= 0 then begin
            write(results,'  THAT''S ODD!  SEQUENCE WITHOUT A NAME:');
         end;
         write(results,' ');
         l := 1;
         while (l <= length) and (l <= quicksequence) do begin
            write(results,bases[l]);
            l := succ(l)
         end;
         if l < length then write(results,'...');
{zzz}
      end
      else write(results,' NO SEQUENCE DATA FOUND! ******');

      writeln(results);
      d := d^.next 
   end;
   writeln(results);
end;
(* end module tod.reportlanes *)
 
(* begin module tod.writeedit *)
procedure writeedit(var f: text; N, P1, P2: integer; S: sequence);
begin
   write(f,'Editing Lane N=',N:1);
   write(f,' from P1=',P1:1);
   write(f,' to P2=',P2:1);
   write(f,' with S="');
   writeseq(f,S);
   write(f,'"');
   writeln(f);
end;
(* end module tod.writeedit *)

(* begin module tod.writeticseq *)
procedure writeticseq(var f: text; var seq: sequence);
(* write the sequence to file f, over the range start to stop
with tick marks. *)
const
  seqline = 80; (* length of sequence lines *)
var
   l: integer; (* position in sequence *)
procedure tic(number: integer);
(* make the tic marks *)
var
   j: integer; (* tic number *)
   k: integer; (* space number *)
begin
   writeln(f);
   for j := 1 to number do begin
      for k := 1 to 4 do write(f,' ');
      write(f,'''');
      for k := 1 to 1 do write(f,' ');
      write(f,(j*10 + l - seqline):3,'^');
   end;
   writeln(f);
end;
begin
   with seq do begin
      if (start <= stop) and (start > 0) and (stop > 0)
      then begin
         l := start;
         while l <= stop do begin
            write(f, bases[l]);
            if (l mod seqline) = 0 then tic(seqline div 10);
            l := succ(l)
         end;
         (* catch the tic marks for the last part of the sequence *)
         (* bug:  it goes beyond the sequence end.  (But so what?) *)
         l := ((stop + seqline) div seqline) * seqline;
         tic(seqline div 10);
      end
      else writeln(f,'NO-SEQUENCE')
   end;
end;
(* end module tod.writeticseq *)

(* begin module capitalize *)
function capitalize(c: char): char; 
(* convert the character c to upper case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 

   if (n >= ord('a')) and 
      (n <= ord('z'))   
   then c := chr( n - ord('a') + ord('A'));

   capitalize := c
end; 
(* end module capitalize version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module decapitalize *)
function decapitalize(c: char): char; 
(* convert the character c to lower case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 

   if (n >= ord('A')) and 
      (n <= ord('Z'))   
   then c := chr( n - ord('A') + ord('a'))
   else c := chr(n);

   decapitalize := c
end; 
(* end module decapitalize version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module tod.nxseq *)
procedure nxseq(var seq: sequence);
(* convert n's to x's in the sequence *)
var
   l: integer; (* position in sequence *)
begin
   with seq do begin
      for l := 1 to seq.length
      do if bases[l] = 'n' then bases[l] := 'x'
   end;
end;
(* end module tod.nxseq *)

(* begin module tod.basecomp *)
function basecomp(c: char): char;
(* return the base complement of the character c.
No change if not identified. *)
begin
        if c = 'a' then basecomp := 't'
   else if c = 'c' then basecomp := 'g'
   else if c = 'g' then basecomp := 'c'
   else if c = 't' then basecomp := 'a'
   else if c = 'x' then basecomp := 'x'
   else basecomp := c
end;
(* end module tod.basecomp *)

(* begin module tod.complement *)
procedure complement(var seq: sequence);
(* complement the sequence *)
var
   hold: sequence; (* the complement held *)
   lower: integer; (* position in sequence *)
   upper: integer; (* position in sequence *)
begin
   for lower := 1 to seq.length do begin
      upper := seq.length-lower+1;
      hold.bases[upper] := basecomp(seq.bases[lower]);
   end;
   for lower := 1 to seq.length do begin
      seq.bases[lower] := hold.bases[lower]
   end;
end;
(* end module tod.complement *)

(* begin module tod.asymmetric *)
function asymmetric(var seq: sequence): boolean;
(* deterimine if the sequence is asymmetric *)
var
   lower: integer; (* position in sequence *)
   crooked: boolean; (* true if the sequence is assymetric *)
   upper: integer; (* position in sequence *)
begin
   if (seq.length div 2) * 2 <> seq.length
   then crooked := true
   else begin
      crooked := false;
      for lower := 1 to seq.length div 2 do begin
         upper := seq.length-lower+1;
         if seq.bases[upper] <> basecomp(seq.bases[lower])
         then crooked := true
      end;
   end;
   asymmetric := crooked
end;
(* end module tod.asymmetric *)

(* begin module tod.upperseq *)
procedure upperseq(var seq: sequence);
(* convert the sequence to upper case *)
var
   l: integer; (* position in sequence *)
begin
   with seq do begin
      for l := 1 to length
      do bases[l] := capitalize(bases[l]);
   end;
end;
(* end module tod.upperseq *)

(* begin module tod.lowerseq *)
procedure lowerseq(var seq: sequence);
(* convert the sequence to lower case *)
var
   l: integer; (* position in sequence *)
begin
   with seq do begin
      for l := 1 to length
      do bases[l] := decapitalize(bases[l]);
   end;
end;
(* end module tod.lowerseq *)

(* begin module tod.locate *)
procedure locate(r: trigger; seq: sequence; var l, finds: integer);
(* locate the trigger r in sequence seq and report its location in l.
finds is the number of times the trigger shot *)
var
   p: integer; (* position in sequence *)
begin
   l := -1;
   with seq do begin
      resettrigger(r);
      finds := 0;
      for p := 1 to seq.length do begin
write(results,seq.bases[p]);
         testfortrigger(seq.bases[p], r);

(* if we missed, try again *)
(* if r.skip then testfortrigger(seq.bases[p], r);*)
write(results,'_');
if r.skip  then write(results,'S') else write(results,'s');
if r.found then write(results,'F') else write(results,'f');
write(results,r.state:1,'_');

         if r.found then begin
            finds := finds + 1;
            l := p;
{did not help: }
            resettrigger(r);
         end
      end;
   end;
writeln(results,'   what we saw l=',l:1,' finds = ',finds:1);
end;
(* end module tod.locate *)

(* begin module tod.listlocations *)
procedure listlocations(var f: text; l: locptr);
(* list the locations given in l *)
var
   location: locptr; (* index to l *)
begin
   location := l;
   while location <> nil do begin
      write(f,' ',location^.spot:1);
      location := location^.next
   end;
end;
(* end module tod.listlocations *)

(* begin module tod.find *)
procedure find(    r: sequence;
               var seq: sequence;
               var allspots: locptr;
               var finds: integer;
                   direction: integer);
(* Find the pattern r in sequence seq and report its location in the
list allspots.
finds is the number of times that r was found.
direction defines the direction of the search: +1 is forward, -1 is backward.
Located sequences r are made capitalized in seq. *)
var
   d: integer; (* position in sequence, forward search *)
   l: locptr; (* current location recorded *)
   last: locptr; (* location of the previous location record *)
   lower: char; (* lower case of the sequence at d *)
   matching: boolean; (* sequence matches pattern *)
   p: integer; (* position in sequence, forward or backward *)
   j: integer; (* position in pattern r *)
begin

{
write(results,'find: search for ');
writeseq(results,r);
writeln(results);
}

   new(allspots);
   allspots^.spot := -1;
   allspots^.next := nil;
{
   new(allspots^.next);
   allspots^.next^.next := nil;
   allspots^.next^.spot := -1;
}
   l := allspots;
   last := l;
{
write(results,'START: ');listlocations(results, allspots); writeln(results);
}
   with seq do begin
      finds := 0;
      for d := 1 to seq.length do begin
         if direction = -1 then p := length - d + 1
                           else p := d;
         matching := true;
         j := 0;
{
write(results,p:1);
}
         while matching and (j < r.length) do begin
            (* conversion to lower case is necessary to assure that
               overlapping sequences are detected properly *)
            lower := decapitalize(seq.bases[p+j]);
            if lower <> r.bases[j+1] then matching := false;
{
write(results,seq.bases[p+j]);
if matching then write(results,'=')
            else write(results,'<>');
write(results,r.bases[j+1],'|');
}
            j := succ(j)
         end;

         if matching then begin
            (* record the finding *)
            finds := finds + 1;
            l^.spot := p;

            (* move into the next empty link *)
            new(l^.next); (* make a blank link for next time *)
            last := l;
            l := l^.next;
            l^.spot := -1;
            l^.next := nil;
{
write(results,'after finding: ');
listlocations(results, allspots); writeln(results,'!');
}
            (* capitalize the found sequence *)
            for j := 0 to r.length-1
            do seq.bases[d+j] := capitalize(seq.bases[d+j]);
         end
      end;
   end;
{
   (* remove the last link, which never had anything put in it *)
   if l = last^.next then writeln(results,'l EQUAL last')
                     else writeln(results,'l NOT EQUAL last');
}
   if last^.next <> nil then begin
      dispose(last^.next); (* ie dispose of l, keep last *)
      last^.next := nil;
   end;
(*
   l^.next := nil;
*)

{
write(results,'GOT: '); listlocations(results,allspots); writeln(results);
writeln(results,'   what we saw l=',l:1,' finds = ',finds:1);
}
end;
(* end module tod.find *)

(* begin module tod.readparameters *)
procedure readparameters(var results: text;
                         var todp: text;
                         var r1,r2: sequence);
(* read the parameters from todp *)
begin
   reset(todp);
   clearseq(r1); readaseq(todp,r1);
   clearseq(r2); readaseq(todp,r2);
   writeln(results);
   write(results,'Parameter (todp) says to grab sequences between ');
   upperseq(r1);
   writeseq(results,r1); (* ,1,r1.length); *)
   write(results,' and ');
   lowerseq(r1);
   upperseq(r2);
   writeseq(results,r2); (* ,1,r2.length); *)
   writeln(results);
   lowerseq(r2);
end;
(* end module tod.readparameters *)

(* begin module tod.coloncheck *)
procedure coloncheck(var abi: text);
(* check that the line has colons as it should if everything is OK *)
begin
   while not eoln(abi) do begin
      if abi^ <> ':' then begin
         writeln(output,'BAD RAW FILE:  MISSING COLONS LINE');
         writeln(output,'character found: ',abi^);
         halt
      end
      else get(abi)
   end;
   readln(abi)
end;
(* end module tod.coloncheck *)

(* begin module tod.readsequences *)
procedure readsequences(var results: text;
                        var abi: text;
                        var data: seqptr);
(* read the sequences from abi into the data structure. *)
var
   found: boolean; (* whether the lane number in abi was found in the ssb *)
   number: integer; (* the lane number of the sequence read in *)

   seq: seqptr; (* a sequence from abi *)
begin
   reset(abi);
   number := 0;
   (* first we expect the first line to be ":" *)
   coloncheck(abi);
   while not eof(abi) do begin
      (* read a sequence in *)
      (* obtain the lane number *)
      while not (abi^ in ['0','1','2','3','4','5','6','7','8','9'])
         do get(abi);
      readln(abi,number);

      (* try to locate this number in the data set! *)
      seq := data;
      found := false;
      while not found do begin
         if number = seq^.lanenumber
         then found := true
         else begin
(* this portion is probably no longer necessary,
as the entire lane sheet is stored. *)
            if seq^.next = nil then begin
               writeln(results,'SEQUENCE ',number:1,' HAS NO CORRESPONDING',
                              ' DATA IN THE SAMPLE SHEET!');
               writeln(results,'Creating new lane, with empty names (etc)');
               new(seq^.next);
               seq := seq^.next;
               clearseq(seq^);
               with seq^ do begin
                  lanenumber := number;
               end;
               found := true
            end
            else seq := seq^.next
         end
      end;

      coloncheck(abi);
      readaseq(abi,seq^);
   end;
end;
(* end module tod.readsequences *)

(* begin module tod.edit *)
procedure edit(var results: text;
               var edits: text;
               var data: seqptr);
(* edit the sequences in the data file. *)
var
   delta: integer; (* size of deleted region *)
   done: boolean; (* done searching for the next lane *)
{
   finished: boolean; (* finished editing *)
}
   N: integer; (* lane number to edit *)
   ok: boolean; (* if all editing commands check out ok, then do the edit *)
   P1, P2: integer; (* positions between which to do the editing *)
   S: sequence; (* a sequence to insert for editing *)
   seq: seqptr; (* a sequence in the data to be edited *)

procedure copycomments(var edits, results: text);
(* copy comments below the command immediately to results *)
var
   havecomments: boolean; (* have comments in edits to print out *)
begin
   havecomments := true;
   while havecomments do begin
      if eof(edits)
      then havecomments := false
      else if edits^ = '*'
           then copyaline(edits,results)  (* copy comments to results *)
           else havecomments := false
   end;
end; (* copycomments *)

procedure editoperation;
var
   p: integer; (* current editing position *)
begin (* editoperation *)
   (* obtain the editing commands *)
   read(edits, N, P1, P2);
   clearseq(S);
   readaseq(edits,S);

   writeedit(results,N,P1,P2,S);
   copycomments(edits,results);

   (* check the editing commands *)
   ok := true;  (* innocent until proven guilty ... *)
   if P1 > P2 then begin
      writeln(results,'Second position must be greater than the first,');
      writeln(results,'but you said P1 = ',P1:1,' > ',P2:1,' = P2.');
      ok := false (* guilty! *)
   end;

   (* force deletion if the "sequence" has a 'd' *)
   if S.bases[1] = 'd' then begin
      writeln(results,'This will be a pure deletion.');
      S.length := 0;
   end;

   (* locate the lane N *)
   seq := data;
   done := false;
   while (seq <> nil) and (not done) do begin
      if seq^.lanenumber = N
      then begin
         if seq^.length > 0
         then done := true
         else begin
            writeln(results,'Sample of Lane N =',
                             N:1,' exists, but has no sequence.');
            ok := false;
            done := true
         end;
      end
      else seq := seq^.next
   end;

   if seq = nil then begin
      writeln(results,'Sample of Lane N = ',N:1,' was not found.');
      ok := false;
   end;

   if ok then with seq^ do begin
      (* Tell them what we are about to edit *)
      writeln(results,'The sequence to edit is length = ',length:1,':');
      writeticseq(results,seq^);
      writeln(results);

      (* make edits at the edges cleaner *)
      if P1 > length then begin (* both off the high end *)
         P1 := length;
         P2 := P1 + 1;
      end
      else if P2 > length then begin (* P1 inside, P2 after end *)
         P2 := length+1;
         if P1 < 1 then P1 := 0; (* both were outside the fragment *)
      end;
      if P2 < 1 then begin (* both off the low end *)
         P2 := 1;
         P1 := P2 - 1;
      end
      else if (P2 > 1) and (P1 < 0) then begin (* P2 inside, P1 before end *)
         P1 := 0;
      end;

      (* We are ready to do the edit! *)

      (* delete *)
      if P2 > P1 + 1 then begin
         write(results,'The sequence being deleted is: "');
         if P2 - P1 > length
         then begin
            write(results,'(entire sequence)');
            editdeleted := true
         end
         else for p := P1 + 1 to P2 - 1 do write(results,bases[p]);
         writeln(results,'"');
       
         delta := - P2 + P1 + 1;
         for p := P2 to length do bases[p+delta] := bases[p];
         length := length + delta;
         stop := length;
         writeln(results,'After the ',-delta:1,
                        ' base DELETION we have length = ',length:1,':');
         writeticseq(results,seq^);
         writeln(results);
      end;

      (* insert *)
      if S.length > 0 then begin
         (* move the sequence to make space for the insert *)
         for p := length downto P1 + 1 do bases[p+S.length] := bases[p];
         (* insert the new sequence *)
         for p := 1 to S.length do bases[P1+p] := S.bases[p];
         length := length + S.length;
         stop := length;
         writeln(results,'After the ',S.length:1,
                        ' base INSERTION we have length = ',length:1,':');
         writeticseq(results,seq^);
         writeln(results);
      end;
   end
   else begin
      writeln(results,'THESE EDIT COMMANDS WERE IGNORED.');
   end;
   writeln(results);
end; (* editoperation *)

begin
   writeln(results);

   bar(results,80,'*'); writeln(results); (* all bars are 80 wide *)
   write(results,'* EDITING THE SAMPLES ********');
   bar(results,50,'*'); writeln(results); (* all bars are 80 wide *)
   copycomments(edits,results);

   while not eof(edits) do editoperation;

{
   while not eof(edits) do begin
      (* skip comments *)
      finished := false;
      while not finished do begin
         if eof(edits)
         then finished := true (* no commands left *)
         else if edits^ = '*'
         then copyaline(edits,results)  (* copy comments to results *)
         else editoperation             (* start editing *)
      end;
   end;
}
   writeln(results,'DONE EDITING THE SAMPLES');
   writeln(results);
end;
(* end module tod.edit *)

(* begin module tod.process *)
procedure process(var results: text; r1, r2: sequence;
                  var data: seqptr);
(* Process the sequences.  Only record data between the locations of the marker
sequences r1 and r2. *)
var
   chopout: boolean; (* whether to chop out the region between r1 and r2 *)
   flip1: boolean; (* r1 is not self complementary, so flip it for searching *)
   flip2: boolean; (* r2 is not self complementary, so flip it for searching *)
   r1finds, r2finds: integer; (* number of times the triggers were found *)
   r1location, r2location: locptr; (* locations of the triggers *)
   seq: seqptr; (* a sequence in the data *)
procedure doasequence;
begin
   (* begin to process the sequence *)
   lowerseq(seq^);
   nxseq(seq^);

   find(r1,seq^,r1location,r1finds,+1);
   if flip1 then if r1finds = 0 then begin
      writeseq(results,r1);
      writeln(results,' not found; trying its complement');
      complement(r1);
      find(r1,seq^,r1location,r1finds,+1);
   end;

   find(r2,seq^,r2location,r2finds,+1);
   if flip2 then if r2finds = 0 then begin
      writeseq(results,r2);
      writeln(results,' not found; trying its complement');
      complement(r2);
      find(r2,seq^,r2location,r2finds,+1);
   end;

   write(results,'- LOCATIONS OF MARKS ---------');
   bar(results,50,'-'); writeln(results);
   writeticseq(results,seq^); writeln(results);
   bar(results,80,'-'); writeln(results);

   if r1finds = 0 then begin
      write(results,'NO copies of ');
      writeseq(results,r1);
      writeln(results,' found');
   end
   else begin (* r1finds > 0 *)
      if r1finds = 1 then begin
         writeseq(results,r1);
         write(results,' is');
      end
      else begin (* r1finds > 1 *)
         write(results,r1finds:1,' copies of ');
         writeseq(results,r1);
         write(results,' found');
      end;
      write(results,' at');
      listlocations(results, r1location);
      writeln(results);
   end;

   if r2finds = 0 then begin
      write(results,'NO copies of ');
      writeseq(results,r2);
      writeln(results,' found');
   end
   else begin (* r2finds > 0 *)
      if r2finds = 1 then begin
         writeseq(results,r2);
         write(results,' is');
      end
      else begin (* r2finds > 1 *)
         write(results,r2finds:1,' copies of ');
         writeseq(results,r2);
         write(results,' found');
      end;
      write(results,' at');
      listlocations(results, r2location);
      writeln(results);
   end;

   chopout := (r1finds = 1) and (r2finds = 1);

   if chopout then begin
      if r1location^.spot > r2location^.spot then begin
         bar(results,80,'-'); writeln(results);
         writeln(results,'Taking the complement');
         lowerseq(seq^);
         complement(seq^);
         if flip1 then complement(r1);
         if flip2 then complement(r2);
         find(r1,seq^,r1location,r1finds,+1);
         find(r2,seq^,r2location,r2finds,+1);
         writeseq(results,seq^); writeln(results);
         bar(results,80,'='); writeln(results);

         chopout := (r1finds = 1) and (r2finds = 1);
         if not chopout then begin
            writeln(results,'I can''t find ');
            writeseq(results,r1);
            write(results,' or ');
            writeseq(results,r2);
            write(results,' in the complement of the sequence!');
            writeln(results);
            halt;

(*
            writeln(results,' I''ll take the complement of them also ...');
            complement(seq^);
            complement(r1);
            complement(r2);
            find(r1,seq^,r1location,r1finds,+1);
            find(r2,seq^,r2location,r2finds,+1);
            chopout := (r1finds = 1) and (r2finds = 1);
            if not chopout then begin
               writeln(result,'that failed, something is horribly wrong!');
               halt
            end
            else writeln(output,'found them!');
*)

         end
      end;

      if chopout then begin
         writeln(results,'Both markers were found once so',
                        ' the middle will be chopped out');
         seq^.start := r1location^.spot;
         seq^.stop  := r2location^.spot + r1.length - 1;

      end
   end
   else writeln(results, 'The middle will NOT be chopped out.');

   writeln(results,'=== THE SEQUENCE THAT WILL BE OUTPUT',
                   ' TO THE DB FILE IS: ==');
   writeseq(results,seq^); writeln(results);
   writeln(results,'Overall length: ',(seq^.stop-seq^.start+1):1);
   if chopout then writeln(results,'length less R1 and R2: ',
                   (seq^.stop-seq^.start+1)-r1.length-r2.length:1);

   bar(results,80,'=');
   writeln(results);
   writeln(results);
   lowerseq(seq^);
end; (* doasequence *)
begin (* process *)
   writeln(results);
   writeln(results);
   bar(results,80,'*'); writeln(results); (* all bars are 80 wide *)
   write(results,'* PROCESSING THE SEQUENCES ***');
   bar(results,50,'*'); writeln(results);
   seq := data;

   flip1 := asymmetric(r1);
   flip2 := asymmetric(r2);

   while seq <> nil do begin
      if seq^.plasmidname.length > 0 then begin

         write(results,'Processing Lane Number ',seq^.lanenumber:1,': ');
         write(results,'plasmid: ');
         writename(results,seq^.plasmidname);
         write(results,', primer: ');
         writename(results,seq^.primername);
         writeln(results);

         if seq^.length > 0 then begin
            write(results,'- RAW SEQUENCE FILE ----------');
            bar(results,50,'-'); writeln(results);
            writeticseq(results,seq^); writeln(results);
            doasequence
         end
         else begin
            if seq^.editdeleted
            then begin
               writeln(results,'(The sequence was deleted by editing.)');
               writeln(results);
            end
            else begin
               bar(results,80,'*'); writeln(results);
               write(results,'* MISSING SEQUENCE FILE! *****');
               bar(results,50,'!'); writeln(results);
               bar(results,80,'*'); writeln(results);
               writeln(results);

               bar(output,80,'*'); writeln(output);
               write(output,'* MISSING SEQUENCE FILE! *****');
               bar(output,50,'*'); writeln(output);
               bar(output,80,'*'); writeln(output);
               write(output,'While Processing Lane Number ',seq^.lanenumber:1,': ');
               write(output,'plasmid: ');
               writename(output,seq^.plasmidname);
               write(output,', primer: ');
               writename(output,seq^.primername);
               writeln(output);
            end
         end;
      end;

      seq := seq^.next
   end;
end; (* process *)
(* end module tod.process *)

(* begin module tod.writesummary *)
procedure writesummary(var db: text; data: seqptr);
(* write a summary of the data to file db *)
var
   datum: seqptr; (* a datum *)
begin
   datum := data;
   while datum <> nil do with datum^ do begin
      write(db,lanenumber:2);
      write(db,' ');
      writename(db,plasmidname);
      write(db,' ');
      writename(db,primername);
      write(db,' ');
      writename(db,date);
      write(db,' ');
      writeseq(db,datum^);
      writeln(db);
      datum := datum ^.next
   end;
end;
(* end module tod.writesummary *)

(* begin module tod.writedata *)
procedure writedata(var db: text; data: seqptr);
(* write out the data to file db *)
var
   datum: seqptr; (* a datum *)
begin
   rewrite(db);
   datum := data;
   while datum <> nil do with datum^ do begin
      if length > 0 then begin (* only write complete sequences out *)
         writeseq(db,datum^);
         write(db,' ');
         writename(db,plasmidname);
         write(db,' ');
         writename(db,primername);
         write(db,' ');
         write(db,lanenumber:1);
         write(db,' ');
         writename(db,date);
         writeln(db);
      end;
      datum := datum ^.next
   end;
end;
(* end module tod.writedata *)

(* begin module tod.themain *)
procedure themain(var abi, ssb, thedate, todp, results, summary, db: text);
(* the main procedure of the program *)
var
   data: seqptr; (* the sequence data *)
   r1, r2: sequence; (* the ends of the sequence printed *)
   now: name; (* today's date *)

begin
   writeln(output, 'tod ', version:4:2);
   rewrite(results);
   writeln(results, 'tod ', version:4:2);

   (* get the parameters *)
   readparameters(results, todp, r1, r2);

   (* read the date(s) *)
   readdate(results, thedate, now);

   (* read ssb table *)
   new(data);
   readssb(ssb, now, data);

   (* read the sequences into the data structure *)
   readsequences(results, abi, data);

   writeln(output);
   write(output,'Date/Name: ');
   writename(output,now);
   writeln(output);
   reportlanes(results, data, true);
   reportlanes(output, data, true);

   checkthedate(thedate, ssb);

   (* edit the sequences *)
   edit(results, todp, data);

   (* process the sequences *)
   process(results, r1, r2, data);

   (* write out the data structure *)
   writedata(db, data);

   (* summarize *)
   rewrite(summary);
   writeln(summary, 'SUMMARY: ');
   writesummary(summary, data);

end;
(* end module tod.themain *)

begin
   themain(abi, ssb, thedate, todp, results, summary, db);
1: end.
