program trex(fin, trexp, fout, fresep, frebp, output);
(* trex: transfac extraction

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of trex.p 1997 April 22
origin 1997 April 21 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.trex *)
(*
name
   trex: transfac extraction

synopsis
   trex(fin: in, trexp: in, fout: out, fresep, output: out)

files
   fin:  transfac matrix.dat file

   fout:  a single matrix

   fresep:  the single matrix ready to be input to the frese program.

   frebp:  the single matrix ready to be input to the freb program.

   trexp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      the id of the matrix to extract

   output: messages to the user

description

   Extract a matrix from the transfac database for further processing.

   Id's that end in '_C' are consindex matrices and cannot be used
   because they do not contain n(b,l) data.

examples

documentation

   http://transfac.gbf-braunschweig.de/TRANSFAC/doc31/toc.html
   ftp://transfac.gbf-braunschweig.de/pub/transfac/ascii/

see also
   frese.p, freb.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.trex *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 4.20; (@ of prgmod.p 1997 March 15 *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module trex.type *)
   idtype = array[1..2] of char; (* id for database lines *)
(* end module trex.type *)

var
   fin, (* file used by this program *)
   trexp, (* file used by this program *)
   fout, (* file used by this program *)
   frebp, (* file used by this program *)
   fresep: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* test for equality between two strings at current positions *)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module skipblanks *)
procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
      while (thefile^ = ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
      while (thefile^ <> ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module copynoreturn *)
procedure copynoreturn(var fin, fout: text);
(* copy a line from file fin to file fout
but don't put a carriage return *)
begin (* copynoreturn *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
{
      readln(fin);
}
(*      writeln(fout); *)
end; (* copynoreturn *)
(* end module copynoreturn *)

function plogp(f: real): real;
(* compute p log2 p for f *)
begin
   if f <= 0.0
   then plogp := 0.0
   else plogp := -f * ln(f)/ln(2)
end;

(* begin module trex.copyto *)
procedure copyto(var fin, fout: text; idwanted: idtype);
(* copy from fin to fout until the idwanted is found *)
var
   astring: string; (* a string from fin *)
   id: idtype; (* id for database lines *)
   gotten: boolean; (* a string was gotten *)
begin
   id := '  ';
   while id <> idwanted do begin
      read(fin, id);
      getstring(fin, astring, gotten);
      if not gotten then begin
         writeln(output,'unexpected end of file');
         halt;
      end;
      write(fout,id);
      writestring(fout,astring);
      writeln(fout);
   end;
end;
(* end module trex.copyto *)

(* begin module trex.getidstring *)
procedure getidstring(var fin: text; var thestring: string);
(* get the string from fin *)
var
   gotten: boolean; (* a string was gotten *)
begin
   skipblanks(fin);
   getstring(fin, thestring, gotten);
   if not gotten then begin
      writeln(output,'unexpected end of file');
      halt;
   end;
end;
(* end module trex.getidstring *)


(* begin module trex.testmatrix *)
procedure testmatrix(nb: real; c: char; var bad: boolean);
(* test whether the real number nb is integer or not.
If it is not, then the matrix is a consindex and should not be used. *)
begin
   if nb <> round(nb) then begin
      bad := true;
      writeln(output,'base ',c,' has value ',nb:4:2);
      writeln(output,'It is from a consindex and the matrix is NO GOOD');
   end;
end;
(* end module trex.testmatrix *)

(* begin module trex.themain *)
procedure themain(var fin, trexp, fout, fresep, frebp: text);
(* the main procedure of the program. *)
const
   wid = 7; (* width of output integers *)
   dec = 0; (* decimal of output numbers *)
   giveinfo = false; (* whether to give R computations to fout *)
var
   acstring: string; (* the accession of a matrix *)
   bad: boolean; (* is the matrix a consindex? *)
   iddesired: string; (* the accession of the desired matrix *)
   idstring: string; (* the id of a matrix *)
   count: integer; (* count of loops done *)
   done: boolean; (* done with looping *)
   H: real; (* H function of na, nc, ng, nt *)
   id: idtype; (* id for database lines *)
   i: integer; (* index for printing information curve *)
   na,nc,ng,nt: real; (* the numbers of 4 bases *)
   n: real; (* sum of na, nc, ng, nt *)
   parameterversion: real; (* parameter version number *)
   R: real; (* 2- H *)
begin
   writeln(output,'trex ',version:4:2);
   reset(trexp);
   readln(trexp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   getidstring(trexp, iddesired);
   write(output,'searching for ');
   writestring(output,iddesired);
   writeln(output);

   reset(fin);
   rewrite(fresep);
   rewrite(frebp);
   rewrite(fout);

   writestring(frebp,iddesired);
   writeln(frebp);

   with iddesired do begin
      if (iddesired.letters[length-1] = '_') and
         (iddesired.letters[length-1] = 'C')
      then begin
         writeln(output,'This is a consindex matrix.');
         writeln(output,'It cannot be used.');
         halt;
      end;
   end;

   count := 0;
   done := false;
   while not done do begin
      if eof(fin)
      then done := true
      else begin

         read(fin, id);
{
writeln(fout,'id = ',id);
}
         if id = 'AC' then begin
            getidstring(fin, acstring);
         end

         else if id = 'ID' then begin
            getidstring(fin, idstring);
write(output,'idstring = ');
writestring(output,idstring);
writeln(output);

            if equalstring(iddesired, idstring) then begin
writeln(output,'identified');

               done := true;

               write(fout,'AC  ');
               writestring(fout,acstring);
               writeln(fout);
               writeln(fout,'XX');

               write(fout,'ID  ');
               writestring(fout,idstring);
               writeln(fout);

               copyto(fin,fout,'P0');
               bad := false;
               while id <> 'XX' do begin
                  read(fin, id);
{
writeln(fout,'id = ',id,' inner');
}
                  if id <> 'XX' then begin
                     read(fin, na, nc, ng, nt);


                     if not bad then begin
                        testmatrix(na,'a',bad);
                        testmatrix(nc,'c',bad);
                        testmatrix(ng,'g',bad);
                        testmatrix(nt,'t',bad);
                        if bad then begin
                           writeln(output,'THE MATRIX IS BAD!');
                           writeln(output,'It is a consindex.');
                           rewrite(fout);
                           rewrite(fresep);
                           rewrite(frebp);
                           halt;
                        end;
                     end;

                     write(fout, id, na:wid:dec, nc:wid:dec,
                                     ng:wid:dec, nt:wid:dec);

                     writeln(fresep, id, na:wid:dec, nc:wid:dec,
                                         ng:wid:dec, nt:wid:dec);

                     writeln(frebp, id, na:wid:dec, nc:wid:dec,
                                        ng:wid:dec, nt:wid:dec);

                     copynoreturn(fin,fout);

                     if giveinfo then begin
                        write(fout,' | ');
   
                        n := na+nc+ng+nt;
                        H :=   plogp(na/n)
                             + plogp(nc/n)
                             + plogp(ng/n)
                             + plogp(nt/n);
                        R := 2 - H;  (* no sample size correction *)

                        write(fout,' R ',R:5:2);
                        write(fout,' | ');
                        for i := 0 to round(20*R) do write(fout,'*');
                        write(fout,'|');
                     end;
 
                     writeln(fout);
                  end;
                  readln(fin);
               end;
               writeln(fout,'XX');
               copyto(fin,fout,'//');
            end
         end
         else readln(fin);
      end
   end;

   writeln(output,'DONE');

end;
(* end module trex.themain *)

begin
   themain(fin, trexp, fout, fresep, frebp);
1: end.




























