program wtch(input, output);
(* wtch: what characters are in a file? (unix version)

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

   module libraries: delman, prgmods *)

const
(* begin module version *)
   version = 1.28; (* of wtch.p 2022 Jul 20
2022 Jul 20, 1.28: give line number of warning
2020 Mar 26, 1.27: An extra CR is being recorded!
2018 Jul 27, 1.26: handle null character (ascii 0)
2018 Apr 16, 1.25: handle control-M better
2015 Aug 05, 1.24: show high order index character (needs more work)
2013 Aug 04, 1.23: increase version for mktoolkit test
2013 Aug 04, 1.22: increase version for mktoolkit test
2013 Aug 04, 1.21: increase version for mktoolkit test
2013 Aug 04, 1.20: give example in the description
2011 Mar 03, 1.19: mark all cases
2011 Mar 03, 1.18: ALWAYS put something in the first column
2001 Apr 10, 1.17: add high bit notation
2000 Dec  8, 1.16: add tab notation
2000 Nov  2, 1.12: previous changes
1997 Feb 26, 1.12: previous changes
1996 Feb 13: switch to input/output to fit unix
1985 apr 20
origin 1982 sep 5 *)
(* end module version *)

(* begin module describe.wtch *)
(*
name
   wtch: what characters are in a file? (unix version)

synopsis
   wtch(input: in, output: out)

files
   input: the file to be studied
   output: an alphabetic list of the characters in the file, giving:
      the character,
      the ordinal number of the character (pascal ord function),
      how many such characters are in the file,
      and the percent of the character in the file.
      Comment lines begin with "**".
      If the character is not printable, the column contains '__'
      (two underscores).

description
   Sometimes it is necessary to determine what characters are in a file.
   if the file is very large, it is not possible to do this by hand.

examples

   For this text:

      "Four score and seven years ago our fathers brought forth on this
      continent, a new nation, conceived in Liberty, and dedicated to the
      proposition that all men are created equal."
      3.14159265358979323846264338327950

   in a file named fourpi, the Unix command
      wtch < fourpi
   gives:

** wtch 1.20
** 238 characters found
**
** char ord(char)  count percent
__         13          4    1.68 CR: Carriage Return
__         32         51   21.43 Space
  "        34          2    0.84
  ,        44          3    1.26
  .        46          2    0.84
  0        48          1    0.42
  1        49          2    0.84
  2        50          4    1.68
  3        51          7    2.94
  4        52          3    1.26
  5        53          4    1.68
  6        54          3    1.26
  7        55          2    0.84
  8        56          3    1.26
  9        57          4    1.68
  F        70          1    0.42
  L        76          1    0.42
  a        97         13    5.46
  b        98          2    0.84
  c        99          6    2.52
  d       100          7    2.94
  e       101         18    7.56
  f       102          2    0.84
  g       103          2    0.84
  h       104          6    2.52
  i       105          9    3.78
  l       108          3    1.26
  m       109          1    0.42
  n       110         14    5.88
  o       111         14    5.88
  p       112          2    0.84
  q       113          1    0.42
  r       114         11    4.62
  s       115          6    2.52
  t       116         15    6.30
  u       117          4    1.68
  v       118          2    0.84
  w       119          1    0.42
  y       121          2    0.84

see also
   whatch.p

author
   Thomas Schneider

bugs
   none known

technical notes
   The constant maxchars determines the number of characters accepted.

*)
(* end module describe.wtch *)


(* more constants *)
      maxchars = 255; (* the last character acceptable *)

type
      charident = record (* the character and how many found *)
         ch: char;  (* a character *)
         nu: integer; (* number of characters ch found *)
      end;

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'prgmod 3.96  85 mar 18 tds'; *)

procedure themain(var fin, fout: text);
const
   NULL = 0; (* null character *)
   CR = 13; (* Carriage Return Control-M *)
   LF = 10; (* Line Feed *)
   TAB = 09; (* Tab character *)
   SPACE = 32; (* space character *)
   HIGH = 127; (* high order bit is on if greater than this value *)
var
   chars: array[0..maxchars] of charident; (* information on the chars *)
   index: 0..maxchars; (* an index for chars *)
   count: integer; (* of all the characters *)
   line: integer; (* line number *)
begin (* wtch *)
   (* initialize *)
(*   writeln(output,'* wtch ',version:4:2);*)
(*    reset(fin); *)
(*    rewrite(fout); *)

   writeln(fout,'** wtch ',version:4:2);
(* cannot reset standard input file
   write(fout,'** ');
   if not eof(fin) then copyaline(fin,fout);
   reset(fin);
   writeln(fout,'**');
*)

   (* clear the chars array *)
   for index := 0 to maxchars do with chars[index] do begin
      ch := ' ';
      nu := 0
   end;
   chars[CR].ch := chr(CR);

   (* scan the fin *)
   count := 0;
   line := 0;
   while not eof(fin) do begin
      if eoln(fin)
      then begin
         if not eof(fin) then count := succ(count);
         with chars[CR] do begin
            (* do not count eof as CR! *)
            if nu > 0 then begin
               if not eof(fin) then begin
                  nu := succ(nu);
               end;
            end
            else begin
               nu := 1;
               ch := fin^
            end;
         end;
         readln(fin);
         line := line + 1;
      end
      else begin
         count := succ(count);
         if ord(fin^) > maxchars then begin
            writeln(fout,'** character with ord = ',ord(fin^):1,
                         ' this is greater than maxchars (',maxchars:1,')');
            writeln(output,'** increase maxchars')
         end
         else if ord(fin^) < 0 then begin (* rare but interesting... *)
            writeln(fout,'** character with negative ord: ',ord(fin^):1)
         end
         else begin
            with chars[ord(fin^)] do begin
               if nu > 0
               then nu := succ(nu)
               else begin
                  nu := 1;
                  ch := fin^
               end
            end
         end;
         get(fin)
      end;
   end;

   (* put out report to fout *)
   writeln(fout,'** ',count:1,' characters found');
   writeln(fout,'**');
   if count <> 0 then begin
      writeln(fout,'** char ord(char)  count percent');
      for index := 0 to maxchars do begin
         if chars[index].nu <> 0 then begin
            (* ensure that there is a column even for unprintable chars *)
            if
            (index = NULL) or (* NOTE: this will overwrite sometimes *)
            (index = CR) or (* NOTE: this will overwrite sometimes *)
            (index = LF) or (* NOTE: this may do odd things *)
            (index = TAB) or
            (index = SPACE) or
            (index > HIGH)
            then begin
                write(fout,'__-') (* don't write odd characters out! *)
            end
            else begin
                write(fout,'  '); (* write printable characters! *)
                (* write out the character *)
                write(fout,chars[index].ch);
            end;
            write(fout,'     ',index:5);
            write(fout,'   ',chars[index].nu:8);
            write(fout,' ',100*(chars[index].nu/count):7:2);
            if chars[index].ch <> chr(index) then begin
               if index = CR then begin
                  write(fout,' CR: Carriage Return');
               end
               else begin
                  if not eof(fin) then begin {shut up about end of file!!}
                     write(fout,' warning: program error,');
                     write(fout,' actual character is different');
                     write(fout,' @line=',line:1)
                  end
               end
            end;
            if index = SPACE then begin
               write(fout,' Space');
            end
            else if index = TAB then begin
               write(fout,' TAB');
            end
            else if index = LF  then begin
               write(fout,' LF ');
            end
            else if index = CR  then begin
               write(fout,' CR = control-M');
            end
            else if index = NULL  then begin
               write(fout,' NULL');
            end
            else if index > HIGH then begin
               if not eof(fin) then begin
                  write(fout,' HIGH ORDER BIT IS ON');
                  write(fout,' @line=',line:1)
                  (* This section needs work: *)
                  if index < 32 then begin
                     write(fout,' char-32="',chr(ord(index)+32),'"');
                     write(fout,' @ line ',line:1)
                  end;
                  if index > 128 then begin
                     write(fout,' char-128="',chr(ord(index)-128),'"');
                     write(fout,' @ line ',line:1)
                  end;
               end
            end;
            writeln(fout)
         end
      end
   end
   else writeln(fout,'**');
end;

begin
   themain(input, output);
end. (* wtch *)
